@@ -13,6 +13,7 @@
 #include "mozilla/dom/CloseWatcherManager.h"
 #include "mozilla/dom/Document.h"
 #include "mozilla/dom/UserActivationIPCUtils.h"
+#include "mozilla/dom/WorkerCommon.h"
 #include "mozilla/PermissionDelegateIPCUtils.h"
 #include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/StaticPtr.h"
@@ -239,13 +240,7 @@ bool WindowContext::CanSet(FieldIndex<IDX_IsSecure>, const bool& aIsSecure,
   return CheckOnlyOwningProcessCanSet(aSource);
 }
 
-bool WindowContext::CanSet(FieldIndex<IDX_AllowMixedContent>,
-                           const bool& aAllowMixedContent,
-                           ContentParent* aSource) {
-  return CheckOnlyOwningProcessCanSet(aSource);
-}
-
-bool WindowContext::CanSet(FieldIndex<IDX_HasBeforeUnload>,
+bool WindowContext::CanSet(FieldIndex<IDX_NeedsBeforeUnload>,
                            const bool& aHasBeforeUnload,
                            ContentParent* aSource) {
   return CheckOnlyOwningProcessCanSet(aSource);
@@ -364,6 +359,17 @@ bool WindowContext::CanSet(FieldIndex<IDX_HasActivePeerConnections>, bool,
   return XRE_IsParentProcess() && IsTop();
 }
 
+void WindowContext::ProcessCloseRequest() {
+  MOZ_ASSERT(XRE_IsParentProcess(), "Window must be Global Parent");
+  BrowsingContext* top = mBrowsingContext->Top();
+  top->PreOrderWalk([&](BrowsingContext* aBrowsingContext) {
+    CanonicalBrowsingContext* canonical = aBrowsingContext->Canonical();
+    if (WindowGlobalParent* parent = canonical->GetCurrentWindowGlobal()) {
+      Unused << parent->SendProcessCloseRequest(aBrowsingContext);
+    }
+  });
+}
+
 void WindowContext::RecomputeCanExecuteScripts(bool aApplyChanges) {
   const bool old = mCanExecuteScripts;
   if (!AllowJavascript()) {
@@ -409,8 +415,28 @@ void WindowContext::DidSet(FieldIndex<IDX_SHEntryHasUserInteraction>,
   }
 }
 
+void WindowContext::DidSet(FieldIndex<IDX_HasActivePeerConnections>,
+                           bool aOldValue) {
+  MOZ_ASSERT(
+      TopWindowContext() == this,
+      "IDX_HasActivePeerConnections can only be set on the top window context");
+
+  BrowsingContext* top = mBrowsingContext->Top();
+
+  top->PreOrderWalk([&](BrowsingContext* aBrowsingContext) {
+    WindowContext* windowContext = aBrowsingContext->GetCurrentWindowContext();
+    if (windowContext) {
+      auto* win{windowContext->GetInnerWindow()};
+      if (win && (aOldValue != win->HasActivePeerConnections())) {
+        dom::UpdateWorkersPeerConnections(*win,
+                                          win->HasActivePeerConnections());
+      }
+    }
+  });
+}
+
 void WindowContext::DidSet(FieldIndex<IDX_UserActivationStateAndModifiers>) {
-  MOZ_ASSERT_IF(!IsInProcess(), mUserGestureStart.IsNull());
+  MOZ_ASSERT_IF(!IsInProcess(), mLastActivationTimestamp.IsNull());
   USER_ACTIVATION_LOG("Set user gesture activation 0x%02" PRIu8
                       " for %s browsing context 0x%08" PRIx64,
                       GetUserActivationStateAndModifiers(),
@@ -420,9 +446,9 @@ void WindowContext::DidSet(FieldIndex<IDX_UserActivationStateAndModifiers>) {
         "Set user gesture start time for %s browsing context 0x%08" PRIx64,
         XRE_IsParentProcess() ? "Parent" : "Child", Id());
     if (GetUserActivationState() == UserActivation::State::FullActivated) {
-      mUserGestureStart = TimeStamp::Now();
+      mLastActivationTimestamp = TimeStamp::Now();
     } else if (GetUserActivationState() == UserActivation::State::None) {
-      mUserGestureStart = TimeStamp();
+      mLastActivationTimestamp = TimeStamp();
     }
   }
 }
@@ -552,45 +578,68 @@ bool WindowContext::HasBeenUserGestureActivated() {
 
 const TimeStamp& WindowContext::GetUserGestureStart() const {
   MOZ_ASSERT(IsInProcess());
-  return mUserGestureStart;
+  return mLastActivationTimestamp;
 }
 
+// https://html.spec.whatwg.org/#transient-activation
 bool WindowContext::HasValidTransientUserGestureActivation() {
   MOZ_ASSERT(IsInProcess());
 
   if (GetUserActivationState() != UserActivation::State::FullActivated) {
-    // mUserGestureStart should be null if the document hasn't ever been
+    // mLastActivationTimestamp should be null if the document hasn't ever been
     // activated by user gesture
     MOZ_ASSERT_IF(GetUserActivationState() == UserActivation::State::None,
-                  mUserGestureStart.IsNull());
+                  mLastActivationTimestamp.IsNull());
     return false;
   }
 
-  MOZ_ASSERT(!mUserGestureStart.IsNull(),
-             "mUserGestureStart shouldn't be null if the document has ever "
-             "been activated by user gesture");
+  MOZ_ASSERT(
+      !mLastActivationTimestamp.IsNull(),
+      "mLastActivationTimestamp shouldn't be null if the document has ever "
+      "been activated by user gesture");
+
   TimeDuration timeout = TimeDuration::FromMilliseconds(
       StaticPrefs::dom_user_activation_transient_timeout());
 
+  // "When the current high resolution time given W is greater than or equal to
+  // the last activation timestamp in W, and less than the last activation
+  // timestamp in W plus the transient activation duration, then W is said to
+  // have transient activation."
   return timeout <= TimeDuration() ||
-         (TimeStamp::Now() - mUserGestureStart) <= timeout;
+         (TimeStamp::Now() - mLastActivationTimestamp) <= timeout;
 }
 
+// https://html.spec.whatwg.org/#consume-user-activation
 bool WindowContext::ConsumeTransientUserGestureActivation() {
   MOZ_ASSERT(IsInProcess());
   MOZ_ASSERT(IsCurrent());
 
+  // 1. If W's navigable is null, then return.
+
   if (!HasValidTransientUserGestureActivation()) {
     return false;
   }
 
+  // 2. Let top be W's navigable's top-level traversable.
   BrowsingContext* top = mBrowsingContext->Top();
+
+  // 3. Let navigables be the inclusive descendant navigables of top's active
+  // document.
   top->PreOrderWalk([&](BrowsingContext* aBrowsingContext) {
+    // 4. Let windows be the list of Window objects constructed by taking the
+    // active window of each item in navigables.
     WindowContext* windowContext = aBrowsingContext->GetCurrentWindowContext();
+
+    // 5. For each window in windows, if window's last activation timestamp is
+    // not positive infinity, then set window's last activation timestamp to
+    // negative infinity.
     if (windowContext && windowContext->GetUserActivationState() ==
                              UserActivation::State::FullActivated) {
       auto stateAndModifiers = UserActivation::StateAndModifiers(
           GetUserActivationStateAndModifiers());
+      // Setting UserActivationStateAndModifiers will trigger
+      // DidSet(FieldIndex<IDX_UserActivationStateAndModifiers>),
+      // which in turn updates mLastActivationTimestamp.
       stateAndModifiers.SetState(UserActivation::State::HasBeenActivated);
       Unused << windowContext->SetUserActivationStateAndModifiers(
           stateAndModifiers.GetRawData());
@@ -603,19 +652,32 @@ bool WindowContext::ConsumeTransientUserGestureActivation() {
 // https://html.spec.whatwg.org/multipage/interaction.html#history-action-activation
 bool WindowContext::HasValidHistoryActivation() const {
   MOZ_ASSERT(IsInProcess());
-  return mHistoryActivation != mUserGestureStart;
+  return mHistoryActivation != mLastActivationTimestamp;
 }
 
 // https://html.spec.whatwg.org/multipage/interaction.html#consume-history-action-user-activation
+// Step 1-2
 void WindowContext::ConsumeHistoryActivation() {
   MOZ_ASSERT(IsInProcess());
 
-  if (!HasValidHistoryActivation()) {
-    return;
-  }
+  // 1. If W's navigable is null, then return.
 
-  mHistoryActivation = mUserGestureStart;
-  return;
+  // 2. Let top be W's navigable's top-level traversable.
+  RefPtr<BrowsingContext> top = mBrowsingContext->Top();
+
+  // Consuming a history activation must happen across all child processes,
+  // including for example cross-origin iframes. As such we need to send an
+  // message over the IPC boundary to ensure out of processes contexts also
+  // consume their activations.
+  MOZ_ASSERT(XRE_IsContentProcess());
+  ContentChild::GetSingleton()->SendConsumeHistoryActivation(top);
+
+  // Update the local process children immediately.
+  top->ConsumeHistoryActivation();
+}
+
+void WindowContext::UpdateLastHistoryActivation() {
+  mHistoryActivation = mLastActivationTimestamp;
 }
 
 bool WindowContext::GetTransientUserGestureActivationModifiers(
