@@ -5,10 +5,13 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "CookieStore.h"
-#include "CookieStoreChild.h"
-#include "CookieStoreNotifier.h"
-#include "CookieStoreNotificationWatcherWrapper.h"
 
+#include "CookieStoreChild.h"
+#include "CookieStoreNotificationWatcherWrapper.h"
+#include "CookieStoreNotifier.h"
+#include "ThirdPartyUtil.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/StorageAccess.h"
 #include "mozilla/dom/Document.h"
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/WorkerCommon.h"
@@ -16,13 +19,13 @@
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/PBackgroundChild.h"
 #include "mozilla/net/CookieCommons.h"
-#include "mozilla/StorageAccess.h"
+#include "mozilla/net/CookiePrefixes.h"
+#include "mozilla/net/NeckoChannelParams.h"
 #include "nsICookie.h"
 #include "nsIGlobalObject.h"
 #include "nsIPrincipal.h"
 #include "nsReadableUtils.h"
 #include "nsSandboxFlags.h"
-#include "ThirdPartyUtil.h"
 
 using namespace mozilla::net;
 
@@ -30,6 +33,15 @@ namespace mozilla::dom {
 
 namespace {
 
+int64_t ComputeExpiry(const CookieInit& aOptions) {
+  if (aOptions.mExpires.IsNull()) {  // Session cookie
+    return INT64_MAX;
+  }
+
+  return CookieCommons::MaybeCapExpiry(PR_Now() / PR_USEC_PER_MSEC,
+                                       aOptions.mExpires.Value());
+}
+
 int32_t SameSiteToConst(const CookieSameSite& aSameSite) {
   switch (aSameSite) {
     case CookieSameSite::Strict:
@@ -48,13 +60,35 @@ bool ValidateCookieNameOrValue(const nsAString& aStr) {
     if (*iter == 0x3B || *iter == 0x7F || (*iter <= 0x1F && *iter != 0x09)) {
       return false;
     }
-    if (*iter == 0x20) {
-      return false;
-    }
   }
   return true;
 }
 
+const nsDependentSubstring TrimTabAndSpace(const nsAString& aStr) {
+  nsAString::const_iterator start, end;
+
+  aStr.BeginReading(start);
+  aStr.EndReading(end);
+
+  auto TabOrSpace = [](char16_t ch) { return ch == 0x09 || ch == 0x20; };
+
+  while (start != end && TabOrSpace(*start)) {
+    ++start;
+  }
+
+  while (end != start) {
+    --end;
+
+    if (!TabOrSpace(*end)) {
+      ++end;
+
+      break;
+    }
+  }
+
+  return Substring(start, end);
+}
+
 bool ValidateCookieNameAndValue(const nsAString& aName, const nsAString& aValue,
                                 Promise* aPromise) {
   MOZ_ASSERT(aPromise);
@@ -69,6 +103,11 @@ bool ValidateCookieNameAndValue(const nsAString& aName, const nsAString& aValue,
     return false;
   }
 
+  if (aName.Contains('=')) {
+    aPromise->MaybeRejectWithTypeError("Cookie name cannot contain '='");
+    return false;
+  }
+
   if (aName.IsEmpty() && aValue.Contains('=')) {
     aPromise->MaybeRejectWithTypeError(
         "Cookie value cannot contain '=' if the name is empty");
@@ -90,21 +129,15 @@ bool ValidateCookieNameAndValue(const nsAString& aName, const nsAString& aValue,
   return true;
 }
 
-bool HasSecurePrefix(const nsAString& aString) {
-  return StringBeginsWith(aString, u"__Secure-"_ns,
-                          nsCaseInsensitiveStringComparator);
-}
-
-bool HasHostPrefix(const nsAString& aString) {
-  return StringBeginsWith(aString, u"__Host-"_ns,
-                          nsCaseInsensitiveStringComparator);
-}
-
 bool ValidateCookieDomain(nsIPrincipal* aPrincipal, const nsAString& aName,
-                          const nsAString& aDomain, Promise* aPromise) {
+                          const nsAString& aDomain, nsAString& aRetDomain,
+                          Promise* aPromise) {
   MOZ_ASSERT(aPromise);
 
-  if (aDomain.IsEmpty()) {
+  aRetDomain.Assign(aDomain);
+  ToLowerCase(aRetDomain);
+
+  if (aRetDomain.IsEmpty()) {
     return true;
   }
 
@@ -116,30 +149,32 @@ bool ValidateCookieDomain(nsIPrincipal* aPrincipal, const nsAString& aName,
     return false;
   }
 
-  // If the name has a __Host- prefix, then aDomain must be empty.
-  if (HasHostPrefix(aName) && !aDomain.IsEmpty()) {
+  // If the name has a __Host- prefix, then aRetDomain must be empty.
+  if (CookiePrefixes::Has(CookiePrefixes::eHost, aName) &&
+      !aRetDomain.IsEmpty()) {
     aPromise->MaybeRejectWithTypeError(
         "Cookie domain is not allowed for cookies with a __Host- prefix");
     return false;
   }
 
-  if (aDomain[0] == '.') {
+  if (aRetDomain[0] == '.') {
     aPromise->MaybeRejectWithTypeError("Cookie domain cannot start with '.'");
     return false;
   }
 
   NS_ConvertUTF8toUTF16 host(utf8Domain);
-  if (host != aDomain) {
-    if ((host.Length() < aDomain.Length() + 1) ||
-        !StringEndsWith(host, aDomain) ||
-        host[host.Length() - aDomain.Length() - 1] != '.') {
+
+  if (host != aRetDomain) {
+    if ((host.Length() < aRetDomain.Length() + 1) ||
+        !StringEndsWith(host, aRetDomain) ||
+        host[host.Length() - aRetDomain.Length() - 1] != '.') {
       aPromise->MaybeRejectWithTypeError(
           "Cookie domain must domain-match current host");
       return false;
     }
   }
 
-  if (aDomain.Length() > 1024) {
+  if (aRetDomain.Length() > 1024) {
     aPromise->MaybeRejectWithTypeError(
         "Cookie domain size cannot be greater than 1024 bytes");
     return false;
@@ -148,8 +183,7 @@ bool ValidateCookieDomain(nsIPrincipal* aPrincipal, const nsAString& aName,
   return true;
 }
 
-bool ValidateCookiePath(const nsAString& aPath, nsAString& retPath,
-                        Promise* aPromise) {
+bool ValidateCookiePath(const nsAString& aPath, Promise* aPromise) {
   MOZ_ASSERT(aPromise);
 
   if (!aPath.IsEmpty() && aPath[0] != '/') {
@@ -157,18 +191,12 @@ bool ValidateCookiePath(const nsAString& aPath, nsAString& retPath,
     return false;
   }
 
-  nsString path(aPath);
-  if (path.IsEmpty() || path[path.Length() - 1] != '/') {
-    path.Append('/');
-  }
-
-  if (path.Length() > 1024) {
+  if (aPath.Length() > 1024) {
     aPromise->MaybeRejectWithTypeError(
         "Cookie domain size cannot be greater than 1024 bytes");
     return false;
   }
 
-  retPath.Assign(path);
   return true;
 }
 
@@ -180,13 +208,24 @@ bool ValidateCookieNamePrefix(const nsAString& aName, const nsAString& aValue,
                               const nsAString& aPath, Promise* aPromise) {
   MOZ_ASSERT(aPromise);
 
-  if (aName.IsEmpty() && (HasHostPrefix(aValue) || HasSecurePrefix(aValue))) {
+  if (aName.IsEmpty() &&
+      (CookiePrefixes::Has(CookiePrefixes::eHost, aValue) ||
+       CookiePrefixes::Has(CookiePrefixes::eHostHttp, aValue) ||
+       CookiePrefixes::Has(CookiePrefixes::eHttp, aValue) ||
+       CookiePrefixes::Has(CookiePrefixes::eSecure, aValue))) {
     aPromise->MaybeRejectWithTypeError(
         "Nameless cookies should not begin with special prefixes");
     return false;
   }
 
-  if (!HasHostPrefix(aName)) {
+  if (CookiePrefixes::Has(CookiePrefixes::eHttp, aName) ||
+      CookiePrefixes::Has(CookiePrefixes::eHostHttp, aName)) {
+    aPromise->MaybeRejectWithTypeError(
+        "Cookie name should not begin with special prefixes");
+    return false;
+  }
+
+  if (!CookiePrefixes::Has(CookiePrefixes::eHost, aName)) {
     return true;
   }
 
@@ -207,16 +246,11 @@ bool ValidateCookieNamePrefix(const nsAString& aName, const nsAString& aValue,
   return true;
 }
 
-void CookieDataToItem(const CookieData& aData, CookieListItem* aItem) {
-  aItem->mName.Construct(aData.name());
-  aItem->mValue.Construct(aData.value());
-}
-
-void CookieDataToList(const nsTArray<CookieData>& aData,
-                      nsTArray<CookieListItem>& aResult) {
-  for (const CookieData& data : aData) {
+void CookieStructToList(const nsTArray<CookieStruct>& aData,
+                        nsTArray<CookieListItem>& aResult) {
+  for (const CookieStruct& data : aData) {
     CookieListItem* item = aResult.AppendElement();
-    CookieDataToItem(data, item);
+    CookieStore::CookieStructToItem(data, item);
   }
 }
 
@@ -383,23 +417,25 @@ already_AddRefed<Promise> CookieStore::Set(const CookieInit& aOptions,
   NS_DispatchToCurrentThread(NS_NewRunnableFunction(
       __func__, [self = RefPtr(this), promise = RefPtr(promise), aOptions,
                  cookiePrincipal = RefPtr(cookiePrincipal.get())]() {
-        if (!ValidateCookieNameAndValue(aOptions.mName, aOptions.mValue,
-                                        promise)) {
+        nsString name(TrimTabAndSpace(aOptions.mName));
+        nsString value(TrimTabAndSpace(aOptions.mValue));
+
+        if (!ValidateCookieNameAndValue(name, value, promise)) {
           return;
         }
 
-        if (!ValidateCookieDomain(cookiePrincipal, aOptions.mName,
-                                  aOptions.mDomain, promise)) {
+        nsString domain;
+        if (!ValidateCookieDomain(cookiePrincipal, name, aOptions.mDomain,
+                                  domain, promise)) {
           return;
         }
 
-        nsString path;
-        if (!ValidateCookiePath(aOptions.mPath, path, promise)) {
+        if (!ValidateCookiePath(aOptions.mPath, promise)) {
           return;
         }
 
-        if (!ValidateCookieNamePrefix(aOptions.mName, aOptions.mValue,
-                                      aOptions.mDomain, path, promise)) {
+        if (!ValidateCookieNamePrefix(name, value, domain, aOptions.mPath,
+                                      promise)) {
           return;
         }
 
@@ -437,16 +473,13 @@ already_AddRefed<Promise> CookieStore::Set(const CookieInit& aOptions,
         nsCOMPtr<nsIURI> cookieURI = cookiePrincipal->GetURI();
         RefPtr<CookieStoreChild::SetRequestPromise> ipcPromise =
             self->mActor->SendSetRequest(
-                cookieURI, cookiePrincipal->OriginAttributesRef(),
-                thirdPartyContext, partitionForeign, usingStorageAccess,
-                isOn3PCBExceptionList, nsString(aOptions.mName),
-                nsString(aOptions.mValue),
+                mozilla::WrapNotNull(cookieURI.get()),
+                cookiePrincipal->OriginAttributesRef(), thirdPartyContext,
+                partitionForeign, usingStorageAccess, isOn3PCBExceptionList,
+                name, value,
                 // If expires is not set, it's a session cookie.
-                aOptions.mExpires.IsNull(),
-                aOptions.mExpires.IsNull()
-                    ? INT64_MAX
-                    : static_cast<int64_t>(aOptions.mExpires.Value() / 1000),
-                aOptions.mDomain, path, SameSiteToConst(aOptions.mSameSite),
+                aOptions.mExpires.IsNull(), ComputeExpiry(aOptions), domain,
+                aOptions.mPath, SameSiteToConst(aOptions.mSameSite),
                 aOptions.mPartitioned, operationID);
         if (NS_WARN_IF(!ipcPromise)) {
           promise->MaybeResolveWithUndefined();
@@ -459,10 +492,27 @@ already_AddRefed<Promise> CookieStore::Set(const CookieInit& aOptions,
              operationID](
                 const CookieStoreChild::SetRequestPromise::ResolveOrRejectValue&
                     aResult) {
-              if (!aResult.IsResolve() || !aResult.ResolveValue()) {
+              auto cleanupNotificationWatcher = MakeScopeExit([&]() {
                 self->mNotificationWatcher->ForgetOperationID(operationID);
+              });
+
+              if (!aResult.IsResolve()) {
                 promise->MaybeResolveWithUndefined();
+                return;
               }
+
+              const CookieStoreResult& result = aResult.ResolveValue();
+              if (!result.success()) {
+                promise->MaybeRejectWithTypeError("Invalid cookie");
+                return;
+              }
+
+              if (!result.waitForNotification()) {
+                promise->MaybeResolveWithUndefined();
+                return;
+              }
+
+              cleanupNotificationWatcher.release();
             });
       }));
 
@@ -505,26 +555,18 @@ already_AddRefed<Promise> CookieStore::Delete(
   NS_DispatchToCurrentThread(NS_NewRunnableFunction(
       __func__, [self = RefPtr(this), promise = RefPtr(promise), aOptions,
                  cookiePrincipal = RefPtr(cookiePrincipal.get())]() {
-        nsAutoCString baseDomainUtf8;
-        nsresult rv =
-            net::CookieCommons::GetBaseDomain(cookiePrincipal, baseDomainUtf8);
-        if (NS_WARN_IF(NS_FAILED(rv))) {
-          promise->MaybeRejectWithNotAllowedError("Permission denied");
-          return;
-        }
-
+        nsString domain;
         if (!ValidateCookieDomain(cookiePrincipal, aOptions.mName,
-                                  aOptions.mDomain, promise)) {
+                                  aOptions.mDomain, domain, promise)) {
           return;
         }
 
-        nsString path;
-        if (!ValidateCookiePath(aOptions.mPath, path, promise)) {
+        if (!ValidateCookiePath(aOptions.mPath, promise)) {
           return;
         }
 
-        if (!ValidateCookieNamePrefix(aOptions.mName, u""_ns, aOptions.mDomain,
-                                      path, promise)) {
+        if (!ValidateCookieNamePrefix(aOptions.mName, u""_ns, domain,
+                                      aOptions.mPath, promise)) {
           return;
         }
 
@@ -550,7 +592,7 @@ already_AddRefed<Promise> CookieStore::Delete(
         }
 
         nsID operationID;
-        rv = nsID::GenerateUUIDInPlace(operationID);
+        nsresult rv = nsID::GenerateUUIDInPlace(operationID);
         if (NS_WARN_IF(NS_FAILED(rv))) {
           promise->MaybeReject(NS_ERROR_UNEXPECTED);
           return;
@@ -561,10 +603,10 @@ already_AddRefed<Promise> CookieStore::Delete(
         nsCOMPtr<nsIURI> cookieURI = cookiePrincipal->GetURI();
         RefPtr<CookieStoreChild::DeleteRequestPromise> ipcPromise =
             self->mActor->SendDeleteRequest(
-                cookieURI, cookiePrincipal->OriginAttributesRef(),
-                thirdPartyContext, partitionForeign, usingStorageAccess,
-                isOn3PCBExceptionList, nsString(aOptions.mName),
-                nsString(aOptions.mDomain), path, aOptions.mPartitioned,
+                mozilla::WrapNotNull(cookieURI.get()),
+                cookiePrincipal->OriginAttributesRef(), thirdPartyContext,
+                partitionForeign, usingStorageAccess, isOn3PCBExceptionList,
+                aOptions.mName, domain, aOptions.mPath, aOptions.mPartitioned,
                 operationID);
         if (NS_WARN_IF(!ipcPromise)) {
           promise->MaybeResolveWithUndefined();
@@ -762,13 +804,14 @@ already_AddRefed<Promise> CookieStore::GetInternal(
         nsCOMPtr<nsIURI> cookieURI = cookiePrincipal->GetURI();
         RefPtr<CookieStoreChild::GetRequestPromise> ipcPromise =
             self->mActor->SendGetRequest(
-                cookieURI, cookiePrincipal->OriginAttributesRef(),
+                mozilla::WrapNotNull(cookieURI.get()),
+                cookiePrincipal->OriginAttributesRef(),
                 partitionedCookiePrincipal
                     ? Some(partitionedCookiePrincipal->OriginAttributesRef())
                     : Nothing(),
                 thirdPartyContext, partitionForeign, usingStorageAccess,
-                isOn3PCBExceptionList, aOptions.mName.WasPassed(),
-                nsString(name), path, aOnlyTheFirstMatch);
+                isOn3PCBExceptionList, aOptions.mName.WasPassed(), name, path,
+                aOnlyTheFirstMatch);
         if (NS_WARN_IF(!ipcPromise)) {
           promise->MaybeResolveWithUndefined();
           return;
@@ -785,7 +828,7 @@ already_AddRefed<Promise> CookieStore::GetInternal(
               }
 
               nsTArray<CookieListItem> list;
-              CookieDataToList(aResult.ResolveValue(), list);
+              CookieStructToList(aResult.ResolveValue(), list);
 
               if (!aOnlyTheFirstMatch) {
                 promise->MaybeResolve(list);
@@ -822,4 +865,44 @@ Document* CookieStore::MaybeGetDocument() const {
   return nullptr;
 }
 
+// static
+void CookieStore::CookieStructToItem(const CookieStruct& aData,
+                                     CookieListItem* aItem) {
+  aItem->mName.Construct(aData.name());
+  aItem->mValue.Construct(aData.value());
+  aItem->mPath.Construct(aData.path());
+
+  if (aData.host().IsEmpty() || aData.host()[0] != '.') {
+    aItem->mDomain.Construct(VoidCString());
+  } else {
+    aItem->mDomain.Construct(nsDependentCSubstring(aData.host(), 1));
+  }
+
+  if (!aData.isSession()) {
+    aItem->mExpires.Construct(aData.expiry());
+  } else {
+    aItem->mExpires.Construct(nullptr);
+  }
+
+  aItem->mSecure.Construct(aData.isSecure());
+
+  CookieSameSite sameSite = CookieSameSite::None;
+  switch (aData.sameSite()) {
+    case nsICookie::SAMESITE_STRICT:
+      sameSite = CookieSameSite::Strict;
+      break;
+
+    case nsICookie::SAMESITE_LAX:
+      sameSite = CookieSameSite::Lax;
+      break;
+
+    default:
+      // FIXME: lax by default?
+      break;
+  }
+
+  aItem->mSameSite.Construct(sameSite);
+  aItem->mPartitioned.Construct(aData.isPartitioned());
+}
+
 }  // namespace mozilla::dom
