@@ -5,53 +5,99 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "FFmpegDataEncoder.h"
-#include "PlatformEncoderModule.h"
 
 #include <utility>
 
 #include "FFmpegLog.h"
+#include "FFmpegUtils.h"
+#include "PlatformEncoderModule.h"
 #include "libavutil/error.h"
 #include "mozilla/StaticMutex.h"
-
-#include "FFmpegUtils.h"
+#include "mozilla/StaticPrefs_media.h"
 
 namespace mozilla {
 
 template <>
 AVCodecID GetFFmpegEncoderCodecId<LIBAV_VER>(CodecType aCodec) {
 #if LIBAVCODEC_VERSION_MAJOR >= 58
-  if (aCodec == CodecType::VP8) {
-    return AV_CODEC_ID_VP8;
+  if (XRE_IsParentProcess() || XRE_IsContentProcess() ||
+      StaticPrefs::media_use_remote_encoder_video()) {
+    if (aCodec == CodecType::VP8) {
+      return AV_CODEC_ID_VP8;
+    }
+
+    if (aCodec == CodecType::VP9) {
+      return AV_CODEC_ID_VP9;
+    }
+
+    if (aCodec == CodecType::H264) {
+      return AV_CODEC_ID_H264;
+    }
+
+    if (aCodec == CodecType::AV1) {
+      return AV_CODEC_ID_AV1;
+    }
   }
 
-  if (aCodec == CodecType::VP9) {
-    return AV_CODEC_ID_VP9;
-  }
+  if (XRE_IsParentProcess() || XRE_IsContentProcess() ||
+      StaticPrefs::media_use_remote_encoder_audio()) {
+    if (aCodec == CodecType::Opus) {
+      return AV_CODEC_ID_OPUS;
+    }
 
-  if (aCodec == CodecType::H264) {
-    return AV_CODEC_ID_H264;
-  }
-
-  if (aCodec == CodecType::AV1) {
-    return AV_CODEC_ID_AV1;
-  }
-
-  if (aCodec == CodecType::Opus) {
-    return AV_CODEC_ID_OPUS;
-  }
-
-  if (aCodec == CodecType::Vorbis) {
-    return AV_CODEC_ID_VORBIS;
+    if (aCodec == CodecType::Vorbis) {
+      return AV_CODEC_ID_VORBIS;
+    }
   }
 #endif
   return AV_CODEC_ID_NONE;
 }
 
 /* static */
-AVCodec* FFmpegDataEncoder<LIBAV_VER>::FindEncoderWithPreference(
+AVCodec* FFmpegDataEncoder<LIBAV_VER>::FindSoftwareEncoder(
     const FFmpegLibWrapper* aLib, AVCodecID aCodecId) {
   MOZ_ASSERT(aLib);
 
+  // We use this instead of MOZ_USE_HWDECODE because it is possible to disable
+  // support for hardware encoding in Firefox, while the system ffmpeg library
+  // still exposes the hardware codecs.
+#if LIBAVCODEC_VERSION_MAJOR >= 58
+  AVCodec* fallbackCodec = nullptr;
+  void* opaque = nullptr;
+  while (AVCodec* codec = aLib->av_codec_iterate(&opaque)) {
+    if (codec->id != aCodecId || !aLib->av_codec_is_encoder(codec)) {
+      continue;
+    }
+
+    // Prioritize libx264 for now since it's the only h264 codec we tested.
+    // Once libopenh264 is supported, we can simply use the first one we find.
+    if (aCodecId == AV_CODEC_ID_H264 && strcmp(codec->name, "libx264") != 0) {
+      if (!fallbackCodec) {
+        fallbackCodec = codec;
+      }
+      continue;
+    }
+
+    if (codec->capabilities & AV_CODEC_CAP_HARDWARE) {
+      continue;
+    }
+
+    if (codec->capabilities & AV_CODEC_CAP_EXPERIMENTAL) {
+      if (!fallbackCodec) {
+        fallbackCodec = codec;
+      }
+      continue;
+    }
+
+    FFMPEGV_LOG("Using preferred software codec %s", codec->name);
+    return codec;
+  }
+
+  if (fallbackCodec) {
+    FFMPEGV_LOG("Using fallback software codec %s", fallbackCodec->name);
+  }
+  return fallbackCodec;
+#else
   // Prioritize libx264 for now since it's the only h264 codec we tested. Once
   // libopenh264 is supported, we can simply use `avcodec_find_encoder` and
   // rename this function.
@@ -65,6 +111,69 @@ AVCodec* FFmpegDataEncoder<LIBAV_VER>::FindEncoderWithPreference(
   }
 
   return aLib->avcodec_find_encoder(aCodecId);
+#endif
+}
+
+#ifdef MOZ_USE_HWDECODE
+/* static */
+AVCodec* FFmpegDataEncoder<LIBAV_VER>::FindHardwareEncoder(
+    const FFmpegLibWrapper* aLib, AVCodecID aCodecId) {
+  MOZ_ASSERT(aLib);
+
+  AVCodec* fallbackCodec = nullptr;
+  void* opaque = nullptr;
+  while (AVCodec* codec = aLib->av_codec_iterate(&opaque)) {
+    if (codec->id != aCodecId || !aLib->av_codec_is_encoder(codec)) {
+      continue;
+    }
+
+    bool hasHwConfig = codec->capabilities & AV_CODEC_CAP_HARDWARE;
+    if (!hasHwConfig) {
+      for (int i = 0; const AVCodecHWConfig* config =
+                          aLib->avcodec_get_hw_config(codec, i);
+           ++i) {
+        if (config->methods & AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX) {
+          hasHwConfig = true;
+          break;
+        }
+      }
+    }
+
+    if (!hasHwConfig) {
+      continue;
+    }
+
+    if (codec->capabilities & AV_CODEC_CAP_EXPERIMENTAL) {
+      if (!fallbackCodec) {
+        fallbackCodec = codec;
+      }
+      continue;
+    }
+
+    FFMPEGV_LOG("Using preferred hardware codec %s", codec->name);
+    return codec;
+  }
+
+  if (fallbackCodec) {
+    FFMPEGV_LOG("Using fallback hardware codec %s", fallbackCodec->name);
+  }
+  return fallbackCodec;
+}
+#endif
+
+/* static */
+Result<RefPtr<MediaRawData>, MediaResult>
+FFmpegDataEncoder<LIBAV_VER>::CreateMediaRawData(AVPacket* aPacket) {
+  MOZ_ASSERT(aPacket);
+
+  // Copy frame data from AVPacket.
+  auto data = MakeRefPtr<MediaRawData>();
+  UniquePtr<MediaRawDataWriter> writer(data->CreateWriter());
+  if (!writer->Append(aPacket->data, static_cast<size_t>(aPacket->size))) {
+    return Err(MediaResult(NS_ERROR_OUT_OF_MEMORY,
+                           "fail to allocate MediaRawData buffer"_ns));
+  }
+  return data;
 }
 
 StaticMutex FFmpegDataEncoder<LIBAV_VER>::sMutex;
@@ -87,12 +196,6 @@ FFmpegDataEncoder<LIBAV_VER>::FFmpegDataEncoder(
 #endif
 };
 
-RefPtr<MediaDataEncoder::InitPromise> FFmpegDataEncoder<LIBAV_VER>::Init() {
-  FFMPEG_LOG("Init");
-  return InvokeAsync(mTaskQueue, this, __func__,
-                     &FFmpegDataEncoder::ProcessInit);
-}
-
 RefPtr<MediaDataEncoder::EncodePromise> FFmpegDataEncoder<LIBAV_VER>::Encode(
     const MediaData* aSample) {
   MOZ_ASSERT(aSample != nullptr);
@@ -131,16 +234,6 @@ RefPtr<GenericPromise> FFmpegDataEncoder<LIBAV_VER>::SetBitrate(
   return GenericPromise::CreateAndReject(NS_ERROR_NOT_IMPLEMENTED, __func__);
 }
 
-RefPtr<MediaDataEncoder::InitPromise>
-FFmpegDataEncoder<LIBAV_VER>::ProcessInit() {
-  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
-
-  FFMPEG_LOG("ProcessInit");
-  nsresult rv = InitSpecific();
-  return NS_FAILED(rv) ? InitPromise::CreateAndReject(rv, __func__)
-                       : InitPromise::CreateAndResolve(true, __func__);
-}
-
 RefPtr<MediaDataEncoder::EncodePromise>
 FFmpegDataEncoder<LIBAV_VER>::ProcessEncode(RefPtr<const MediaData> aSample) {
   MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
@@ -155,7 +248,9 @@ FFmpegDataEncoder<LIBAV_VER>::ProcessEncode(RefPtr<const MediaData> aSample) {
 
   auto rv = EncodeInputWithModernAPIs(std::move(aSample));
   if (rv.isErr()) {
-    return EncodePromise::CreateAndReject(rv.inspectErr(), __func__);
+    MediaResult e = rv.unwrapErr();
+    FFMPEG_LOG("%s", e.Description().get());
+    return EncodePromise::CreateAndReject(e, __func__);
   }
 
   return EncodePromise::CreateAndResolve(rv.unwrap(), __func__);
@@ -215,6 +310,8 @@ FFmpegDataEncoder<LIBAV_VER>::ProcessDrain() {
 #else
   auto rv = DrainWithModernAPIs();
   if (rv.isErr()) {
+    MediaResult e = rv.unwrapErr();
+    FFMPEG_LOG("%s", e.Description().get());
     return EncodePromise::CreateAndReject(rv.inspectErr(), __func__);
   }
   return EncodePromise::CreateAndResolve(rv.unwrap(), __func__);
@@ -232,31 +329,10 @@ RefPtr<ShutdownPromise> FFmpegDataEncoder<LIBAV_VER>::ProcessShutdown() {
   return ShutdownPromise::CreateAndResolve(true, __func__);
 }
 
-AVCodec* FFmpegDataEncoder<LIBAV_VER>::InitCommon() {
+void FFmpegDataEncoder<LIBAV_VER>::SetContextBitrate() {
   MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
+  MOZ_ASSERT(mCodecContext);
 
-  FFMPEG_LOG("FFmpegDataEncoder::InitCommon");
-
-  AVCodec* codec = FindEncoderWithPreference(mLib, mCodecID);
-  if (!codec) {
-    FFMPEG_LOG("failed to find ffmpeg encoder for codec id %d", mCodecID);
-    return nullptr;
-  }
-  FFMPEG_LOG("found codec: %s", codec->name);
-  mCodecName = codec->name;
-
-  ForceEnablingFFmpegDebugLogs();
-
-  MOZ_ASSERT(!mCodecContext);
-  if (!(mCodecContext = mLib->avcodec_alloc_context3(codec))) {
-    FFMPEG_LOG("failed to allocate ffmpeg context for codec %s", codec->name);
-    return nullptr;
-  }
-
-  return codec;
-}
-
-MediaResult FFmpegDataEncoder<LIBAV_VER>::FinishInitCommon(AVCodec* aCodec) {
   if (mConfig.mBitrateMode == BitrateMode::Constant) {
     mCodecContext->rc_max_rate = static_cast<FFmpegBitRate>(mConfig.mBitrate);
     mCodecContext->rc_min_rate = static_cast<FFmpegBitRate>(mConfig.mBitrate);
@@ -269,20 +345,6 @@ MediaResult FFmpegDataEncoder<LIBAV_VER>::FinishInitCommon(AVCodec* aCodec) {
     FFMPEG_LOG("Encoding in VBR: [%d;%d]", (int)mCodecContext->rc_min_rate,
                (int)mCodecContext->rc_max_rate);
   }
-#if LIBAVCODEC_VERSION_MAJOR >= 60
-  mCodecContext->flags |= AV_CODEC_FLAG_FRAME_DURATION;
-#endif
-
-  AVDictionary* options = nullptr;
-  if (int ret = OpenCodecContext(aCodec, &options); ret < 0) {
-    FFMPEG_LOG("failed to open %s avcodec: %s", aCodec->name,
-               MakeErrorString(mLib, ret).get());
-    return MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
-                       RESULT_DETAIL("avcodec_open2 error"));
-  }
-  mLib->av_dict_free(&options);
-
-  return MediaResult(NS_OK);
 }
 
 void FFmpegDataEncoder<LIBAV_VER>::ShutdownInternal() {
@@ -293,12 +355,41 @@ void FFmpegDataEncoder<LIBAV_VER>::ShutdownInternal() {
   DestroyFrame();
 
   if (mCodecContext) {
-    CloseCodecContext();
-    mLib->av_freep(&mCodecContext);
+    ReleaseCodecContext();
     mCodecContext = nullptr;
   }
 }
 
+Result<AVCodecContext*, MediaResult>
+FFmpegDataEncoder<LIBAV_VER>::AllocateCodecContext(bool aHardware) {
+  AVCodec* codec = nullptr;
+  if (aHardware) {
+#ifdef MOZ_USE_HWDECODE
+    codec = FindHardwareEncoder(mLib, mCodecID);
+#endif
+  } else {
+    codec = FindSoftwareEncoder(mLib, mCodecID);
+  }
+  if (!codec) {
+    return Err(MediaResult(
+        NS_ERROR_DOM_MEDIA_FATAL_ERR,
+        RESULT_DETAIL("failed to find ffmpeg encoder for codec id %d",
+                      mCodecID)));
+  }
+
+  AVCodecContext* ctx = mLib->avcodec_alloc_context3(codec);
+  if (!ctx) {
+    return Err(MediaResult(
+        NS_ERROR_OUT_OF_MEMORY,
+        RESULT_DETAIL("failed to allocate ffmpeg context for codec %s",
+                      codec->name)));
+  }
+
+  MOZ_ASSERT(ctx->codec == codec);
+
+  return ctx;
+}
+
 int FFmpegDataEncoder<LIBAV_VER>::OpenCodecContext(const AVCodec* aCodec,
                                                    AVDictionary** aOptions) {
   MOZ_ASSERT(mCodecContext);
@@ -307,11 +398,18 @@ int FFmpegDataEncoder<LIBAV_VER>::OpenCodecContext(const AVCodec* aCodec,
   return mLib->avcodec_open2(mCodecContext, aCodec, aOptions);
 }
 
-void FFmpegDataEncoder<LIBAV_VER>::CloseCodecContext() {
-  MOZ_ASSERT(mCodecContext);
-
+void FFmpegDataEncoder<LIBAV_VER>::ReleaseCodecContext() {
   StaticMutexAutoLock mon(sMutex);
+  if (!mCodecContext) {
+    return;
+  }
+
+#if LIBAVCODEC_VERSION_MAJOR < 57
   mLib->avcodec_close(mCodecContext);
+  mLib->av_freep(&mCodecContext);
+#else
+  mLib->avcodec_free_context(&mCodecContext);
+#endif
 }
 
 bool FFmpegDataEncoder<LIBAV_VER>::PrepareFrame() {
@@ -354,14 +452,14 @@ void FFmpegDataEncoder<LIBAV_VER>::DestroyFrame() {
 
 // avcodec_send_frame and avcodec_receive_packet were introduced in version 58.
 #if LIBAVCODEC_VERSION_MAJOR >= 58
-Result<MediaDataEncoder::EncodedData, nsresult>
+Result<MediaDataEncoder::EncodedData, MediaResult>
 FFmpegDataEncoder<LIBAV_VER>::EncodeWithModernAPIs() {
   // Initialize AVPacket.
   AVPacket* pkt = mLib->av_packet_alloc();
 
   if (!pkt) {
-    FFMPEG_LOG("failed to allocate packet");
-    return Err(NS_ERROR_OUT_OF_MEMORY);
+    return Err(
+        MediaResult(NS_ERROR_OUT_OF_MEMORY, "failed to allocate packet"_ns));
   }
 
   auto freePacket = MakeScopeExit([this, &pkt] { mLib->av_packet_free(&pkt); });
@@ -372,9 +470,9 @@ FFmpegDataEncoder<LIBAV_VER>::EncodeWithModernAPIs() {
     // buffers is full. In practice this can't happen as we only feed one frame
     // at a time, and we immediately call avcodec_receive_packet right after.
     // TODO: Create a NS_ERROR_DOM_MEDIA_ENCODE_ERR in ErrorList.py?
-    FFMPEG_LOG("avcodec_send_frame error: %s",
-               MakeErrorString(mLib, ret).get());
-    return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+    return Err(MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                           RESULT_DETAIL("avcodec_send_frame error: %s",
+                                         MakeErrorString(mLib, ret).get())));
   }
 
   EncodedData output;
@@ -389,13 +487,20 @@ FFmpegDataEncoder<LIBAV_VER>::EncodeWithModernAPIs() {
     if (ret < 0) {
       // AVERROR_EOF is returned when the encoder has been fully flushed, but it
       // shouldn't happen here.
-      FFMPEG_LOG("avcodec_receive_packet error: %s",
-                 MakeErrorString(mLib, ret).get());
-      return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+      return Err(MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                             RESULT_DETAIL("avcodec_receive_packet error: %s",
+                                           MakeErrorString(mLib, ret).get())));
     }
 
-    RefPtr<MediaRawData> d = ToMediaRawData(pkt);
+    auto r = ToMediaRawData(pkt);
     mLib->av_packet_unref(pkt);
+    if (r.isErr()) {
+      MediaResult e = r.unwrapErr();
+      FFMPEG_LOG("%s", e.Description().get());
+      return Err(e);
+    }
+
+    RefPtr<MediaRawData> d = r.unwrap();
     if (!d) {
       // This can happen if e.g. DTX is enabled
       FFMPEG_LOG("No encoded packet output");
@@ -408,19 +513,19 @@ FFmpegDataEncoder<LIBAV_VER>::EncodeWithModernAPIs() {
   return std::move(output);
 }
 
-Result<MediaDataEncoder::EncodedData, nsresult>
+Result<MediaDataEncoder::EncodedData, MediaResult>
 FFmpegDataEncoder<LIBAV_VER>::DrainWithModernAPIs() {
   MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
   MOZ_ASSERT(mCodecContext);
 
-  // TODO: Create a Result<EncodedData, nsresult> EncodeWithModernAPIs(AVFrame
-  // *aFrame) to merge the duplicate code below with EncodeWithModernAPIs above.
+  // TODO: Create a common utility to merge the duplicate code below with
+  // EncodeWithModernAPIs above.
 
   // Initialize AVPacket.
   AVPacket* pkt = mLib->av_packet_alloc();
   if (!pkt) {
-    FFMPEG_LOG("failed to allocate packet");
-    return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+    return Err(
+        MediaResult(NS_ERROR_OUT_OF_MEMORY, "failed to allocate packet"_ns));
   }
   auto freePacket = MakeScopeExit([this, &pkt] { mLib->av_packet_free(&pkt); });
 
@@ -439,9 +544,9 @@ FFmpegDataEncoder<LIBAV_VER>::DrainWithModernAPIs() {
       return EncodedData();
     }
 
-    FFMPEG_LOG("avcodec_send_frame error: %s",
-               MakeErrorString(mLib, ret).get());
-    return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+    return Err(MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                           RESULT_DETAIL("avcodec_send_frame error: %s",
+                                         MakeErrorString(mLib, ret).get())));
   }
 
   EncodedData output;
@@ -455,16 +560,24 @@ FFmpegDataEncoder<LIBAV_VER>::DrainWithModernAPIs() {
     if (ret < 0) {
       // avcodec_receive_packet should not result in a -EAGAIN once it's in
       // draining mode.
-      FFMPEG_LOG("avcodec_receive_packet error: %s",
-                 MakeErrorString(mLib, ret).get());
-      return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+      return Err(MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                             RESULT_DETAIL("avcodec_receive_packet error: %s",
+                                           MakeErrorString(mLib, ret).get())));
     }
 
-    RefPtr<MediaRawData> d = ToMediaRawData(pkt);
+    auto r = ToMediaRawData(pkt);
     mLib->av_packet_unref(pkt);
+    if (r.isErr()) {
+      MediaResult e = r.unwrapErr();
+      FFMPEG_LOG("%s", e.Description().get());
+      return Err(e);
+    }
+
+    RefPtr<MediaRawData> d = r.unwrap();
     if (!d) {
-      FFMPEG_LOG("failed to create a MediaRawData from the AVPacket");
-      return Err(NS_ERROR_DOM_MEDIA_FATAL_ERR);
+      return Err(
+          MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                      "failed to create a MediaRawData from the AVPacket"_ns));
     }
     output.AppendElement(std::move(d));
   }
@@ -475,57 +588,14 @@ FFmpegDataEncoder<LIBAV_VER>::DrainWithModernAPIs() {
   // TODO: Only re-create AVCodecContext when avcodec_flush_buffers is
   // unavailable.
   ShutdownInternal();
-  nsresult r = InitSpecific();
-  return NS_FAILED(r) ? Result<MediaDataEncoder::EncodedData, nsresult>(
-                            NS_ERROR_DOM_MEDIA_FATAL_ERR)
-                      : Result<MediaDataEncoder::EncodedData, nsresult>(
-                            std::move(output));
+  MediaResult r = InitEncoder();
+  if (NS_FAILED(r.Code())) {
+    FFMPEG_LOG("%s", r.Description().get());
+    return Err(r);
+  }
+
+  return std::move(output);
 }
 #endif  // LIBAVCODEC_VERSION_MAJOR >= 58
 
-RefPtr<MediaRawData> FFmpegDataEncoder<LIBAV_VER>::ToMediaRawDataCommon(
-    AVPacket* aPacket) {
-  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
-  MOZ_ASSERT(aPacket);
-
-  // Copy frame data from AVPacket.
-  auto data = MakeRefPtr<MediaRawData>();
-  UniquePtr<MediaRawDataWriter> writer(data->CreateWriter());
-  if (!writer->Append(aPacket->data, static_cast<size_t>(aPacket->size))) {
-    FFMPEG_LOG("fail to allocate MediaRawData buffer");
-    return nullptr;  // OOM
-  }
-
-  data->mKeyframe = (aPacket->flags & AV_PKT_FLAG_KEY) != 0;
-  // TODO(bug 1869560): The unit of pts, dts, and duration is time_base, which
-  // is recommended to be the reciprocal of the frame rate, but we set it to
-  // microsecond for now.
-  data->mTime = media::TimeUnit::FromMicroseconds(aPacket->pts);
-#if LIBAVCODEC_VERSION_MAJOR >= 60
-  data->mDuration = media::TimeUnit::FromMicroseconds(aPacket->duration);
-#else
-  int64_t duration;
-  if (mDurationMap.Find(aPacket->pts, duration)) {
-    data->mDuration = media::TimeUnit::FromMicroseconds(duration);
-  } else {
-    data->mDuration = media::TimeUnit::FromMicroseconds(aPacket->duration);
-  }
-#endif
-  data->mTimecode = media::TimeUnit::FromMicroseconds(aPacket->dts);
-
-  if (auto r = GetExtraData(aPacket); r.isOk()) {
-    data->mExtraData = r.unwrap();
-  }
-
-  return data;
-}
-void FFmpegDataEncoder<LIBAV_VER>::ForceEnablingFFmpegDebugLogs() {
-#if DEBUG
-  if (!getenv("MOZ_AV_LOG_LEVEL") &&
-      MOZ_LOG_TEST(sFFmpegVideoLog, LogLevel::Debug)) {
-    mLib->av_log_set_level(AV_LOG_DEBUG);
-  }
-#endif  // DEBUG
-}
-
 }  // namespace mozilla
