@@ -8,14 +8,16 @@
 
 #include <utility>
 
+#include "Navigator.h"
+#include "NotificationUtils.h"
 #include "mozilla/Encoding.h"
 #include "mozilla/HoldDropJSObjects.h"
 #include "mozilla/OwningNonNull.h"
 #include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/dom/BindingUtils.h"
 #include "mozilla/dom/Document.h"
-#include "mozilla/dom/Promise.h"
 #include "mozilla/dom/Promise-inl.h"
+#include "mozilla/dom/Promise.h"
 #include "mozilla/dom/RootedDictionary.h"
 #include "mozilla/dom/ServiceWorkerGlobalScopeBinding.h"
 #include "mozilla/dom/WorkerRunnable.h"
@@ -24,8 +26,6 @@
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/BackgroundUtils.h"
 #include "mozilla/ipc/PBackgroundChild.h"
-#include "Navigator.h"
-#include "NotificationUtils.h"
 #include "nsContentPermissionHelper.h"
 #include "nsContentUtils.h"
 #include "nsIContentPermissionPrompt.h"
@@ -37,18 +37,6 @@ namespace mozilla::dom {
 
 using namespace notification;
 
-struct NotificationStrings {
-  const nsString mID;
-  const nsString mTitle;
-  const nsString mDir;
-  const nsString mLang;
-  const nsString mBody;
-  const nsString mTag;
-  const nsString mIcon;
-  const nsString mData;
-  const nsString mServiceWorkerRegistrationScope;
-};
-
 class NotificationPermissionRequest : public ContentPermissionRequestBase,
                                       public nsIRunnable,
                                       public nsINamed {
@@ -404,8 +392,8 @@ already_AddRefed<Notification> Notification::ValidateAndCreate(
   // Step 12: If options["icon"] exists, then parse it using baseURL, and if
   // that does not return failure, set notification’s icon URL to the return
   // value. (Otherwise icon URL is not set.)
-  nsString iconUrl = aOptions.mIcon;
-  ResolveIconURL(aGlobal, iconUrl);
+  nsAutoString iconUrl;
+  ResolveIconURL(aGlobal, aOptions.mIcon, iconUrl);
 
   // Step 19: Set notification’s actions to « ».
   nsTArray<IPCNotificationAction> actions;
@@ -573,49 +561,46 @@ uint32_t Notification::MaxActions(const GlobalObject& aGlobal) {
 }
 
 nsresult Notification::ResolveIconURL(nsIGlobalObject* aGlobal,
-                                      nsString& aIconUrl) {
+                                      const nsAString& aIconUrl,
+                                      nsString& aDecodedUrl) {
   nsresult rv = NS_OK;
 
   if (aIconUrl.IsEmpty()) {
     return rv;
   }
 
-  nsCOMPtr<nsIURI> baseUri = nullptr;
-
-  // XXXnsm If I understand correctly, the character encoding for resolving
-  // URIs in new specs is dictated by the URL spec, which states that unless
-  // the URL parser is passed an override encoding, the charset to be used is
-  // UTF-8. The new Notification icon/sound specification just says to use the
-  // Fetch API, where the Request constructor defers to URL parsing specifying
-  // the API base URL and no override encoding. So we've to use UTF-8 on
-  // workers, but for backwards compat keeping it document charset on main
-  // thread.
-  auto encoding = UTF_8_ENCODING;
-
-  if (nsCOMPtr<nsPIDOMWindowInner> window = aGlobal->GetAsInnerWindow()) {
-    if (RefPtr<Document> doc = window->GetExtantDoc()) {
-      baseUri = doc->GetBaseURI();
-      encoding = doc->GetDocumentCharacterSet();
-    } else {
-      NS_WARNING("No document found for main thread notification!");
-      return NS_ERROR_FAILURE;
-    }
-  } else if (WorkerPrivate* workerPrivate = GetCurrentThreadWorkerPrivate()) {
-    baseUri = workerPrivate->GetBaseURI();
-  }
-
+  nsCOMPtr<nsIURI> baseUri = aGlobal->GetBaseURI();
   if (!baseUri) {
     return rv;
   }
 
+  auto encoding = UTF_8_ENCODING;
+
+  // TODO(krosylight): Ultimately we can use UTF8String for icon definition of
+  // Web IDL when we remove this.
+  if (!StaticPrefs::dom_webnotifications_icon_encoding_utf8_enabled()) {
+    if (nsCOMPtr<nsPIDOMWindowInner> window = aGlobal->GetAsInnerWindow()) {
+      if (RefPtr<Document> doc = window->GetExtantDoc()) {
+        encoding = doc->GetDocumentCharacterSet();
+      } else {
+        NS_WARNING("No document found for main thread notification!");
+        return NS_ERROR_FAILURE;
+      }
+    }
+  }
+
   nsCOMPtr<nsIURI> srcUri;
   rv = NS_NewURI(getter_AddRefs(srcUri), aIconUrl, encoding, baseUri);
   if (NS_SUCCEEDED(rv)) {
     nsAutoCString src;
     srcUri->GetSpec(src);
-    CopyUTF8toUTF16(src, aIconUrl);
+    // XXX(krosylight): We should be able to pass UTF8 as-is, or ideally the URI
+    // object itself.
+    CopyUTF8toUTF16(src, aDecodedUrl);
   }
 
+  // TODO(krosylight): We should be able to remove the following when removing
+  // the non-UTF8 branch above.
   if (encoding == UTF_8_ENCODING) {
     return rv;
   }
@@ -628,12 +613,12 @@ nsresult Notification::ResolveIconURL(nsIGlobalObject* aGlobal,
 
   nsCOMPtr<nsIURI> srcUriUtf8;
   nsresult rvUtf8 =
-      NS_NewURI(getter_AddRefs(srcUri), aIconUrl, UTF_8_ENCODING, baseUri);
+      NS_NewURI(getter_AddRefs(srcUriUtf8), aIconUrl, UTF_8_ENCODING, baseUri);
 
   if (NS_SUCCEEDED(rv)) {
     if (NS_SUCCEEDED(rvUtf8)) {
       bool equals = false;
-      if (NS_SUCCEEDED(baseUri->Equals(srcUri, &equals))) {
+      if (NS_SUCCEEDED(srcUri->Equals(srcUriUtf8, &equals))) {
         if (equals) {
           // Okay to be parsed with UTF8
           label = glean::web_notification::IconUrlEncodingLabel::eUtf8;
