@@ -15,7 +15,6 @@
 
 #include "ETWTools.h"
 #include "GeckoProfiler.h"
-#include "nsRFPService.h"
 #include "PerformanceEntry.h"
 #include "PerformanceMainThread.h"
 #include "PerformanceMark.h"
@@ -26,19 +25,20 @@
 #include "PerformanceWorker.h"
 #include "mozilla/BasePrincipal.h"
 #include "mozilla/ErrorResult.h"
-#include "mozilla/dom/MessagePortBinding.h"
-#include "mozilla/dom/PerformanceBinding.h"
-#include "mozilla/dom/PerformanceEntryEvent.h"
-#include "mozilla/dom/PerformanceNavigationBinding.h"
-#include "mozilla/dom/PerformanceObserverBinding.h"
-#include "mozilla/dom/PerformanceNavigationTiming.h"
 #include "mozilla/IntegerPrintfMacros.h"
 #include "mozilla/Perfetto.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/TimeStamp.h"
+#include "mozilla/dom/MessagePortBinding.h"
+#include "mozilla/dom/PerformanceBinding.h"
+#include "mozilla/dom/PerformanceEntryEvent.h"
+#include "mozilla/dom/PerformanceNavigationBinding.h"
+#include "mozilla/dom/PerformanceNavigationTiming.h"
+#include "mozilla/dom/PerformanceObserverBinding.h"
 #include "mozilla/dom/WorkerPrivate.h"
 #include "mozilla/dom/WorkerRunnable.h"
 #include "mozilla/dom/WorkerScope.h"
+#include "nsRFPService.h"
 
 #define PERFLOG(msg, ...) printf_stderr(msg, ##__VA_ARGS__)
 
@@ -590,22 +590,38 @@ static std::string GetMarkerFilename() {
   return s.str();
 }
 
-std::pair<TimeStamp, TimeStamp> Performance::GetTimeStampsForMarker(
+Maybe<std::pair<TimeStamp, TimeStamp>> Performance::GetTimeStampsForMarker(
     const Maybe<const nsAString&>& aStartMark,
     const Optional<nsAString>& aEndMark,
-    const Maybe<const PerformanceMeasureOptions&>& aOptions, ErrorResult& aRv) {
+    const Maybe<const PerformanceMeasureOptions&>& aOptions) {
+  ErrorResult err;
   const DOMHighResTimeStamp unclampedStartTime = ResolveStartTimeForMeasure(
-      aStartMark, aOptions, aRv, /* aReturnUnclamped */ true);
+      aStartMark, aOptions, err, /* aReturnUnclamped */ true);
   const DOMHighResTimeStamp unclampedEndTime =
-      ResolveEndTimeForMeasure(aEndMark, aOptions, aRv, /* aReturnUnclamped */
-                               true);
+      ResolveEndTimeForMeasure(aEndMark, aOptions, /* aReturnUnclamped */
+                               err, true);
+
+  if (err.Failed()) {
+    return Nothing();
+  }
+
+  // Performance.measure() can receive user-supplied timestamps and those
+  // timestamps might not be relative to 'navigation start'. This is
+  // (potentially) valid but, if we treat them as relative, we will end up
+  // placing them far into the future which causes problems for the profiler
+  // later so we report that as an error. (See bug 1925191 for details.)
+  // kMaxFuture_ms represents approximately 10 years worth of milliseconds.
+  static constexpr double kMaxFuture_ms = 31536000000.0;
+  if (unclampedStartTime > kMaxFuture_ms || unclampedEndTime > kMaxFuture_ms) {
+    return Nothing();
+  }
 
   TimeStamp startTimeStamp =
       CreationTimeStamp() + TimeDuration::FromMilliseconds(unclampedStartTime);
   TimeStamp endTimeStamp =
       CreationTimeStamp() + TimeDuration::FromMilliseconds(unclampedEndTime);
 
-  return std::make_pair(startTimeStamp, endTimeStamp);
+  return Some(std::make_pair(startTimeStamp, endTimeStamp));
 }
 
 // Try to open the marker file for writing performance markers.
@@ -679,13 +695,12 @@ void Performance::MaybeEmitExternalProfilerMarker(
   }
 
 #if defined(XP_LINUX) || defined(XP_WIN) || defined(XP_MACOSX)
-  ErrorResult rv;
-  auto [startTimeStamp, endTimeStamp] =
-      GetTimeStampsForMarker(aStartMark, aEndMark, aOptions, rv);
-
-  if (NS_WARN_IF(rv.Failed())) {
+  Maybe<std::pair<TimeStamp, TimeStamp>> tsPair =
+      GetTimeStampsForMarker(aStartMark, aEndMark, aOptions);
+  if (tsPair.isNothing()) {
     return;
   }
+  auto [startTimeStamp, endTimeStamp] = tsPair.value();
 #endif
 
 #ifdef XP_LINUX
@@ -713,6 +728,44 @@ void Performance::MaybeEmitExternalProfilerMarker(
   fflush(sMarkerFile);
 }
 
+void MOZ_ALWAYS_INLINE Performance::MaybeAddProfileMarker(
+    const nsAString& aName,
+    const Maybe<const PerformanceMeasureOptions&>& options,
+    const Maybe<const nsAString&>& aStartMark,
+    const Optional<nsAString>& aEndMark) {
+  if (profiler_thread_is_being_profiled_for_markers()) {
+    AddProfileMarker(aName, options, aStartMark, aEndMark);
+  }
+}
+
+void MOZ_NEVER_INLINE Performance::AddProfileMarker(
+    const nsAString& aName,
+    const Maybe<const PerformanceMeasureOptions&>& options,
+    const Maybe<const nsAString&>& aStartMark,
+    const Optional<nsAString>& aEndMark) {
+  Maybe<std::pair<TimeStamp, TimeStamp>> tsPair =
+      GetTimeStampsForMarker(aStartMark, aEndMark, options);
+  if (tsPair.isNothing()) {
+    return;
+  }
+  auto [startTimeStamp, endTimeStamp] = tsPair.value();
+
+  Maybe<nsString> endMark;
+  if (aEndMark.WasPassed()) {
+    endMark.emplace(aEndMark.Value());
+  }
+
+  Maybe<uint64_t> innerWindowId;
+  if (nsGlobalWindowInner* owner = GetOwnerWindow()) {
+    innerWindowId = Some(owner->WindowID());
+  }
+  profiler_add_marker("UserTiming", geckoprofiler::category::DOM,
+                      {MarkerTiming::Interval(startTimeStamp, endTimeStamp),
+                       MarkerInnerWindowId(innerWindowId)},
+                      UserTimingMarker{}, aName, /* aIsMeasure */ true,
+                      aStartMark, endMark);
+}
+
 already_AddRefed<PerformanceMeasure> Performance::Measure(
     JSContext* aCx, const nsAString& aName,
     const StringOrPerformanceMeasureOptions& aStartOrMeasureOptions,
@@ -810,25 +863,7 @@ already_AddRefed<PerformanceMeasure> Performance::Measure(
 
   MaybeEmitExternalProfilerMarker(aName, options, startMark, aEndMark);
 
-  if (profiler_thread_is_being_profiled_for_markers()) {
-    auto [startTimeStamp, endTimeStamp] =
-        GetTimeStampsForMarker(startMark, aEndMark, options, aRv);
-
-    Maybe<nsString> endMark;
-    if (aEndMark.WasPassed()) {
-      endMark.emplace(aEndMark.Value());
-    }
-
-    Maybe<uint64_t> innerWindowId;
-    if (nsGlobalWindowInner* owner = GetOwnerWindow()) {
-      innerWindowId = Some(owner->WindowID());
-    }
-    profiler_add_marker("UserTiming", geckoprofiler::category::DOM,
-                        {MarkerTiming::Interval(startTimeStamp, endTimeStamp),
-                         MarkerInnerWindowId(innerWindowId)},
-                        UserTimingMarker{}, aName, /* aIsMeasure */ true,
-                        startMark, endMark);
-  }
+  MaybeAddProfileMarker(aName, options, startMark, aEndMark);
 
   return performanceMeasure.forget();
 }
@@ -1096,7 +1131,7 @@ void Performance::QueueEntry(PerformanceEntry* aEntry) {
   if (!mObservers.IsEmpty()) {
     const auto [begin, end] = mObservers.NonObservingRange();
     std::copy_if(begin, end, MakeBackInserter(interestedObservers),
-                 [aEntry](PerformanceObserver* observer) {
+                 [&](PerformanceObserver* observer) {
                    return observer->ObservesTypeOfEntry(aEntry);
                  });
   }
