@@ -6,70 +6,72 @@
 
 #include "RuntimeService.h"
 
-#include "nsContentSecurityUtils.h"
-#include "nsIContentSecurityPolicy.h"
-#include "mozilla/dom/Document.h"
-#include "nsIObserverService.h"
-#include "nsIScriptContext.h"
-#include "nsIStreamTransportService.h"
-#include "nsISupportsPriority.h"
-#include "nsITimer.h"
-#include "nsIURI.h"
-#include "nsIXULRuntime.h"
-#include "nsPIDOMWindow.h"
-
 #include <algorithm>
-#include "mozilla/ipc/BackgroundChild.h"
+
 #include "GeckoProfiler.h"
+#include "XPCSelfHostedShmem.h"
 #include "js/ColumnNumber.h"  // JS::ColumnNumberOneOrigin
-#include "js/experimental/CTypes.h"  // JS::CTypesActivityType, JS::SetCTypesActivityCallback
-#include "jsfriendapi.h"
-#include "js/friend/ErrorMessages.h"  // js::GetErrorMessage, JSMSG_*
 #include "js/ContextOptions.h"
 #include "js/GCVector.h"
 #include "js/Initialization.h"
 #include "js/LocaleSensitive.h"
 #include "js/Value.h"
 #include "js/WasmFeatures.h"
+#include "js/experimental/CTypes.h"  // JS::CTypesActivityType, JS::SetCTypesActivityCallback
+#include "js/friend/ErrorMessages.h"  // js::GetErrorMessage, JSMSG_*
+#include "jsfriendapi.h"
 #include "mozilla/ArrayUtils.h"
 #include "mozilla/Atomics.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/CycleCollectedJSContext.h"
 #include "mozilla/CycleCollectedJSRuntime.h"
-#include "mozilla/glean/DomWorkersMetrics.h"
-#include "mozilla/glean/DomServiceworkersMetrics.h"
+#include "mozilla/DebugOnly.h"
+#include "mozilla/FlowMarkers.h"
+#include "mozilla/Monitor.h"
+#include "mozilla/Preferences.h"
+#include "mozilla/ScopeExit.h"
 #include "mozilla/TimeStamp.h"
 #include "mozilla/dom/AtomList.h"
 #include "mozilla/dom/BindingUtils.h"
+#include "mozilla/dom/Document.h"
 #include "mozilla/dom/ErrorEventBinding.h"
 #include "mozilla/dom/EventTargetBinding.h"
 #include "mozilla/dom/FetchUtil.h"
+#include "mozilla/dom/IndexedDatabaseManager.h"
 #include "mozilla/dom/MessageChannel.h"
 #include "mozilla/dom/MessageEventBinding.h"
+#include "mozilla/dom/Navigator.h"
 #include "mozilla/dom/PerformanceService.h"
 #include "mozilla/dom/RemoteWorkerChild.h"
-#include "mozilla/dom/WorkerBinding.h"
 #include "mozilla/dom/ScriptSettings.h"
 #include "mozilla/dom/ShadowRealmGlobalScope.h"
 #include "mozilla/dom/TrustedTypeUtils.h"
-#include "mozilla/dom/IndexedDatabaseManager.h"
-#include "mozilla/DebugOnly.h"
-#include "mozilla/Preferences.h"
-#include "mozilla/ScopeExit.h"
-#include "mozilla/dom/Navigator.h"
-#include "mozilla/Monitor.h"
+#include "mozilla/dom/WorkerBinding.h"
+#include "mozilla/extensions/WebExtensionPolicy.h"
+#include "mozilla/glean/DomServiceworkersMetrics.h"
+#include "mozilla/glean/DomWorkersMetrics.h"
+#include "mozilla/ipc/BackgroundChild.h"
+#include "nsContentSecurityUtils.h"
 #include "nsContentUtils.h"
 #include "nsCycleCollector.h"
 #include "nsDOMJSUtils.h"
+#include "nsIContentSecurityPolicy.h"
+#include "nsIObserverService.h"
+#include "nsIScriptContext.h"
+#include "nsIStreamTransportService.h"
 #include "nsISupportsImpl.h"
+#include "nsISupportsPriority.h"
+#include "nsITimer.h"
+#include "nsIURI.h"
+#include "nsIXULRuntime.h"
 #include "nsLayoutStatics.h"
 #include "nsNetUtil.h"
+#include "nsPIDOMWindow.h"
 #include "nsServiceManagerUtils.h"
 #include "nsThreadUtils.h"
 #include "nsXPCOM.h"
 #include "nsXPCOMPrivate.h"
 #include "xpcpublic.h"
-#include "XPCSelfHostedShmem.h"
 
 #if defined(XP_MACOSX)
 #  include "nsMacUtilsImpl.h"
@@ -374,6 +376,7 @@ void LoadJSGCMemoryOptions(const char* aPrefName, void* /* aClosure */) {
 #ifdef NIGHTLY_BUILD
       PREF("gc_experimental_semispace_nursery", JSGC_SEMISPACE_NURSERY_ENABLED),
 #endif
+      PREF("nursery_max_time_goal_ms", JSGC_NURSERY_MAX_TIME_GOAL_MS),
       // Note: Workers do not currently trigger eager minor GC, but if that is
       // desired the following parameters should be added:
       // javascript.options.mem.nursery_eager_collection_threshold_kb
@@ -453,6 +456,7 @@ void LoadJSGCMemoryOptions(const char* aPrefName, void* /* aClosure */) {
       case JSGC_HEAP_GROWTH_FACTOR:
       case JSGC_PARALLEL_MARKING_THRESHOLD_MB:
       case JSGC_MAX_MARKING_THREADS:
+      case JSGC_NURSERY_MAX_TIME_GOAL_MS:
         UpdateCommonJSGCMemoryOption(rts, pref->fullName, pref->key);
         break;
       default:
@@ -511,8 +515,9 @@ MOZ_CAN_RUN_SCRIPT_FOR_DEFINITION bool ContentSecurityPolicyAllows(
   WorkerPrivate* worker = GetWorkerPrivateFromContext(aCx);
   worker->AssertIsOnWorkerThread();
 
-  bool evalOK;
-  bool reportViolation;
+  // Allow eval by default without a CSP.
+  bool evalOK = true;
+  bool reportViolation = false;
   uint16_t violationType;
   nsAutoJSString scriptSample;
   if (aKind == JS::RuntimeCode::JS) {
@@ -539,12 +544,27 @@ MOZ_CAN_RUN_SCRIPT_FOR_DEFINITION bool ContentSecurityPolicyAllows(
       return true;
     }
 
-    evalOK = worker->IsEvalAllowed();
-    reportViolation = worker->GetReportEvalCSPViolations();
+    if (WorkerCSPContext* ctx = worker->GetCSPContext()) {
+      evalOK = ctx->IsEvalAllowed(reportViolation);
+    }
     violationType = nsIContentSecurityPolicy::VIOLATION_TYPE_EVAL;
   } else {
-    evalOK = worker->IsWasmEvalAllowed();
-    reportViolation = worker->GetReportWasmEvalCSPViolations();
+    if (WorkerCSPContext* ctx = worker->GetCSPContext()) {
+      evalOK = ctx->IsWasmEvalAllowed(reportViolation);
+    }
+
+    // As for nsScriptSecurityManager::ContentSecurityPolicyPermitsJSAction,
+    // for MV2 extensions we have to allow wasm by default and report violations
+    // for historical reasons.
+    // TODO bug 1770909: remove this exception.
+    auto* principal = BasePrincipal::Cast(worker->GetPrincipal());
+    RefPtr<extensions::WebExtensionPolicyCore> policy =
+        principal ? principal->AddonPolicyCore() : nullptr;
+    if (!evalOK && policy && policy->ManifestVersion() == 2) {
+      evalOK = true;
+      reportViolation = true;
+    }
+
     violationType = nsIContentSecurityPolicy::VIOLATION_TYPE_WASM_EVAL;
   }
 
@@ -604,7 +624,7 @@ void CTypesActivityCallback(JSContext* aCx, JS::CTypesActivityType aType) {
 // being called, DispatchToEventLoopCallback failure is expected to happen
 // during shutdown.
 class JSDispatchableRunnable final : public WorkerThreadRunnable {
-  JS::Dispatchable* mDispatchable;
+  js::UniquePtr<JS::Dispatchable> mDispatchable;
 
   ~JSDispatchableRunnable() { MOZ_ASSERT(!mDispatchable); }
 
@@ -615,17 +635,19 @@ class JSDispatchableRunnable final : public WorkerThreadRunnable {
 
   void PostDispatch(WorkerPrivate* aWorkerPrivate,
                     bool aDispatchResult) override {
-    // For the benefit of the destructor assert.
     if (!aDispatchResult) {
-      mDispatchable = nullptr;
+      // It is possible (for example in WASM failed compilation) that a
+      // worker will not run, and in this case we need to
+      // release the task as a failed task for deletion by the JS runtime.
+      JS::Dispatchable::ReleaseFailedTask(std::move(mDispatchable));
     }
   }
 
  public:
   JSDispatchableRunnable(WorkerPrivate* aWorkerPrivate,
-                         JS::Dispatchable* aDispatchable)
+                         js::UniquePtr<JS::Dispatchable>&& aDispatchable)
       : WorkerThreadRunnable("JSDispatchableRunnable"),
-        mDispatchable(aDispatchable) {
+        mDispatchable(std::move(aDispatchable)) {
     MOZ_ASSERT(mDispatchable);
   }
 
@@ -636,9 +658,11 @@ class JSDispatchableRunnable final : public WorkerThreadRunnable {
     AutoJSAPI jsapi;
     jsapi.Init();
 
-    mDispatchable->run(aWorkerPrivate->GetJSContext(),
-                       JS::Dispatchable::NotShuttingDown);
-    mDispatchable = nullptr;  // mDispatchable may delete itself
+    JS::Dispatchable::Run(aWorkerPrivate->GetJSContext(),
+                          std::move(mDispatchable),
+                          JS::Dispatchable::NotShuttingDown);
+    // mDispatchable is no longer valid after this point.
+    // The delete has been handled on the JS engine side
 
     return true;
   }
@@ -649,16 +673,23 @@ class JSDispatchableRunnable final : public WorkerThreadRunnable {
     AutoJSAPI jsapi;
     jsapi.Init();
 
-    mDispatchable->run(GetCurrentThreadWorkerPrivate()->GetJSContext(),
-                       JS::Dispatchable::ShuttingDown);
-    mDispatchable = nullptr;  // mDispatchable may delete itself
+    // TODO: Make this make more sense
+    // Why are we calling Run here? Because the way the API was designed
+    // is so that once control is passed to the runnable, then both cancellation
+    // and running are handled through `Run` by either passing NotShuttingDown
+    // or ShuttingDown (for cancellation).
+    JS::Dispatchable::Run(GetCurrentThreadWorkerPrivate()->GetJSContext(),
+                          std::move(mDispatchable),
+                          JS::Dispatchable::ShuttingDown);
+    // mDispatchable is no longer valid after this point.
+    // The delete has been handled on the JS engine side
 
     return NS_OK;
   }
 };
 
-static bool DispatchToEventLoop(void* aClosure,
-                                JS::Dispatchable* aDispatchable) {
+static bool DispatchToEventLoop(
+    void* aClosure, js::UniquePtr<JS::Dispatchable>&& aDispatchable) {
   // This callback may execute either on the worker thread or a random
   // JS-internal helper thread.
 
@@ -669,10 +700,28 @@ static bool DispatchToEventLoop(void* aClosure,
   // Dispatch is expected to fail during shutdown for the reasons outlined in
   // the JSDispatchableRunnable comment above.
   RefPtr<JSDispatchableRunnable> r =
-      new JSDispatchableRunnable(workerPrivate, aDispatchable);
+      new JSDispatchableRunnable(workerPrivate, std::move(aDispatchable));
   return r->Dispatch(workerPrivate);
 }
 
+static bool DelayedDispatchToEventLoop(
+    void* aClosure, js::UniquePtr<JS::Dispatchable>&& aDispatchable,
+    uint32_t delay) {
+  // See comment at JS::InitDispatchsToEventLoop() below for how we know the
+  // WorkerPrivate is alive.
+  WorkerPrivate* workerPrivate = reinterpret_cast<WorkerPrivate*>(aClosure);
+
+  workerPrivate->AssertIsOnWorkerThread();
+
+  JSContext* cx = workerPrivate->GetJSContext();
+  TimeoutHandler* handler =
+      new DelayedJSDispatchableHandler(cx, std::move(aDispatchable));
+  workerPrivate->SetTimeout(cx, handler, delay, /* aIsInterval */ false,
+                            Timeout::Reason::eJSTimeout, IgnoreErrors());
+
+  return true;
+}
+
 static bool ConsumeStream(JSContext* aCx, JS::Handle<JSObject*> aObj,
                           JS::MimeType aMimeType,
                           JS::StreamConsumer* aConsumer) {
@@ -714,8 +763,9 @@ bool InitJSContextForWorker(WorkerPrivate* aWorkerPrivate,
 
   // A WorkerPrivate lives strictly longer than its JSRuntime so we can safely
   // store a raw pointer as the callback's closure argument on the JSRuntime.
-  JS::InitDispatchToEventLoop(aWorkerCx, DispatchToEventLoop,
-                              (void*)aWorkerPrivate);
+  JS::InitDispatchsToEventLoop(aWorkerCx, DispatchToEventLoop,
+                               DelayedDispatchToEventLoop,
+                               (void*)aWorkerPrivate);
 
   JS::InitConsumeStreamCallback(aWorkerCx, ConsumeStream,
                                 FetchUtil::ReportJSStreamError);
@@ -842,7 +892,7 @@ class WorkerJSRuntime final : public mozilla::CycleCollectedJSRuntime {
     }
   }
 
-  void TraceNativeBlackRoots(JSTracer* aTracer) override {
+  void TraceAdditionalNativeBlackRoots(JSTracer* aTracer) override {
     if (!mWorkerPrivate || !mWorkerPrivate->MayContinueRunning()) {
       return;
     }
@@ -957,6 +1007,12 @@ class WorkerJSContext final : public mozilla::CycleCollectedJSContext {
     }
 
     JS::JobQueueMayNotBeEmpty(cx);
+    if (!runnable->isInList()) {
+      // A recycled object may be in the list already.
+      mMicrotasksToTrace.insertBack(runnable);
+    }
+    PROFILER_MARKER_FLOW_ONLY("WorkerJSContext::DispatchToMicroTask", OTHER, {},
+                              FlowMarker, Flow::FromPointer(runnable.get()));
     microTaskQueue->push_back(std::move(runnable));
   }
 
@@ -1195,6 +1251,8 @@ bool RuntimeService::RegisterWorker(WorkerPrivate& aWorkerPrivate) {
     }
   }
 
+  aWorkerPrivate.SetIsQueued(queued);
+
   // From here on out we must call UnregisterWorker if something fails!
   if (parent) {
     if (!parent->AddChildWorker(aWorkerPrivate)) {
@@ -1362,6 +1420,7 @@ bool RuntimeService::ScheduleWorker(WorkerPrivate& aWorkerPrivate) {
 
   aWorkerPrivate.SetThread(thread.unsafeGetRawPtr());
   JSContext* cx = CycleCollectedJSContext::Get()->Context();
+
   nsCOMPtr<nsIRunnable> runnable = new WorkerThreadPrimaryRunnable(
       &aWorkerPrivate, thread.clonePtr(), JS_GetParentRuntime(cx));
   if (NS_FAILED(
@@ -1370,6 +1429,12 @@ bool RuntimeService::ScheduleWorker(WorkerPrivate& aWorkerPrivate) {
     return false;
   }
 
+  // The worker was queued when creating, so enable remote debugger now.
+  if (aWorkerPrivate.IsQueued()) {
+    aWorkerPrivate.SetIsQueued(false);
+    aWorkerPrivate.EnableRemoteDebugger();
+  }
+
   return true;
 }
 
@@ -1918,13 +1983,18 @@ void RuntimeService::MemoryPressureAllWorkers() {
   BroadcastAllWorkers([](auto& worker) { worker.MemoryPressure(); });
 }
 
-uint32_t RuntimeService::ClampedHardwareConcurrency(
-    bool aShouldResistFingerprinting) const {
-  // The Firefox Hardware Report says 70% of Firefox users have exactly 2 cores.
+uint32_t RuntimeService::ClampedHardwareConcurrency(bool aRFPHardcoded,
+                                                    bool aRFPTiered) const {
+  // The Firefox Hardware Report says 34% of Firefox users have exactly 4 cores.
   // When the resistFingerprinting pref is set, we want to blend into the crowd
-  // so spoof navigator.hardwareConcurrency = 2 to reduce user uniqueness.
-  if (MOZ_UNLIKELY(aShouldResistFingerprinting)) {
-    return 2;
+  // so spoof navigator.hardwareConcurrency = 4 to reduce user uniqueness. On
+  // OSX, the majority of Macs have 8 cores.
+  if (MOZ_UNLIKELY(aRFPHardcoded)) {
+#ifdef XP_MACOSX
+    return 8;
+#else
+    return 4;
+#endif
   }
 
   // This needs to be atomic, because multiple workers, and even mainthread,
@@ -1952,6 +2022,13 @@ uint32_t RuntimeService::ClampedHardwareConcurrency(
                                                            numberOfProcessors);
   }
 
+  if (MOZ_UNLIKELY(aRFPTiered)) {
+    if (unclampedHardwareConcurrency >= 8) {
+      return 8;
+    }
+    return 4;
+  }
+
   return std::min(uint32_t(unclampedHardwareConcurrency),
                   StaticPrefs::dom_maxHardwareConcurrency());
 }
@@ -2082,6 +2159,26 @@ void RuntimeService::DumpRunningWorkers() {
   }
 }
 
+void RuntimeService::UpdateWorkersPlaybackState(
+    const nsPIDOMWindowInner& aWindow, bool aIsPlayingAudio) {
+  AssertIsOnMainThread();
+
+  for (WorkerPrivate* const worker : GetWorkersForWindow(aWindow)) {
+    MOZ_ASSERT(!worker->IsSharedWorker());
+    worker->SetIsPlayingAudio(aIsPlayingAudio);
+  }
+}
+
+void RuntimeService::UpdateWorkersPeerConnections(
+    const nsPIDOMWindowInner& aWindow, bool aHasPeerConnections) {
+  AssertIsOnMainThread();
+
+  for (WorkerPrivate* const worker : GetWorkersForWindow(aWindow)) {
+    MOZ_ASSERT(!worker->IsSharedWorker());
+    worker->SetActivePeerConnections(aHasPeerConnections);
+  }
+}
+
 bool LogViolationDetailsRunnable::MainThreadRun() {
   AssertIsOnMainThread();
   MOZ_ASSERT(mWorkerRef);
@@ -2162,6 +2259,11 @@ WorkerThreadPrimaryRunnable::Run() {
       }
 
       failureCleanup.release();
+
+      // Binding the RemoteWorkerDebugger child endpoint after initailzation
+      // successfully.
+      // mWorkerPrivate->BindRemoteWorkerDebuggerChild();
+
       runLoopRan = true;
 
       {
@@ -2441,5 +2543,23 @@ JSObject* GetCurrentThreadWorkerDebuggerGlobal() {
   return scope->GetGlobalJSObject();
 }
 
+void UpdateWorkersPlaybackState(const nsPIDOMWindowInner& aWindow,
+                                bool aIsPlayingAudio) {
+  AssertIsOnMainThread();
+  RuntimeService* runtime = RuntimeService::GetService();
+  if (runtime) {
+    runtime->UpdateWorkersPlaybackState(aWindow, aIsPlayingAudio);
+  }
+}
+
+void UpdateWorkersPeerConnections(const nsPIDOMWindowInner& aWindow,
+                                  bool aHasPeerConnections) {
+  AssertIsOnMainThread();
+  RuntimeService* runtime = RuntimeService::GetService();
+  if (runtime) {
+    runtime->UpdateWorkersPeerConnections(aWindow, aHasPeerConnections);
+  }
+}
+
 }  // namespace dom
 }  // namespace mozilla
