@@ -15,16 +15,16 @@
 #include "mozilla/Maybe.h"
 #include "mozilla/NotNull.h"
 #include "mozilla/RefPtr.h"
-#include "mozilla/UniquePtr.h"
 #include "mozilla/TimeStamp.h"
+#include "mozilla/UniquePtr.h"
 #include "mozilla/dom/AnimationFrameProvider.h"
 #include "mozilla/dom/ImageBitmapBinding.h"
 #include "mozilla/dom/ImageBitmapSource.h"
 #include "mozilla/dom/PerformanceWorker.h"
 #include "mozilla/dom/SafeRefPtr.h"
+#include "mozilla/dom/TimeoutManager.h"
 #include "mozilla/dom/TrustedTypePolicyFactory.h"
 #include "mozilla/dom/WorkerPrivate.h"
-#include "mozilla/dom/TimeoutManager.h"
 #include "nsCOMPtr.h"
 #include "nsCycleCollectionParticipant.h"
 #include "nsIGlobalObject.h"
@@ -86,6 +86,7 @@ class WorkerPrivate;
 class VsyncWorkerChild;
 class WebTaskScheduler;
 class WebTaskSchedulerWorker;
+class WebTaskSchedulingState;
 struct RequestInit;
 
 namespace cache {
@@ -185,10 +186,6 @@ class WorkerGlobalScopeBase : public DOMEventTargetHelper,
 
   ClientSource& MutableClientSourceRef() const { return *mClientSource; }
 
-  // WorkerPrivate wants to be able to forbid script when its state machine
-  // demands it.
-  void WorkerPrivateSaysForbidScript() { StartForbiddingScript(); }
-  void WorkerPrivateSaysAllowScript() { StopForbiddingScript(); }
   bool IsBackgroundInternal() const override {
     MOZ_ASSERT(mWorkerPrivate);
     return mWorkerPrivate->IsRunningInBackground();
@@ -206,6 +203,41 @@ class WorkerGlobalScopeBase : public DOMEventTargetHelper,
     return mWorkerPrivate->IsFrozenForWorkerThread();
   }
 
+  void UpdateWebSocketCount(int32_t aDelta) override {
+    if (aDelta == 0) {
+      return;
+    }
+
+    MOZ_DIAGNOSTIC_ASSERT(
+        aDelta > 0 || ((aDelta + mNumOfOpenWebSockets) < mNumOfOpenWebSockets));
+
+    mNumOfOpenWebSockets += aDelta;
+  }
+
+  // Increase/Decrease the number of active IndexedDB databases for the
+  // decision making of timeout-throttling.
+  void UpdateActiveIndexedDBDatabaseCount(int32_t aDelta) override {
+    AssertIsOnWorkerThread();
+    mNumOfIndexedDBDatabases += aDelta;
+  }
+
+  bool HasOpenWebSockets() const override { return mNumOfOpenWebSockets; }
+
+  bool HasActiveIndexedDBDatabases() const override {
+    AssertIsOnWorkerThread();
+    return mNumOfIndexedDBDatabases;
+  }
+
+  bool IsPlayingAudio() override {
+    AssertIsOnWorkerThread();
+    return mWorkerPrivate && mWorkerPrivate->IsPlayingAudio();
+  }
+
+  bool HasActivePeerConnections() override {
+    AssertIsOnWorkerThread();
+    return mWorkerPrivate && mWorkerPrivate->HasActivePeerConnections();
+  }
+
   void TriggerUpdateCCFlag() override {
     mWorkerPrivate->UpdateCCFlag(WorkerPrivate::CCFlag::EligibleForTimeout);
   }
@@ -232,6 +264,8 @@ class WorkerGlobalScopeBase : public DOMEventTargetHelper,
   PRThread* mWorkerThreadUsedOnlyForAssert;
 #endif
   mozilla::UniquePtr<mozilla::dom::TimeoutManager> mTimeoutManager;
+  uint32_t mNumOfOpenWebSockets{};
+  uint32_t mNumOfIndexedDBDatabases{};
 };
 
 namespace workerinternals {
@@ -345,7 +379,7 @@ class WorkerGlobalScope : public WorkerGlobalScopeBase {
   int32_t SetTimeout(JSContext* aCx,
                      const FunctionOrTrustedScriptOrString& aHandler,
                      int32_t aTimeout, const Sequence<JS::Value>& aArguments,
-                     ErrorResult& aRv);
+                     nsIPrincipal* aSubjectPrincipal, ErrorResult& aRv);
 
   MOZ_CAN_RUN_SCRIPT
   void ClearTimeout(int32_t aHandle);
@@ -354,7 +388,7 @@ class WorkerGlobalScope : public WorkerGlobalScopeBase {
   int32_t SetInterval(JSContext* aCx,
                       const FunctionOrTrustedScriptOrString& aHandler,
                       int32_t aTimeout, const Sequence<JS::Value>& aArguments,
-                      ErrorResult& aRv);
+                      nsIPrincipal* aSubjectPrincipal, ErrorResult& aRv);
 
   MOZ_CAN_RUN_SCRIPT
   void ClearInterval(int32_t aHandle);
@@ -385,6 +419,12 @@ class WorkerGlobalScope : public WorkerGlobalScopeBase {
 
   WebTaskScheduler* Scheduler();
   WebTaskScheduler* GetExistingScheduler() const;
+  void SetWebTaskSchedulingState(WebTaskSchedulingState* aState) override;
+  bool HasScheduledNormalOrHighPriorityWebTasks() const override;
+
+  WebTaskSchedulingState* GetWebTaskSchedulingState() const override {
+    return mWebTaskSchedulingState;
+  }
 
   bool WindowInteractionAllowed() const;
 
@@ -406,11 +446,10 @@ class WorkerGlobalScope : public WorkerGlobalScopeBase {
 
  private:
   MOZ_CAN_RUN_SCRIPT
-  int32_t SetTimeoutOrInterval(JSContext* aCx,
-                               const FunctionOrTrustedScriptOrString& aHandler,
-                               int32_t aTimeout,
-                               const Sequence<JS::Value>& aArguments,
-                               bool aIsInterval, ErrorResult& aRv);
+  int32_t SetTimeoutOrInterval(
+      JSContext* aCx, const FunctionOrTrustedScriptOrString& aHandler,
+      int32_t aTimeout, const Sequence<JS::Value>& aArguments, bool aIsInterval,
+      nsIPrincipal* aSubjectPrincipal, ErrorResult& aRv);
 
   RefPtr<Crypto> mCrypto;
   RefPtr<WorkerLocation> mLocation;
@@ -421,6 +460,7 @@ class WorkerGlobalScope : public WorkerGlobalScopeBase {
   RefPtr<cache::CacheStorage> mCacheStorage;
   RefPtr<DebuggerNotificationManager> mDebuggerNotificationManager;
   RefPtr<WebTaskSchedulerWorker> mWebTaskScheduler;
+  RefPtr<WebTaskSchedulingState> mWebTaskSchedulingState;
   RefPtr<TrustedTypePolicyFactory> mTrustedTypePolicyFactory;
   uint32_t mWindowInteractionsAllowed = 0;
   bool mIsEligibleForMessaging{true};
