@@ -22,6 +22,14 @@
 #include "malloc_decls.h"
 #include "GLContext.h"
 
+#include "source-repo.h"
+
+#ifdef MOZ_SOURCE_STAMP
+#  define MOZ_SOURCE_STAMP_VALUE MOZ_STRINGIFY(MOZ_SOURCE_STAMP)
+#else
+#  define MOZ_SOURCE_STAMP_VALUE nullptr
+#endif
+
 static mozilla::LazyLogModule sWrDLLog("wr.dl");
 #define WRDL_LOG(...) \
   MOZ_LOG(sWrDLLog, LogLevel::Debug, ("WRDL(%p): " __VA_ARGS__))
@@ -40,7 +48,8 @@ class NewRenderer : public RendererEvent {
               layers::CompositorBridgeParent* aBridge,
               WebRenderBackend* aBackend, WebRenderCompositor* aCompositor,
               int32_t* aMaxTextureSize, bool* aUseANGLE, bool* aUseDComp,
-              bool* aUseTripleBuffering, bool* aSupportsExternalBufferTextures,
+              bool* aUseLayerCompositor, bool* aUseTripleBuffering,
+              bool* aSupportsExternalBufferTextures,
               RefPtr<widget::CompositorWidget>&& aWidget,
               layers::SynchronousTask* aTask, LayoutDeviceIntSize aSize,
               layers::WindowKind aWindowKind, layers::SyncHandle* aHandle,
@@ -51,6 +60,7 @@ class NewRenderer : public RendererEvent {
         mMaxTextureSize(aMaxTextureSize),
         mUseANGLE(aUseANGLE),
         mUseDComp(aUseDComp),
+        mUseLayerCompositor(aUseLayerCompositor),
         mUseTripleBuffering(aUseTripleBuffering),
         mSupportsExternalBufferTextures(aSupportsExternalBufferTextures),
         mBridge(aBridge),
@@ -83,6 +93,7 @@ class NewRenderer : public RendererEvent {
     *mCompositor = compositor->CompositorType();
     *mUseANGLE = compositor->UseANGLE();
     *mUseDComp = compositor->UseDComp();
+    *mUseLayerCompositor = compositor->ShouldUseLayerCompositor();
     *mUseTripleBuffering = compositor->UseTripleBuffering();
     *mSupportsExternalBufferTextures =
         compositor->SupportsExternalBufferTextures();
@@ -131,13 +142,6 @@ class NewRenderer : public RendererEvent {
       }
     }
 
-    // Only allow the layer compositor in nightly builds, for now.
-    bool use_layer_compositor = false;
-#ifdef NIGHTLY_BUILD
-    use_layer_compositor =
-        StaticPrefs::gfx_webrender_layer_compositor_AtStartup();
-#endif
-
     if (!wr_window_new(
             aWindowId, mSize.width, mSize.height,
             mWindowKind == WindowKind::MAIN, supportLowPriorityTransactions,
@@ -159,7 +163,7 @@ class NewRenderer : public RendererEvent {
             StaticPrefs::gfx_webrender_low_quality_pinch_zoom_AtStartup(),
             StaticPrefs::gfx_webrender_max_shared_surface_size_AtStartup(),
             StaticPrefs::gfx_webrender_enable_subpixel_aa_AtStartup(),
-            use_layer_compositor)) {
+            compositor->ShouldUseLayerCompositor())) {
       // wr_window_new puts a message into gfxCriticalNote if it returns false
       MOZ_ASSERT(errorMessage);
       mError->AssignASCII(errorMessage);
@@ -196,6 +200,7 @@ class NewRenderer : public RendererEvent {
   int32_t* mMaxTextureSize;
   bool* mUseANGLE;
   bool* mUseDComp;
+  bool* mUseLayerCompositor;
   bool* mUseTripleBuffering;
   bool* mSupportsExternalBufferTextures;
   layers::CompositorBridgeParent* mBridge;
@@ -288,8 +293,10 @@ void TransactionBuilder::ClearDisplayList(Epoch aEpoch,
 }
 
 void TransactionBuilder::GenerateFrame(const VsyncId& aVsyncId, bool aPresent,
+                                       bool aTracked,
                                        wr::RenderReasons aReasons) {
-  wr_transaction_generate_frame(mTxn, aVsyncId.mId, aPresent, aReasons);
+  wr_transaction_generate_frame(mTxn, aVsyncId.mId, aPresent, aTracked,
+                                aReasons);
 }
 
 void TransactionBuilder::InvalidateRenderedFrame(wr::RenderReasons aReasons) {
@@ -318,6 +325,10 @@ void TransactionBuilder::SetDocumentView(
   wr_transaction_set_document_view(mTxn, &wrDocRect);
 }
 
+void TransactionBuilder::RenderOffscreen(wr::WrPipelineId aPipelineId) {
+  wr_transaction_render_offscreen(mTxn, aPipelineId);
+}
+
 TransactionWrapper::TransactionWrapper(Transaction* aTxn) : mTxn(aTxn) {}
 
 void TransactionWrapper::AppendDynamicProperties(
@@ -374,6 +385,7 @@ already_AddRefed<WebRenderAPI> WebRenderAPI::Create(
   int32_t maxTextureSize = 0;
   bool useANGLE = false;
   bool useDComp = false;
+  bool useLayerCompositor = false;
   bool useTripleBuffering = false;
   bool supportsExternalBufferTextures = false;
   layers::SyncHandle syncHandle = {};
@@ -384,8 +396,9 @@ already_AddRefed<WebRenderAPI> WebRenderAPI::Create(
   layers::SynchronousTask task("Create Renderer");
   auto event = MakeUnique<NewRenderer>(
       &docHandle, aBridge, &backend, &compositor, &maxTextureSize, &useANGLE,
-      &useDComp, &useTripleBuffering, &supportsExternalBufferTextures,
-      std::move(aWidget), &task, aSize, aWindowKind, &syncHandle, &aError);
+      &useDComp, &useLayerCompositor, &useTripleBuffering,
+      &supportsExternalBufferTextures, std::move(aWidget), &task, aSize,
+      aWindowKind, &syncHandle, &aError);
   RenderThread::Get()->PostEvent(aWindowId, std::move(event));
 
   task.Wait();
@@ -397,7 +410,7 @@ already_AddRefed<WebRenderAPI> WebRenderAPI::Create(
   return RefPtr<WebRenderAPI>(
              new WebRenderAPI(docHandle, aWindowId, backend, compositor,
                               maxTextureSize, useANGLE, useDComp,
-                              useTripleBuffering,
+                              useLayerCompositor, useTripleBuffering,
                               supportsExternalBufferTextures, syncHandle))
       .forget();
 }
@@ -408,8 +421,8 @@ already_AddRefed<WebRenderAPI> WebRenderAPI::Clone() {
 
   RefPtr<WebRenderAPI> renderApi = new WebRenderAPI(
       docHandle, mId, mBackend, mCompositor, mMaxTextureSize, mUseANGLE,
-      mUseDComp, mUseTripleBuffering, mSupportsExternalBufferTextures,
-      mSyncHandle, this, this);
+      mUseDComp, mUseLayerCompositor, mUseTripleBuffering,
+      mSupportsExternalBufferTextures, mSyncHandle, this, this);
 
   return renderApi.forget();
 }
@@ -421,7 +434,7 @@ wr::WrIdNamespace WebRenderAPI::GetNamespace() {
 WebRenderAPI::WebRenderAPI(
     wr::DocumentHandle* aHandle, wr::WindowId aId, WebRenderBackend aBackend,
     WebRenderCompositor aCompositor, uint32_t aMaxTextureSize, bool aUseANGLE,
-    bool aUseDComp, bool aUseTripleBuffering,
+    bool aUseDComp, bool aUseLayerCompositor, bool aUseTripleBuffering,
     bool aSupportsExternalBufferTextures, layers::SyncHandle aSyncHandle,
     wr::WebRenderAPI* aRootApi, wr::WebRenderAPI* aRootDocumentApi)
     : mDocHandle(aHandle),
@@ -431,6 +444,7 @@ WebRenderAPI::WebRenderAPI(
       mMaxTextureSize(aMaxTextureSize),
       mUseANGLE(aUseANGLE),
       mUseDComp(aUseDComp),
+      mUseLayerCompositor(aUseLayerCompositor),
       mUseTripleBuffering(aUseTripleBuffering),
       mSupportsExternalBufferTextures(aSupportsExternalBufferTextures),
       mCaptureSequence(false),
@@ -708,8 +722,14 @@ void WebRenderAPI::Readback(const TimeStamp& aStartTime, gfx::IntSize size,
 
     void Run(RenderThread& aRenderThread, WindowId aWindowId) override {
       RendererStats stats = {0};
-      aRenderThread.UpdateAndRender(aWindowId, VsyncId(), mStartTime,
-                                    /* aRender */ true, Some(mSize),
+      wr::FrameReadyParams params = {
+          .present = true,
+          .render = true,
+          .scrolled = false,
+          .tracked = false,
+      };
+      aRenderThread.UpdateAndRender(aWindowId, VsyncId(), mStartTime, params,
+                                    Some(mSize),
                                     wr::SurfaceFormatToImageFormat(mFormat),
                                     Some(mBuffer), &stats, mNeedsYFlip);
       layers::AutoCompleteTask complete(mTask);
@@ -880,7 +900,8 @@ void WebRenderAPI::Capture() {
   // SCENE | FRAME | TILE_CACHE
   uint8_t bits = 15;                // TODO: get from JavaScript
   const char* path = "wr-capture";  // TODO: get from JavaScript
-  wr_api_capture(mDocHandle, path, bits);
+  const char* revision = MOZ_SOURCE_STAMP_VALUE;
+  wr_api_capture(mDocHandle, path, revision, bits);
 }
 
 void WebRenderAPI::StartCaptureSequence(const nsACString& aPath,
@@ -890,7 +911,7 @@ void WebRenderAPI::StartCaptureSequence(const nsACString& aPath,
   }
 
   wr_api_start_capture_sequence(mDocHandle, PromiseFlatCString(aPath).get(),
-                                aFlags);
+                                MOZ_SOURCE_STAMP_VALUE, aFlags);
 
   mCaptureSequence = true;
 }
