@@ -35,7 +35,7 @@ use crate::hit_test::{HitTest, HitTester, SharedHitTester};
 use crate::intern::DataStore;
 #[cfg(any(feature = "capture", feature = "replay"))]
 use crate::internal_types::DebugOutput;
-use crate::internal_types::{FastHashMap, FrameId, FrameMemory, FrameStamp, RenderedDocument, ResultMsg};
+use crate::internal_types::{FastHashMap, FrameId, FrameStamp, RenderedDocument, ResultMsg};
 use malloc_size_of::{MallocSizeOf, MallocSizeOfOps};
 use crate::picture::{PictureScratchBuffer, SliceId, TileCacheInstance, TileCacheParams, SurfaceInfo, RasterConfig};
 use crate::picture::PicturePrimitive;
@@ -71,7 +71,6 @@ use std::{mem, u32};
 use std::path::PathBuf;
 #[cfg(feature = "replay")]
 use crate::frame_builder::Frame;
-use time::precise_time_ns;
 use core::time::Duration;
 use crate::util::{Recycler, VecHelper, drain_filter};
 
@@ -517,9 +516,9 @@ impl Document {
         frame_stats: Option<FullFrameStats>,
         present: bool,
         render_reasons: RenderReasons,
-        frame_memory: FrameMemory,
+        chunk_pool: Arc<ChunkPool>,
     ) -> RenderedDocument {
-        let frame_build_start_time = precise_time_ns();
+        let frame_build_start_time = zeitstempel::now();
 
         // Advance to the next frame.
         self.stamp.advance();
@@ -548,7 +547,7 @@ impl Document {
                 // on the next frame, it will add new entries to the minimap
                 // data during sampling.
                 mem::take(&mut self.minimap_data),
-                frame_memory,
+                chunk_pool,
             );
 
             frame
@@ -560,7 +559,7 @@ impl Document {
         self.has_built_scene = false;
 
         let frame_build_time_ms =
-            profiler::ns_to_ms(precise_time_ns() - frame_build_start_time);
+            profiler::ns_to_ms(zeitstempel::now() - frame_build_start_time);
         self.profile.set(profiler::FRAME_BUILDING_TIME, frame_build_time_ms);
         self.profile.start_time(profiler::FRAME_SEND_TIME);
 
@@ -577,6 +576,69 @@ impl Document {
         }
     }
 
+    /// Build a frame without changing the state of the current scene.
+    ///
+    /// This is useful to render arbitrary content into to images in
+    /// the resource cache for later use without affecting what is
+    /// currently being displayed.
+    fn process_offscreen_scene(
+        &mut self,
+        mut txn: OffscreenBuiltScene,
+        resource_cache: &mut ResourceCache,
+        gpu_cache: &mut GpuCache,
+        chunk_pool: Arc<ChunkPool>,
+        debug_flags: DebugFlags,
+    ) -> RenderedDocument {
+        let mut profile = TransactionProfile::new();
+        self.stamp.advance();
+
+        let mut data_stores = DataStores::default();
+        data_stores.apply_updates(txn.interner_updates, &mut profile);
+
+        let mut spatial_tree = SpatialTree::new();
+        spatial_tree.apply_updates(txn.spatial_tree_updates);
+
+        let mut tile_caches = FastHashMap::default();
+        self.update_tile_caches_for_new_scene(
+            mem::take(&mut txn.scene.tile_cache_config.tile_caches),
+            &mut tile_caches,
+            resource_cache,
+        );
+
+        let present = false;
+
+        let frame = self.frame_builder.build(
+            &mut txn.scene,
+            present,
+            resource_cache,
+            gpu_cache,
+            &mut self.rg_builder,
+            self.stamp, // TODO(nical)
+            self.view.scene.device_rect.min,
+            &self.dynamic_properties,
+            &mut data_stores,
+            &mut self.scratch,
+            debug_flags,
+            &mut tile_caches,
+            &mut spatial_tree,
+            self.dirty_rects_are_valid,
+            &mut profile,
+            // Consume the minimap data. If APZ wants a minimap rendered
+            // on the next frame, it will add new entries to the minimap
+            // data during sampling.
+            mem::take(&mut self.minimap_data),
+            chunk_pool,
+        );
+
+        RenderedDocument {
+            frame,
+            profile,
+            render_reasons: RenderReasons::SNAPSHOT,
+            frame_stats: None,
+        }
+    }
+
+
     fn rebuild_hit_tester(&mut self) {
         self.spatial_tree.update_tree(&self.dynamic_properties);
 
@@ -945,6 +1007,46 @@ impl RenderBackend {
                     &mut doc.profile,
                 );
 
+                for offscreen_scene in txn.offscreen_scenes.drain(..) {
+                    self.resource_cache.post_scene_building_update(
+                        txn.resource_updates.take(),
+                        &mut doc.profile,
+                    );
+
+                    let rendered_document = doc.process_offscreen_scene(
+                        offscreen_scene,
+                        &mut self.resource_cache,
+                        &mut self.gpu_cache,
+                        self.chunk_pool.clone(),
+                        self.debug_flags,
+                    );
+
+                    let msg = ResultMsg::UpdateGpuCache(self.gpu_cache.extract_updates());
+                    self.result_tx.send(msg).unwrap();
+
+                    let pending_update = self.resource_cache.pending_updates();
+
+                    let msg = ResultMsg::PublishDocument(
+                        self.frame_publish_id,
+                        txn.document_id,
+                        rendered_document,
+                        pending_update,
+                    );
+                    self.result_tx.send(msg).unwrap();
+
+                    let params = api::FrameReadyParams {
+                        present: false,
+                        render: true,
+                        scrolled: false,
+                        tracked: false,
+                    };
+
+                    self.notifier.new_frame_ready(
+                        txn.document_id,
+                        self.frame_publish_id,
+                        &params
+                    );
+                }
             } else {
                 // The document was removed while we were building it, skip it.
                 // TODO: we might want to just ensure that removed documents are
@@ -962,6 +1064,7 @@ impl RenderBackend {
                 txn.notifications.take(),
                 txn.render_frame,
                 txn.present,
+                txn.tracked,
                 RenderReasons::SCENE,
                 None,
                 txn.invalidate_rendered_frame,
@@ -1313,6 +1416,7 @@ impl RenderBackend {
                 txn.notifications.take(),
                 txn.generate_frame.as_bool(),
                 txn.generate_frame.present(),
+                txn.generate_frame.tracked(),
                 txn.render_reasons,
                 txn.generate_frame.id(),
                 txn.invalidate_rendered_frame,
@@ -1353,6 +1457,7 @@ impl RenderBackend {
                     Vec::default(),
                     false,
                     false,
+                    false,
                     RenderReasons::empty(),
                     None,
                     false,
@@ -1376,6 +1481,7 @@ impl RenderBackend {
         mut notifications: Vec<NotificationRequest>,
         mut render_frame: bool,
         mut present: bool,
+        tracked: bool,
         render_reasons: RenderReasons,
         generated_frame_id: Option<u64>,
         invalidate_rendered_frame: bool,
@@ -1383,7 +1489,7 @@ impl RenderBackend {
         has_built_scene: bool,
         start_time: Option<u64>
     ) -> bool {
-        let update_doc_start = precise_time_ns();
+        let update_doc_start = zeitstempel::now();
 
         let requested_frame = render_frame;
 
@@ -1463,7 +1569,7 @@ impl RenderBackend {
             }
 
             if start_time.is_some() {
-              Telemetry::record_time_to_frame_build(Duration::from_nanos(precise_time_ns() - start_time.unwrap()));
+              Telemetry::record_time_to_frame_build(Duration::from_nanos(zeitstempel::now() - start_time.unwrap()));
             }
             profile_scope!("generate frame");
 
@@ -1475,8 +1581,6 @@ impl RenderBackend {
 
                 let frame_stats = doc.frame_stats.take();
 
-                let frame_memory = FrameMemory::new(self.chunk_pool.clone());
-
                 let rendered_document = doc.build_frame(
                     &mut self.resource_cache,
                     &mut self.gpu_cache,
@@ -1485,7 +1589,7 @@ impl RenderBackend {
                     frame_stats,
                     present,
                     render_reasons,
-                    frame_memory,
+                    self.chunk_pool.clone(),
                 );
 
                 debug!("generated frame for document {:?} with {} passes",
@@ -1546,7 +1650,7 @@ impl RenderBackend {
                 None => {},
             }
 
-            let update_doc_time = profiler::ns_to_ms(precise_time_ns() - update_doc_start);
+            let update_doc_time = profiler::ns_to_ms(zeitstempel::now() - update_doc_start);
             rendered_document.profile.set(profiler::UPDATE_DOCUMENT_TIME, update_doc_time);
 
             let msg = ResultMsg::PublishPipelineInfo(doc.updated_pipeline_info());
@@ -1590,7 +1694,13 @@ impl RenderBackend {
             } else if render_frame {
                 doc.rendered_frame_is_valid = true;
             }
-            self.notifier.new_frame_ready(document_id, scroll, render_frame, self.frame_publish_id);
+            let params = api::FrameReadyParams {
+                present,
+                render: render_frame,
+                scrolled: scroll,
+                tracked,
+            };
+            self.notifier.new_frame_ready(document_id, self.frame_publish_id, &params);
         }
 
         if !doc.hit_tester_is_valid {
@@ -1683,7 +1793,6 @@ impl RenderBackend {
             if config.bits.contains(CaptureBits::FRAME) {
                 // Temporarily force invalidation otherwise the render task graph dump is empty.
                 let force_invalidation = std::mem::replace(&mut doc.scene.config.force_invalidation, true);
-                let frame_memory = FrameMemory::new(self.chunk_pool.clone());
                 let rendered_document = doc.build_frame(
                     &mut self.resource_cache,
                     &mut self.gpu_cache,
@@ -1692,7 +1801,7 @@ impl RenderBackend {
                     None,
                     true,
                     RenderReasons::empty(),
-                    frame_memory,
+                    self.chunk_pool.clone(),
                 );
 
                 doc.scene.config.force_invalidation = force_invalidation;
@@ -1974,7 +2083,13 @@ impl RenderBackend {
                     );
                     self.result_tx.send(msg_publish).unwrap();
 
-                    self.notifier.new_frame_ready(id, false, true, self.frame_publish_id);
+                    let params = api::FrameReadyParams {
+                        present: true,
+                        render: true,
+                        scrolled: false,
+                        tracked: false,
+                    };
+                    self.notifier.new_frame_ready(id, self.frame_publish_id, &params);
 
                     // We deserialized the state of the frame so we don't want to build
                     // it (but we do want to update the scene builder's state)
