@@ -38,11 +38,6 @@ using namespace mozilla::ipc;
 static NS_DEFINE_CID(kThisIconURIImplementationCID,
                      NS_THIS_ICONURI_IMPLEMENTATION_CID);
 
-static const char* const kSizeStrings[] = {"button", "toolbar", "toolbarsmall",
-                                           "menu",   "dnd",     "dialog"};
-
-static const char* const kStateStrings[] = {"normal", "disabled"};
-
 ////////////////////////////////////////////////////////////////////////////////
 
 NS_IMPL_CLASSINFO(nsMozIconURI, nullptr, nsIClassInfo::THREADSAFE,
@@ -50,10 +45,9 @@ NS_IMPL_CLASSINFO(nsMozIconURI, nullptr, nsIClassInfo::THREADSAFE,
 // Empty CI getter. We only need nsIClassInfo for Serialization
 NS_IMPL_CI_INTERFACE_GETTER0(nsMozIconURI)
 
-nsMozIconURI::nsMozIconURI()
-    : mSize(DEFAULT_IMAGE_SIZE), mIconSize(-1), mIconState(-1) {}
+nsMozIconURI::nsMozIconURI() : mSize(DEFAULT_IMAGE_SIZE) {}
 
-nsMozIconURI::~nsMozIconURI() {}
+nsMozIconURI::~nsMozIconURI() = default;
 
 NS_IMPL_ADDREF(nsMozIconURI)
 NS_IMPL_RELEASE(nsMozIconURI)
@@ -94,24 +88,27 @@ nsMozIconURI::GetSpec(nsACString& aSpec) {
   }
 
   aSpec += "?size=";
-  if (mIconSize >= 0) {
-    aSpec += kSizeStrings[mIconSize];
-  } else {
+  {
     char buf[20];
     SprintfLiteral(buf, "%d", mSize);
     aSpec.Append(buf);
   }
 
-  if (mIconState >= 0) {
-    aSpec += "&state=";
-    aSpec += kStateStrings[mIconState];
-  }
-
   if (!mContentType.IsEmpty()) {
     aSpec += "&contentType=";
     aSpec += mContentType.get();
   }
 
+  if (mScale != 1) {
+    aSpec += "&scale=";
+    aSpec.AppendInt(mScale);
+  }
+
+  if (mDark) {
+    aSpec += "&dark=";
+    aSpec.AppendInt(*mDark ? 1 : 0);
+  }
+
   return NS_OK;
 }
 
@@ -208,8 +205,8 @@ nsresult nsMozIconURI::SetSpecInternal(const nsACString& aSpec) {
   mContentType.Truncate();
   mFileName.Truncate();
   mStockIcon.Truncate();
-  mIconSize = -1;
-  mIconState = -1;
+  mScale = 1;
+  mDark.reset();
 
   nsAutoCString iconSpec(aSpec);
   if (!Substring(iconSpec, 0, MOZICON_SCHEME_LEN)
@@ -229,31 +226,27 @@ nsresult nsMozIconURI::SetSpecInternal(const nsACString& aSpec) {
     nsAutoCString sizeString;
     extractAttributeValue(iconSpec.get(), "size=", sizeString);
     if (!sizeString.IsEmpty()) {
-      const char* sizeStr = sizeString.get();
-      for (uint32_t i = 0; i < std::size(kSizeStrings); i++) {
-        if (nsCRT::strcasecmp(sizeStr, kSizeStrings[i]) == 0) {
-          mIconSize = i;
-          break;
-        }
-      }
-
       int32_t sizeValue = atoi(sizeString.get());
       if (sizeValue > 0) {
         mSize = sizeValue;
       }
     }
 
-    nsAutoCString stateString;
-    extractAttributeValue(iconSpec.get(), "state=", stateString);
-    if (!stateString.IsEmpty()) {
-      const char* stateStr = stateString.get();
-      for (uint32_t i = 0; i < std::size(kStateStrings); i++) {
-        if (nsCRT::strcasecmp(stateStr, kStateStrings[i]) == 0) {
-          mIconState = i;
-          break;
-        }
+    nsAutoCString scaleString;
+    extractAttributeValue(iconSpec.get(), "scale=", scaleString);
+    if (!scaleString.IsEmpty()) {
+      int32_t scaleValue = atoi(scaleString.get());
+      if (scaleValue > 0) {
+        mScale = scaleValue;
       }
     }
+
+    nsAutoCString darkString;
+    extractAttributeValue(iconSpec.get(), "dark=", darkString);
+    if (!darkString.IsEmpty()) {
+      int32_t darkValue = atoi(darkString.get());
+      mDark = Some(darkValue != 0);
+    }
   }
 
   int32_t pathLength = iconSpec.Length() - MOZICON_SCHEME_LEN;
@@ -464,8 +457,6 @@ nsresult nsMozIconURI::Clone(nsIURI** result) {
   uri->mContentType = mContentType;
   uri->mFileName = mFileName;
   uri->mStockIcon = mStockIcon;
-  uri->mIconSize = mIconSize;
-  uri->mIconState = mIconState;
   uri.forget(result);
 
   return NS_OK;
@@ -498,13 +489,26 @@ nsMozIconURI::GetIconURL(nsIURL** aFileUrl) {
 }
 
 NS_IMETHODIMP
-nsMozIconURI::GetImageSize(uint32_t* aImageSize)
-// measured by # of pixels in a row. defaults to 16.
-{
+nsMozIconURI::GetImageSize(uint32_t* aImageSize) {
   *aImageSize = mSize;
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsMozIconURI::GetImageScale(uint32_t* aImageScale) {
+  *aImageScale = mScale;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsMozIconURI::GetImageDark(bool* aImageDark) {
+  if (!mDark) {
+    return NS_ERROR_FAILURE;
+  }
+  *aImageDark = *mDark;
+  return NS_OK;
+}
+
 NS_IMETHODIMP
 nsMozIconURI::GetContentType(nsACString& aContentType) {
   aContentType = mContentType;
@@ -546,26 +550,6 @@ nsMozIconURI::GetStockIcon(nsACString& aStockIcon) {
   return NS_OK;
 }
 
-NS_IMETHODIMP
-nsMozIconURI::GetIconSize(nsACString& aSize) {
-  if (mIconSize >= 0) {
-    aSize = kSizeStrings[mIconSize];
-  } else {
-    aSize.Truncate();
-  }
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-nsMozIconURI::GetIconState(nsACString& aState) {
-  if (mIconState >= 0) {
-    aState = kStateStrings[mIconState];
-  } else {
-    aState.Truncate();
-  }
-  return NS_OK;
-}
-
 void nsMozIconURI::Serialize(URIParams& aParams) {
   IconURIParams params;
 
@@ -583,10 +567,12 @@ void nsMozIconURI::Serialize(URIParams& aParams) {
   }
 
   params.size() = mSize;
+  params.contentType() = mContentType;
   params.fileName() = mFileName;
   params.stockIcon() = mStockIcon;
-  params.iconSize() = mIconSize;
-  params.iconState() = mIconState;
+
+  params.iconScale() = mScale;
+  params.iconDark() = mDark;
 
   aParams = params;
 }
@@ -612,17 +598,8 @@ bool nsMozIconURI::Deserialize(const URIParams& aParams) {
   mFileName = params.fileName();
   mStockIcon = params.stockIcon();
 
-  if (params.iconSize() < -1 ||
-      params.iconSize() >= (int32_t)std::size(kSizeStrings)) {
-    return false;
-  }
-  mIconSize = params.iconSize();
-
-  if (params.iconState() < -1 ||
-      params.iconState() >= (int32_t)std::size(kStateStrings)) {
-    return false;
-  }
-  mIconState = params.iconState();
+  mScale = params.iconScale();
+  mDark = params.iconDark();
 
   return true;
 }
