@@ -15,7 +15,6 @@
 #include "imgINotificationObserver.h"
 #include "imgLoader.h"
 #include "mozilla/dom/Document.h"
-#include "mozilla/Telemetry.h"     // for Telemetry
 #include "mozilla/dom/DocGroup.h"  // for DocGroup
 #include "nsCRTGlue.h"
 #include "nsError.h"
@@ -523,6 +522,12 @@ imgIContainer::DecodeResult imgRequestProxy::RequestDecodeWithResult(
   return imgIContainer::DECODE_REQUESTED;
 }
 
+NS_IMETHODIMP
+imgRequestProxy::GetHasAnimationConsumers(bool* aIsLocked) {
+  *aIsLocked = !!mAnimationConsumers;
+  return NS_OK;
+}
+
 NS_IMETHODIMP
 imgRequestProxy::LockImage() {
   mLockCount++;
