@@ -27,6 +27,7 @@
 #include "builtin/OrderedHashTableObject-inl.h"
 #include "gc/GCContext-inl.h"
 #include "gc/Marking-inl.h"
+#include "gc/ObjectKind-inl.h"
 #include "vm/GeckoProfiler-inl.h"
 #include "vm/NativeObject-inl.h"
 
@@ -116,9 +117,7 @@ bool HashableValue::equals(const HashableValue& other) const {
 #ifdef DEBUG
   bool same;
   JSContext* cx = TlsContext.get();
-  RootedValue valueRoot(cx, value);
-  RootedValue otherRoot(cx, other.value);
-  MOZ_ASSERT(SameValueZero(cx, valueRoot, otherRoot, &same));
+  MOZ_ASSERT(SameValueZero(cx, value, other.value, &same));
   MOZ_ASSERT(same == b);
 #endif
   return b;
@@ -176,7 +175,7 @@ bool GlobalObject::initMapIteratorProto(JSContext* cx,
       !DefineToStringTag(cx, proto, cx->names().Map_Iterator_)) {
     return false;
   }
-  if (!JSObject::setHasFuseProperty(cx, proto)) {
+  if (!JSObject::setHasRealmFuseProperty(cx, proto)) {
     return false;
   }
   global->initBuiltinProto(ProtoKind::MapIteratorProto, proto);
@@ -296,8 +295,8 @@ bool MapIteratorObject::next(MapIteratorObject* mapIterator,
 
 /* static */
 JSObject* MapIteratorObject::createResultPair(JSContext* cx) {
-  Rooted<ArrayObject*> resultPairObj(
-      cx, NewDenseFullyAllocatedArray(cx, 2, TenuredObject));
+  ArrayObject* resultPairObj =
+      NewDenseFullyAllocatedArray(cx, 2, TenuredObject);
   if (!resultPairObj) {
     return nullptr;
   }
@@ -426,7 +425,7 @@ const JSFunctionSpec MapObject::staticMethods[] = {
     return false;
   }
 
-  return JSObject::setHasFuseProperty(cx, nativeProto);
+  return JSObject::setHasRealmFuseProperty(cx, nativeProto);
 }
 
 void MapObject::trace(JSTracer* trc, JSObject* obj) {
@@ -610,6 +609,8 @@ MapObject* MapObject::createWithProto(JSContext* cx, HandleObject proto,
                                       NewObjectKind newKind) {
   MOZ_ASSERT(proto);
 
+  gc::AllocKind allocKind = gc::GetGCObjectKind(SlotCount);
+
   AutoSetNewObjectMetadata metadata(cx);
   auto* mapObj =
       NewObjectWithGivenProtoAndKinds<MapObject>(cx, proto, allocKind, newKind);
@@ -625,10 +626,6 @@ MapObject* MapObject::createWithProto(JSContext* cx, HandleObject proto,
 
 MapObject* MapObject::create(JSContext* cx,
                              HandleObject proto /* = nullptr */) {
-  MOZ_ASSERT(gc::ForegroundToBackgroundAllocKind(
-                 gc::GetGCObjectKind(SlotCount)) == allocKind,
-             "allocKind constant doesn't match SlotCount");
-
   if (proto) {
     return createWithProto(cx, proto, GenericObject);
   }
@@ -640,6 +637,10 @@ MapObject* MapObject::create(JSContext* cx,
     return nullptr;
   }
 
+  gc::AllocKind allocKind = templateObj->asTenured().getAllocKind();
+  MOZ_ASSERT(gc::GetGCKindSlots(allocKind) >= SlotCount);
+  MOZ_ASSERT(gc::IsBackgroundFinalized(allocKind));
+
   AutoSetNewObjectMetadata metadata(cx);
   Rooted<SharedShape*> shape(cx, templateObj->sharedShape());
   auto* mapObj =
@@ -1119,7 +1120,7 @@ bool GlobalObject::initSetIteratorProto(JSContext* cx,
       !DefineToStringTag(cx, proto, cx->names().Set_Iterator_)) {
     return false;
   }
-  if (!JSObject::setHasFuseProperty(cx, proto)) {
+  if (!JSObject::setHasRealmFuseProperty(cx, proto)) {
     return false;
   }
   global->initBuiltinProto(ProtoKind::SetIteratorProto, proto);
@@ -1215,8 +1216,7 @@ bool SetIteratorObject::next(SetIteratorObject* setIterator,
 
 /* static */
 JSObject* SetIteratorObject::createResult(JSContext* cx) {
-  Rooted<ArrayObject*> resultObj(
-      cx, NewDenseFullyAllocatedArray(cx, 1, TenuredObject));
+  ArrayObject* resultObj = NewDenseFullyAllocatedArray(cx, 1, TenuredObject);
   if (!resultObj) {
     return nullptr;
   }
@@ -1331,7 +1331,7 @@ const JSPropertySpec SetObject::staticProperties[] = {
     return false;
   }
 
-  return JSObject::setHasFuseProperty(cx, nativeProto);
+  return JSObject::setHasRealmFuseProperty(cx, nativeProto);
 }
 
 bool SetObject::keys(JS::MutableHandle<GCVector<JS::Value>> keys) {
@@ -1360,6 +1360,8 @@ SetObject* SetObject::createWithProto(JSContext* cx, HandleObject proto,
                                       NewObjectKind newKind) {
   MOZ_ASSERT(proto);
 
+  gc::AllocKind allocKind = gc::GetGCObjectKind(SlotCount);
+
   AutoSetNewObjectMetadata metadata(cx);
   auto* setObj =
       NewObjectWithGivenProtoAndKinds<SetObject>(cx, proto, allocKind, newKind);
@@ -1375,10 +1377,6 @@ SetObject* SetObject::createWithProto(JSContext* cx, HandleObject proto,
 
 SetObject* SetObject::create(JSContext* cx,
                              HandleObject proto /* = nullptr */) {
-  MOZ_ASSERT(gc::ForegroundToBackgroundAllocKind(
-                 gc::GetGCObjectKind(SlotCount)) == allocKind,
-             "allocKind constant doesn't match SlotCount");
-
   if (proto) {
     return createWithProto(cx, proto, GenericObject);
   }
@@ -1390,6 +1388,10 @@ SetObject* SetObject::create(JSContext* cx,
     return nullptr;
   }
 
+  gc::AllocKind allocKind = templateObj->asTenured().getAllocKind();
+  MOZ_ASSERT(gc::GetGCKindSlots(allocKind) >= SlotCount);
+  MOZ_ASSERT(gc::IsBackgroundFinalized(allocKind));
+
   AutoSetNewObjectMetadata metadata(cx);
   Rooted<SharedShape*> shape(cx, templateObj->sharedShape());
   auto* setObj =
