@@ -1539,6 +1539,8 @@ static bool BytecodeIsEffectful(JSScript* script, size_t offset) {
     case JSOp::Ne:
     case JSOp::StrictEq:
     case JSOp::StrictNe:
+    case JSOp::StrictConstantEq:
+    case JSOp::StrictConstantNe:
     case JSOp::Lt:
     case JSOp::Le:
     case JSOp::Gt:
@@ -2113,6 +2115,7 @@ struct DebuggerScript::SetBreakpointMatcher {
 
     if (!cx_->zone()->new_<Breakpoint>(dbg_, debuggerObject_, site, handler_)) {
       site->destroyIfEmpty(cx_->runtime()->gcContext());
+      ReportOutOfMemory(cx_);
       return false;
     }
     AddCellMemory(script, sizeof(Breakpoint), MemoryUse::Breakpoint);
@@ -2142,6 +2145,7 @@ struct DebuggerScript::SetBreakpointMatcher {
 
     if (!cx_->zone()->new_<Breakpoint>(dbg_, debuggerObject_, site, handler_)) {
       site->destroyIfEmpty(cx_->runtime()->gcContext());
+      ReportOutOfMemory(cx_);
       return false;
     }
     AddCellMemory(wasmInstance, sizeof(Breakpoint), MemoryUse::Breakpoint);
