@@ -85,17 +85,6 @@ struct LargeBuffer;
 class StoreBuffer;
 class TenuringTracer;
 
-// A set of cells that need to be swept at the end of a minor GC,
-// represented as a linked list of ArenaCellSet structs extracted from a
-// WholeCellBuffer.
-struct CellSweepSet {
-  UniquePtr<LifoAlloc> storage_;
-  ArenaCellSet* head_ = nullptr;
-
-  // Fixup the tenured dependent strings stored in the ArenaCellSet list.
-  void sweep();
-};
-
 }  // namespace gc
 
 class Nursery {
@@ -381,7 +370,7 @@ class Nursery {
 
   bool shouldTenureEverything(JS::GCReason reason);
 
-  inline bool inCollectedRegion(gc::Cell* cell) const;
+  inline bool inCollectedRegion(const gc::Cell* cell) const;
   inline bool inCollectedRegion(void* ptr) const;
 
   void trackMallocedBufferOnPromotion(void* buffer, gc::Cell* owner,
@@ -399,6 +388,8 @@ class Nursery {
     return mallocedBlockCache_.sizeOfExcludingThis(mallocSizeOf);
   }
 
+  inline void addMallocedBufferBytes(size_t nbytes);
+
   mozilla::TimeStamp lastCollectionEndTime() const;
 
  private:
@@ -447,8 +438,6 @@ class Nursery {
            (maxChunkCount() - currentChunk() - 1) * gc::ChunkSize;
   }
 
-  inline void addMallocedBufferBytes(size_t nbytes);
-
   // Calculate the promotion rate of the most recent minor GC.
   // The valid_for_tenuring parameter is used to return whether this
   // promotion rate is accurate enough (the nursery was full enough) to be
@@ -487,7 +476,8 @@ class Nursery {
   const js::gc::GCSchedulingTunables& tunables() const;
 
   void getAllocFlagsForZone(JS::Zone* zone, bool* allocObjectsOut,
-                            bool* allocStringsOut, bool* allocBigIntsOut);
+                            bool* allocStringsOut, bool* allocBigIntsOut,
+                            bool* allocGetterSettersOut);
   void updateAllZoneAllocFlags();
   void updateAllocFlagsForZone(JS::Zone* zone);
   void discardCodeAndSetJitFlagsForZone(JS::Zone* zone);
@@ -531,8 +521,7 @@ class Nursery {
   bool checkForwardingPointerInsideNursery(void* ptr);
 #endif
 
-  // Updates pointers to nursery objects that have been tenured and discards
-  // pointers to objects that have been freed.
+  // Discard pointers to objects that have been freed.
   void sweep();
 
   // In a minor GC, resets the start and end positions, the current chunk and
@@ -565,6 +554,9 @@ class Nursery {
   // vector. Shrinks the vector but does not update maxChunkCount().
   void freeChunksFrom(Space& space, unsigned firstFreeChunk);
 
+  // During a semispace nursery collection, return whether a cell in fromspace
+  // was in the tospace of the previous collection, meaning that it should be
+  // tenured in this collection.
   inline bool shouldTenure(gc::Cell* cell);
 
   void sendTelemetry(JS::GCReason reason, mozilla::TimeDuration totalTime,
@@ -733,8 +725,6 @@ class Nursery {
       HashMap<void*, void*, PointerHasher<void*>, SystemAllocPolicy>;
   ForwardedBufferMap forwardedBuffers;
 
-  gc::CellSweepSet cellsToSweep;
-
   // When we assign a unique id to cell in the nursery, that almost always
   // means that the cell will be in a hash table, and thus, held live,
   // automatically moving the uid from the nursery to its new home in
@@ -777,9 +767,6 @@ class Nursery {
       Vector<mozilla::StringBuffer*, 8, SystemAllocPolicy>;
   StringBufferVector stringBuffersToReleaseAfterMinorGC_;
 
-  using LargeAllocList = SlimLinkedList<gc::LargeBuffer>;
-  LargeAllocList largeAllocsToFreeAfterMinorGC_;
-
   UniquePtr<NurserySweepTask> sweepTask;
   UniquePtr<NurseryDecommitTask> decommitTask;
 
@@ -816,13 +803,6 @@ MOZ_ALWAYS_INLINE bool Nursery::Space::isInside(const void* p) const {
   return false;
 }
 
-// Test whether a GC cell or buffer is in the nursery. Equivalent to
-// IsInsideNursery but take care not to call this with malloc memory. Faster
-// than Nursery::isInside.
-MOZ_ALWAYS_INLINE bool ChunkPtrIsInsideNursery(void* ptr) {
-  return gc::detail::ChunkPtrHasStoreBuffer(ptr);
-}
-
 }  // namespace js
 
 #endif  // gc_Nursery_h
