@@ -43,6 +43,7 @@
 #include "vm/BigIntType.h"
 #include "vm/HelperThreads.h"
 #include "vm/JSContext.h"
+#include "vm/Probes.h"
 #include "vm/Time.h"
 #include "vm/WrapperObject.h"
 
@@ -61,48 +62,71 @@ using mozilla::TimeStamp;
 
 using JS::SliceBudget;
 
-struct js::gc::FinalizePhase {
-  gcstats::PhaseKind statsPhase;
-  AllocKinds kinds;
-};
-
 /*
- * Finalization order for objects swept incrementally on the main thread.
+ * Sweeping of arenas and possible finalization of dead cells proceeds in a
+ * sequence of phases.
+ *
+ *  1. ForegroundObjectFinalizePhase
+ *     JSObjects with finalizers. Swept incrementally on the main thread.
+ *
+ *  2. ForegroundNonObjectFinalizePhase
+ *     Non-JSObjects with finalizers. Swept incrementally on the main thread.
+ *
+ *  3. BackgroundObjectFinalizePhase
+ *     JSObjects with finalizers that can run off main thread. Swept
+ *     non-incrementally on a helper thread.
+ *
+ *  4. BackgroundTrivialFinalizePhase
+ *     Everything else. These may or may not have finalizers. Any finalizers
+ *     must not delete HeapPtrs. Swept non-incrementally on a helper thread.
  */
-static constexpr FinalizePhase ForegroundObjectFinalizePhase = {
-    gcstats::PhaseKind::FINALIZE_OBJECT,
-    {AllocKind::OBJECT0, AllocKind::OBJECT2, AllocKind::OBJECT4,
-     AllocKind::OBJECT8, AllocKind::OBJECT12, AllocKind::OBJECT16}};
 
-/*
- * Finalization order for GC things swept incrementally on the main thread.
- */
-static constexpr FinalizePhase ForegroundNonObjectFinalizePhase = {
-    gcstats::PhaseKind::FINALIZE_NON_OBJECT,
-    {AllocKind::SCRIPT, AllocKind::JITCODE}};
+static constexpr AllocKinds ForegroundObjectFinalizePhase = {
+    AllocKind::OBJECT0_FOREGROUND,  AllocKind::OBJECT2_FOREGROUND,
+    AllocKind::OBJECT4_FOREGROUND,  AllocKind::OBJECT8_FOREGROUND,
+    AllocKind::OBJECT12_FOREGROUND, AllocKind::OBJECT16_FOREGROUND};
 
-/*
- * Finalization order for GC things swept on the background thread.
- */
-static constexpr FinalizePhase BackgroundFinalizePhases[] = {
-    {gcstats::PhaseKind::FINALIZE_OBJECT,
-     {AllocKind::FUNCTION, AllocKind::FUNCTION_EXTENDED,
-      AllocKind::OBJECT0_BACKGROUND, AllocKind::OBJECT2_BACKGROUND,
-      AllocKind::ARRAYBUFFER4, AllocKind::OBJECT4_BACKGROUND,
-      AllocKind::ARRAYBUFFER8, AllocKind::OBJECT8_BACKGROUND,
-      AllocKind::ARRAYBUFFER12, AllocKind::OBJECT12_BACKGROUND,
-      AllocKind::ARRAYBUFFER16, AllocKind::OBJECT16_BACKGROUND}},
-    {gcstats::PhaseKind::FINALIZE_NON_OBJECT,
-     {AllocKind::BUFFER16, AllocKind::BUFFER32, AllocKind::BUFFER64,
-      AllocKind::BUFFER128, AllocKind::SCOPE, AllocKind::REGEXP_SHARED,
-      AllocKind::FAT_INLINE_STRING, AllocKind::STRING,
-      AllocKind::EXTERNAL_STRING, AllocKind::FAT_INLINE_ATOM, AllocKind::ATOM,
-      AllocKind::SYMBOL, AllocKind::BIGINT, AllocKind::SHAPE,
-      AllocKind::BASE_SHAPE, AllocKind::GETTER_SETTER,
-      AllocKind::COMPACT_PROP_MAP, AllocKind::NORMAL_PROP_MAP,
-      AllocKind::DICT_PROP_MAP}}};
+static constexpr AllocKinds ForegroundNonObjectFinalizePhase = {
+    AllocKind::SCRIPT, AllocKind::JITCODE};
 
-template <typename T>
+static constexpr AllocKinds BackgroundObjectFinalizePhase = {
+    AllocKind::OBJECT0_BACKGROUND, AllocKind::OBJECT2_BACKGROUND,
+    AllocKind::ARRAYBUFFER4,       AllocKind::OBJECT4_BACKGROUND,
+    AllocKind::ARRAYBUFFER8,       AllocKind::OBJECT8_BACKGROUND,
+    AllocKind::ARRAYBUFFER12,      AllocKind::OBJECT12_BACKGROUND,
+    AllocKind::ARRAYBUFFER16,      AllocKind::OBJECT16_BACKGROUND};
+
+static constexpr AllocKinds BackgroundTrivialFinalizePhase = {
+    AllocKind::FUNCTION,
+    AllocKind::FUNCTION_EXTENDED,
+    AllocKind::OBJECT0,
+    AllocKind::OBJECT2,
+    AllocKind::OBJECT4,
+    AllocKind::OBJECT8,
+    AllocKind::OBJECT12,
+    AllocKind::OBJECT16,
+    AllocKind::SCOPE,
+    AllocKind::REGEXP_SHARED,
+    AllocKind::FAT_INLINE_STRING,
+    AllocKind::STRING,
+    AllocKind::EXTERNAL_STRING,
+    AllocKind::FAT_INLINE_ATOM,
+    AllocKind::ATOM,
+    AllocKind::SYMBOL,
+    AllocKind::BIGINT,
+    AllocKind::SHAPE,
+    AllocKind::BASE_SHAPE,
+    AllocKind::GETTER_SETTER,
+    AllocKind::COMPACT_PROP_MAP,
+    AllocKind::NORMAL_PROP_MAP,
+    AllocKind::DICT_PROP_MAP};
+
+static constexpr AllocKinds AllBackgroundSweptKinds =
+    BackgroundObjectFinalizePhase + BackgroundTrivialFinalizePhase;
+
+static constexpr size_t ArenaReleaseBatchSize = 32;
+
+template <typename T, FinalizeKind finalizeKind>
 inline size_t Arena::finalize(JS::GCContext* gcx, AllocKind thingKind,
                               size_t thingSize) {
   /* Enforce requirements on size of T. */
@@ -115,6 +139,8 @@ inline size_t Arena::finalize(JS::GCContext* gcx, AllocKind thingKind,
   MOZ_ASSERT(thingSize == getThingSize());
   MOZ_ASSERT(!onDelayedMarkingList_);
 
+  MOZ_ASSERT(finalizeKind == GetFinalizeKind(thingKind));
+
   uint_fast16_t freeStart = firstThingOffset(thingKind);
 
   // Update the free list as we go along. The cell iterator will always be ahead
@@ -138,7 +164,12 @@ inline size_t Arena::finalize(JS::GCContext* gcx, AllocKind thingKind,
       freeStart = thing + thingSize;
       nmarked++;
     } else {
-      t->finalize(gcx);
+      if constexpr (std::is_same_v<T, JSObject>) {
+        js::probes::FinalizeObject(t);
+      }
+      if constexpr (finalizeKind != FinalizeKind::None) {
+        t->finalize(gcx);
+      }
       AlwaysPoison(t, JS_SWEPT_TENURED_PATTERN, thingSize,
                    MemCheckKind::MakeUndefined);
       gcprobes::TenuredFinalize(t);
@@ -176,7 +207,7 @@ inline size_t Arena::finalize(JS::GCContext* gcx, AllocKind thingKind,
 // Finalize arenas from src list, releasing empty arenas if keepArenas wasn't
 // specified and inserting the others into the appropriate destination size
 // bins.
-template <typename T>
+template <typename T, FinalizeKind finalizeKind, ReleaseEmpty releaseEmpty>
 static inline bool FinalizeTypedArenas(JS::GCContext* gcx, ArenaList& src,
                                        SortedArenaList& dest,
                                        AllocKind thingKind,
@@ -186,41 +217,69 @@ static inline bool FinalizeTypedArenas(JS::GCContext* gcx, ArenaList& src,
   size_t thingSize = Arena::thingSize(thingKind);
   size_t thingsPerArena = Arena::thingsPerArena(thingKind);
   size_t markCount = 0;
+  size_t emptyCount = 0;
 
-  auto updateMarkCount = mozilla::MakeScopeExit([&] {
-    GCRuntime* gc = &gcx->runtimeFromAnyThread()->gc;
-    gc->stats().addCount(gcstats::COUNT_CELLS_MARKED, markCount);
-  });
+  GCRuntime* gc = &gcx->runtimeFromAnyThread()->gc;
+  auto updateMarkCount = mozilla::MakeScopeExit(
+      [&] { gc->stats().addCount(gcstats::COUNT_CELLS_MARKED, markCount); });
 
   while (!src.isEmpty()) {
     Arena* arena = src.popFront();
-    size_t nmarked = arena->finalize<T>(gcx, thingKind, thingSize);
+    size_t nmarked =
+        arena->finalize<T, finalizeKind>(gcx, thingKind, thingSize);
     size_t nfree = thingsPerArena - nmarked;
 
     markCount += nmarked;
 
     dest.insertAt(arena, nfree);
 
+    if constexpr (bool(releaseEmpty)) {
+      if (nmarked == 0) {
+        emptyCount++;
+        MOZ_ASSERT(emptyCount <= ArenaReleaseBatchSize);
+        if (emptyCount == ArenaReleaseBatchSize) {
+          Arena* emptyArenas = nullptr;
+          dest.extractEmptyTo(&emptyArenas);
+          emptyArenas =
+              gc->releaseSomeEmptyArenas(emptyArenas->zone(), emptyArenas);
+          MOZ_ASSERT(!emptyArenas);
+          emptyCount = 0;
+        }
+      }
+    }
+
     budget.step(thingsPerArena);
     if (budget.isOverBudget()) {
       return false;
     }
   }
 
+  if constexpr (bool(releaseEmpty)) {
+    if (emptyCount) {
+      Arena* emptyArenas = nullptr;
+      dest.extractEmptyTo(&emptyArenas);
+      emptyArenas =
+          gc->releaseSomeEmptyArenas(emptyArenas->zone(), emptyArenas);
+      MOZ_ASSERT(!emptyArenas);
+    }
+  }
+
   return true;
 }
 
 /*
  * Finalize the list of areans.
  */
+template <ReleaseEmpty releaseEmpty>
 static bool FinalizeArenas(JS::GCContext* gcx, ArenaList& src,
                            SortedArenaList& dest, AllocKind thingKind,
                            SliceBudget& budget) {
   switch (thingKind) {
-#define EXPAND_CASE(allocKind, traceKind, type, sizedType, bgFinal, nursery, \
-                    compact)                                                 \
-  case AllocKind::allocKind:                                                 \
-    return FinalizeTypedArenas<type>(gcx, src, dest, thingKind, budget);
+#define EXPAND_CASE(allocKind, _1, type, _2, finalizeKind, _3, _4)      \
+  case AllocKind::allocKind:                                            \
+    return FinalizeTypedArenas<type, FinalizeKind::finalizeKind,        \
+                               releaseEmpty>(gcx, src, dest, thingKind, \
+                                             budget);
     FOR_EACH_ALLOCKIND(EXPAND_CASE)
 #undef EXPAND_CASE
 
@@ -230,15 +289,14 @@ static bool FinalizeArenas(JS::GCContext* gcx, ArenaList& src,
 }
 
 void GCRuntime::initBackgroundSweep(Zone* zone, JS::GCContext* gcx,
-                                    const FinalizePhase& phase) {
-  gcstats::AutoPhase ap(stats(), phase.statsPhase);
-  for (auto kind : phase.kinds) {
+                                    const AllocKinds& kinds) {
+  for (AllocKind kind : kinds) {
     zone->arenas.initBackgroundSweep(kind);
   }
 }
 
 void ArenaLists::initBackgroundSweep(AllocKind thingKind) {
-  MOZ_ASSERT(IsBackgroundFinalized(thingKind));
+  MOZ_ASSERT(IsBackgroundSwept(thingKind));
   MOZ_ASSERT(concurrentUse(thingKind) == ConcurrentUse::None);
 
   if (!collectingArenaList(thingKind).isEmpty()) {
@@ -246,69 +304,78 @@ void ArenaLists::initBackgroundSweep(AllocKind thingKind) {
   }
 }
 
-void GCRuntime::backgroundFinalize(JS::GCContext* gcx, Zone* zone,
-                                   AllocKind kind, Arena** empty) {
-  MOZ_ASSERT(empty);
+template <ReleaseEmpty releaseEmpty>
+void ArenaLists::backgroundFinalize(JS::GCContext* gcx, AllocKind kind,
+                                    Arena** empty) {
+  MOZ_ASSERT(IsBackgroundSwept(kind));
+  MOZ_ASSERT(bool(empty) != bool(releaseEmpty));
 
-  ArenaLists* lists = &zone->arenas;
-  ArenaList& arenas = lists->collectingArenaList(kind);
+  ArenaList& arenas = collectingArenaList(kind);
   if (arenas.isEmpty()) {
-    MOZ_ASSERT(lists->concurrentUse(kind) == ArenaLists::ConcurrentUse::None);
+    MOZ_ASSERT(concurrentUse(kind) == ConcurrentUse::None);
     return;
   }
+  MOZ_ASSERT(concurrentUse(kind) == ConcurrentUse::BackgroundFinalize);
 
   SortedArenaList finalizedSorted(kind);
 
   auto unlimited = SliceBudget::unlimited();
-  FinalizeArenas(gcx, arenas, finalizedSorted, kind, unlimited);
+  FinalizeArenas<releaseEmpty>(gcx, arenas, finalizedSorted, kind, unlimited);
   MOZ_ASSERT(arenas.isEmpty());
 
-  finalizedSorted.extractEmptyTo(empty);
-
-  // When marking begins, all arenas are moved from arenaLists to
-  // collectingArenaLists. When the mutator runs, new arenas are allocated in
-  // arenaLists. Now that finalization is complete, we want to merge these lists
-  // back together.
-
-  // We must take the GC lock to be able to safely modify the ArenaList;
-  // however, this does not by itself make the changes visible to all threads,
-  // as not all threads take the GC lock to read the ArenaLists.
-  // That safety is provided by the ReleaseAcquire memory ordering of the
-  // background finalize state, which we explicitly set as the final step.
-  {
-    AutoLockGC lock(rt);
-    MOZ_ASSERT(lists->concurrentUse(kind) ==
-               ArenaLists::ConcurrentUse::BackgroundFinalize);
-    lists->mergeFinalizedArenas(kind, finalizedSorted);
+  if constexpr (!bool(releaseEmpty)) {
+    finalizedSorted.extractEmptyTo(empty);
   }
+  MOZ_ASSERT(!finalizedSorted.hasEmptyArenas());
 
-  lists->concurrentUse(kind) = ArenaLists::ConcurrentUse::None;
+  // Set the collectingArenaList to the possibly empty list of swept arenas
+  // while holding the GC lock. Set concurrentUse to indicate to the main thread
+  // that sweeping has finished.
+  ArenaList sweptArenas = finalizedSorted.convertToArenaList();
+
+  AutoLockGC lock(gcx->runtimeFromAnyThread());
+  collectingArenaList(kind) = std::move(sweptArenas);
+  concurrentUse(kind) = ConcurrentUse::BackgroundFinalizeFinished;
 }
 
-// After finalizing arenas, merge the following to get the final state of an
-// arena list:
+void ArenaLists::mergeBackgroundSweptArenas() {
+  // Merge swept arenas into main arena lists on the main thread.
+  MOZ_ASSERT(CurrentThreadCanAccessRuntime(runtime()));
+
+  for (AllocKind kind : AllBackgroundSweptKinds) {
+    maybeMergeSweptArenas(kind);
+  }
+}
+
+void ArenaLists::maybeMergeSweptArenas(AllocKind kind) {
+  MOZ_ASSERT(zone_->isGCFinished());
+  MOZ_ASSERT(concurrentUse(kind) != ConcurrentUse::BackgroundFinalize);
+
+  if (concurrentUse(kind) == ConcurrentUse::BackgroundFinalizeFinished) {
+    concurrentUse(kind) = ConcurrentUse::None;
+    mergeSweptArenas(kind, collectingArenaList(kind));
+  }
+
+  MOZ_ASSERT(collectingArenaList(kind).isEmpty());
+}
+
+// This methods merges the following to get the final state of an arena
+// list:
+//  - swept arenas
 //  - arenas allocated during marking
 //  - arenas allocated during sweeping
-//  - finalized arenas
-void ArenaLists::mergeFinalizedArenas(AllocKind kind,
-                                      SortedArenaList& finalizedArenas) {
-#ifdef DEBUG
-  // Updating arena lists off-thread requires taking the GC lock because the
-  // main thread uses these when allocating.
-  if (IsBackgroundFinalized(kind)) {
-    runtimeFromAnyThread()->gc.assertCurrentThreadHasLockedGC();
-  }
-  MOZ_ASSERT(collectingArenaList(kind).isEmpty());
-#endif
+void ArenaLists::mergeSweptArenas(AllocKind kind, ArenaList& sweptArenas) {
+  MOZ_ASSERT(CurrentThreadCanAccessRuntime(runtime()));
+  MOZ_ASSERT(concurrentUse(kind) == ConcurrentUse::None);
 
-  arenaList(kind).prepend(finalizedArenas.convertToArenaList());
+  arenaList(kind).prepend(std::move(sweptArenas));
 }
 
 void ArenaLists::queueForegroundThingsForSweep() {
   gcCompactPropMapArenasToUpdate =
-      collectingArenaList(AllocKind::COMPACT_PROP_MAP).first();
+      collectingArenaList(AllocKind::COMPACT_PROP_MAP).getFirst();
   gcNormalPropMapArenasToUpdate =
-      collectingArenaList(AllocKind::NORMAL_PROP_MAP).first();
+      collectingArenaList(AllocKind::NORMAL_PROP_MAP).getFirst();
 }
 
 void GCRuntime::sweepBackgroundThings(ZoneList& zones) {
@@ -327,85 +394,100 @@ void GCRuntime::sweepBackgroundThings(ZoneList& zones) {
 
     TimeStamp startTime = TimeStamp::Now();
 
-    Arena* emptyArenas = zone->arenas.takeSweptEmptyArenas();
+    ArenaLists& arenaLists = zone->arenas;
+    Arena* emptyArenas = arenaLists.takeSweptEmptyArenas();
 
-    // We must finalize thing kinds in the order specified by
-    // BackgroundFinalizePhases.
+    // We must finalize kinds in the order specified at the top of this file.
 
-    for (const auto& phase : BackgroundFinalizePhases) {
-      for (auto kind : phase.kinds) {
-        backgroundFinalize(gcx, zone, kind, &emptyArenas);
-      }
+    for (auto kind : BackgroundObjectFinalizePhase) {
+      MOZ_ASSERT(IsBackgroundFinalized(kind));
+      arenaLists.backgroundFinalize<ReleaseEmpty::No>(gcx, kind, &emptyArenas);
     }
 
     // Release any arenas that are now empty.
     //
-    // Empty arenas are only released after everything has been finalized so
-    // that it's still possible to get a thing's zone after the thing has been
-    // finalized. The HeapPtr destructor depends on this, and this allows
-    // HeapPtrs between things of different alloc kind regardless of
-    // finalization order.
+    // Finalizers can still access the zone pointer in now-empty arenas because
+    // of the write barrier in the HeapPtr destructor. This means we can't
+    // release any empty arenas until all such finalizers have been run.
     //
-    // Batch releases so as to periodically drop and reaquire the GC lock every
-    // so often to avoid blocking the main thread from allocating arenas. This
-    // is important for allocation-heavy workloads such as the splay benchmark.
-    //
-    // This block is equivalent to calling GCRuntime::releaseArena on each arena
-    // individually.
-    static constexpr size_t BatchSize = 32;
+    // At this point this has happened and we can release empty arenas
+    // immediately from now on.
+
+    AutoDisallowPreWriteBarrier disallowBarrier(gcx);
+
     while (emptyArenas) {
-      Arena* arenasToRelease[BatchSize];
-      size_t count = 0;
-
-      size_t gcHeapBytesFreed = 0;
-      size_t mallocHeapBytesFreed = 0;
-
-      {
-        mozilla::Maybe<AutoLockGC> maybeLock;
-        if (zone->isAtomsZone()) {
-          // Required to synchronize access to AtomMarkingRuntime.
-          maybeLock.emplace(this);
-        }
-
-        // Take up to BatchSize arenas from emptyArenas list.
-        for (size_t i = 0; emptyArenas && i < BatchSize; i++) {
-          Arena* arena = emptyArenas;
-          emptyArenas = arena->next;
-
-          if (IsBufferAllocKind(arena->getAllocKind())) {
-            mallocHeapBytesFreed += ArenaSize - arena->getFirstThingOffset();
-          } else {
-            gcHeapBytesFreed += ArenaSize;
-          }
-
-          arena->release(this, maybeLock.ptrOr(nullptr));
-          arenasToRelease[i] = arena;
-          count++;
-        }
-      }
-
-      zone->mallocHeapSize.removeBytes(mallocHeapBytesFreed, true);
-      zone->gcHeapSize.removeBytes(gcHeapBytesFreed, true, heapSize);
-
-      AutoLockGC lock(this);
-      for (size_t i = 0; i < count; i++) {
-        Arena* arena = arenasToRelease[i];
-        arena->chunk()->releaseArena(this, arena, lock);
-      }
+      emptyArenas = releaseSomeEmptyArenas(zone, emptyArenas);
     }
 
-    // Bug 1930497: If we had a separate phase for finalizing GC things with
-    // trivial finalizers, we could run this before it, potentially freeing more
-    // memory sooner. This could also happen in parallel with it.
+    // Now everything with a non-trivial finalizer has been finalized we can
+    // sweep buffer memory.
+    //
+    // Note we depend on this happening before the BUFFER alloc kinds in
+    // BackgroundTrivialFinalizePhase are swept!
     bool decommit = shouldDecommit() && DecommitEnabled();
     zone->bufferAllocator.sweepForMajorCollection(decommit);
 
+    // TODO: The remaining sweeping work can be parallelised between multiple
+    // threads.
+    for (AllocKind kind : BackgroundTrivialFinalizePhase) {
+      MOZ_ASSERT(IsBackgroundSwept(kind));
+      arenaLists.backgroundFinalize<ReleaseEmpty::Yes>(gcx, kind);
+    }
+
     // Record time spent sweeping this zone.
     TimeStamp endTime = TimeStamp::Now();
     zone->perZoneGCTime += endTime - startTime;
   }
 }
 
+Arena* GCRuntime::releaseSomeEmptyArenas(Zone* zone, Arena* emptyArenas) {
+  // Batch releases so as to periodically drop and reaquire the GC lock to
+  // avoid blocking the main thread from allocating arenas. This is important
+  // for allocation-heavy workloads such as the splay benchmark.
+  //
+  // This block is equivalent to calling GCRuntime::releaseArena on each arena
+  // individually.
+  bool isAtomsZone = zone->isAtomsZone();
+
+  Arena* arenasToRelease[ArenaReleaseBatchSize];
+  size_t atomsBitmapIndexes[ArenaReleaseBatchSize];
+  size_t count = 0;
+
+  size_t gcHeapBytesFreed = 0;
+
+  // Take up to ArenaReleaseBatchSize arenas from emptyArenas list.
+  for (size_t i = 0; emptyArenas && i < ArenaReleaseBatchSize; i++) {
+    Arena* arena = emptyArenas;
+    emptyArenas = arena->next;
+
+    gcHeapBytesFreed += ArenaSize;
+
+    if (isAtomsZone) {
+      atomsBitmapIndexes[i] = arena->atomBitmapStart();
+#ifdef DEBUG
+      arena->atomBitmapStart() = 0;
+#endif
+    }
+
+    arena->release();
+    arenasToRelease[i] = arena;
+    count++;
+  }
+
+  zone->gcHeapSize.removeBytes(gcHeapBytesFreed, true, heapSize);
+
+  AutoLockGC lock(this);
+  for (size_t i = 0; i < count; i++) {
+    Arena* arena = arenasToRelease[i];
+    if (isAtomsZone) {
+      atomMarking.freeIndex(atomsBitmapIndexes[i], lock);
+    }
+    arena->chunk()->releaseArena(this, arena, lock);
+  }
+
+  return emptyArenas;
+}
+
 void GCRuntime::assertBackgroundSweepingFinished() {
 #ifdef DEBUG
   {
@@ -415,7 +497,7 @@ void GCRuntime::assertBackgroundSweepingFinished() {
 
   for (ZonesIter zone(this, WithAtoms); !zone.done(); zone.next()) {
     for (auto kind : AllAllocKinds()) {
-      MOZ_ASSERT_IF(!(state() >= State::Prepare && state() <= State::Sweep),
+      MOZ_ASSERT_IF(state() == State::NotActive || state() >= State::Compact,
                     zone->arenas.collectingArenaList(kind).isEmpty());
       MOZ_ASSERT(zone->arenas.doneBackgroundFinalize(kind));
     }
@@ -463,6 +545,11 @@ void GCRuntime::sweepFromBackgroundThread(AutoLockHelperThreadState& lock) {
 
 void GCRuntime::waitBackgroundSweepEnd() {
   sweepTask.join();
+  for (AllZonesIter zone(this); !zone.done(); zone.next()) {
+    if (zone->isGCFinished()) {
+      zone->arenas.mergeBackgroundSweptArenas();
+    }
+  }
   if (state() != State::Sweep) {
     assertBackgroundSweepingFinished();
   }
@@ -501,9 +588,6 @@ void GCRuntime::freeFromBackgroundThread(AutoLockHelperThreadState& lock) {
     Nursery::StringBufferVector stringBuffers;
     std::swap(stringBuffers, stringBuffersToReleaseAfterMinorGC.ref());
 
-    SlimLinkedList<LargeBuffer> largeBuffers;
-    std::swap(largeBuffers, largeBuffersToFreeAfterMinorGC.ref());
-
     AutoUnlockHelperThreadState unlock(lock);
 
     lifoBlocks.freeAll();
@@ -519,8 +603,6 @@ void GCRuntime::freeFromBackgroundThread(AutoLockHelperThreadState& lock) {
     for (auto* buffer : stringBuffers) {
       buffer->Release();
     }
-
-    BufferAllocator::FreeLargeAllocs(largeBuffers);
   } while (hasBuffersForBackgroundFree());
 }
 
@@ -1269,20 +1351,14 @@ class ImmediateSweepWeakCacheTask : public GCParallelTask {
 };
 
 void GCRuntime::updateAtomsBitmap() {
-  size_t collectedZones = 0;
-  size_t uncollectedZones = 0;
-  for (ZonesIter zone(this, SkipAtoms); !zone.done(); zone.next()) {
-    if (zone->isCollecting()) {
-      collectedZones++;
-    } else {
-      uncollectedZones++;
-    }
+  atomMarking.refineZoneBitmapsForCollectedZones(this);
+
+  // Mark atoms used by uncollected zones after refining the atoms bitmaps.
+  auto& atomsToMark = atomsUsedByUncollectedZones.ref();
+  if (atomsToMark) {
+    atomMarking.markAtomsUsedByUncollectedZones(this, std::move(atomsToMark));
   }
 
-  atomMarking.refineZoneBitmapsForCollectedZones(this, collectedZones);
-
-  atomMarking.markAtomsUsedByUncollectedZones(this, uncollectedZones);
-
   // For convenience sweep these tables non-incrementally as part of bitmap
   // sweeping; they are likely to be much smaller than the main atoms table.
   SweepingTracer trc(rt);
@@ -1675,11 +1751,10 @@ IncrementalProgress GCRuntime::beginSweepingSweepGroup(JS::GCContext* gcx,
   // or on the background thread.
 
   for (SweepGroupZonesIter zone(this); !zone.done(); zone.next()) {
-    for (const auto& phase : BackgroundFinalizePhases) {
-      initBackgroundSweep(zone, gcx, phase);
-    }
-
     zone->arenas.queueForegroundThingsForSweep();
+    constexpr AllocKinds backgroundKinds =
+        BackgroundObjectFinalizePhase + BackgroundTrivialFinalizePhase;
+    initBackgroundSweep(zone, gcx, backgroundKinds);
   }
 
   MOZ_ASSERT(!sweepZone);
@@ -1817,25 +1892,24 @@ void GCRuntime::beginSweepPhase(JS::GCReason reason, AutoGCSession& session) {
   sweepActions->assertFinished();
 }
 
-bool GCRuntime::foregroundFinalize(JS::GCContext* gcx, Zone* zone,
-                                   AllocKind thingKind,
-                                   SliceBudget& sliceBudget,
-                                   SortedArenaList& sweepList) {
-  ArenaLists& lists = zone->arenas;
-  lists.checkNoArenasToUpdateForKind(thingKind);
+bool ArenaLists::foregroundFinalize(JS::GCContext* gcx, AllocKind thingKind,
+                                    SliceBudget& sliceBudget,
+                                    SortedArenaList& sweepList) {
+  checkNoArenasToUpdateForKind(thingKind);
 
   // Non-empty arenas are reused for use for new allocations as soon as the
   // finalizers for that allocation kind have run. Empty arenas are only
   // released when everything in the zone has been swept (see
   // GCRuntime::sweepBackgroundThings for more details).
-  if (!FinalizeArenas(gcx, lists.collectingArenaList(thingKind), sweepList,
-                      thingKind, sliceBudget)) {
+  ArenaList& arenas = collectingArenaList(thingKind);
+  if (!FinalizeArenas<ReleaseEmpty::No>(gcx, arenas, sweepList, thingKind,
+                                        sliceBudget)) {
     return false;
   }
 
-  sweepList.extractEmptyTo(&lists.savedEmptyArenas.ref());
-  lists.mergeFinalizedArenas(thingKind, sweepList);
-
+  sweepList.extractEmptyTo(&savedEmptyArenas.ref());
+  ArenaList sweptArenas = sweepList.convertToArenaList();
+  mergeSweptArenas(thingKind, sweptArenas);
   return true;
 }
 
@@ -2042,8 +2116,9 @@ IncrementalProgress GCRuntime::finalizeAllocKind(JS::GCContext* gcx,
   }
 
   AutoSetThreadIsFinalizing threadIsFinalizing(gcx);
-  if (!foregroundFinalize(gcx, sweepZone, sweepAllocKind, budget,
-                          finalizedArenas.ref())) {
+  ArenaLists& arenaLists = sweepZone->arenas;
+  if (!arenaLists.foregroundFinalize(gcx, sweepAllocKind, budget,
+                                     finalizedArenas.ref())) {
     return NotFinished;
   }
 
@@ -2349,7 +2424,7 @@ static UniquePtr<SweepAction> ForEachZoneInSweepGroup(
   return js::MakeUnique<Action>(rt, zoneOut, std::move(action));
 }
 
-static UniquePtr<SweepAction> ForEachAllocKind(AllocKinds kinds,
+static UniquePtr<SweepAction> ForEachAllocKind(const AllocKinds& kinds,
                                                AllocKind* kindOut,
                                                UniquePtr<SweepAction> action) {
   if (!action) {
@@ -2382,11 +2457,11 @@ bool GCRuntime::initSweepActions() {
           ForEachZoneInSweepGroup(
               rt, &sweepZone.ref(),
               Sequence(MaybeYield(ZealMode::YieldBeforeSweepingObjects),
-                       ForEachAllocKind(ForegroundObjectFinalizePhase.kinds,
+                       ForEachAllocKind(ForegroundObjectFinalizePhase,
                                         &sweepAllocKind.ref(),
                                         Call(&GCRuntime::finalizeAllocKind)),
                        MaybeYield(ZealMode::YieldBeforeSweepingNonObjects),
-                       ForEachAllocKind(ForegroundNonObjectFinalizePhase.kinds,
+                       ForEachAllocKind(ForegroundNonObjectFinalizePhase,
                                         &sweepAllocKind.ref(),
                                         Call(&GCRuntime::finalizeAllocKind)),
                        MaybeYield(ZealMode::YieldBeforeSweepingPropMapTrees),
@@ -2524,14 +2599,5 @@ void GCRuntime::endSweepPhase(bool destroyingRuntime) {
   finishMarkingValidation();
 #endif
 
-#ifdef DEBUG
-  for (ZonesIter zone(this, WithAtoms); !zone.done(); zone.next()) {
-    for (auto i : AllAllocKinds()) {
-      MOZ_ASSERT_IF(!IsBackgroundFinalized(i) || !useBackgroundThreads,
-                    zone->arenas.collectingArenaList(i).isEmpty());
-    }
-  }
-#endif
-
   AssertNoWrappersInGrayList(rt);
 }
