@@ -165,6 +165,7 @@ JS::Zone::Zone(JSRuntime* rt, Kind kind)
       allocNurseryObjects_(true),
       allocNurseryStrings_(true),
       allocNurseryBigInts_(true),
+      allocNurseryGetterSetters_(true),
       pretenuring(this),
       crossZoneStringWrappers_(this),
       shapeZone_(this),
@@ -174,7 +175,8 @@ JS::Zone::Zone(JSRuntime* rt, Kind kind)
       keepPropMapTables_(false),
       wasCollected_(false),
       listNext_(NotOnList),
-      keptObjects(this) {
+      keptObjects(this),
+      objectFuses(rt) {
   /* Ensure that there are no vtables to mess us up here. */
   MOZ_ASSERT(reinterpret_cast<JS::shadow::Zone*>(this) ==
              static_cast<JS::shadow::Zone*>(this));
@@ -602,6 +604,13 @@ Zone* Zone::nextZone() const {
   return listNext_;
 }
 
+void Zone::prepareForMovingGC() {
+  JS::GCContext* gcx = runtimeFromMainThread()->gcContext();
+
+  MOZ_ASSERT(!isPreservingCode());
+  forceDiscardJitCode(gcx);
+}
+
 void Zone::fixupAfterMovingGC() {
   ZoneAllocator::fixupAfterMovingGC();
   shapeZone().fixupPropMapShapeTableAfterMovingGC();
@@ -620,8 +629,8 @@ void Zone::purgeAtomCache() {
 void Zone::addSizeOfIncludingThis(
     mozilla::MallocSizeOf mallocSizeOf, size_t* zoneObject, JS::CodeSizes* code,
     size_t* regexpZone, size_t* jitZone, size_t* cacheIRStubs,
-    size_t* uniqueIdMap, size_t* initialPropMapTable, size_t* shapeTables,
-    size_t* atomsMarkBitmaps, size_t* compartmentObjects,
+    size_t* objectFusesArg, size_t* uniqueIdMap, size_t* initialPropMapTable,
+    size_t* shapeTables, size_t* atomsMarkBitmaps, size_t* compartmentObjects,
     size_t* crossCompartmentWrappersTables, size_t* compartmentsPrivateData,
     size_t* scriptCountsMapArg) {
   *zoneObject += mallocSizeOf(this);
@@ -629,6 +638,7 @@ void Zone::addSizeOfIncludingThis(
   if (jitZone_) {
     jitZone_->addSizeOfIncludingThis(mallocSizeOf, code, jitZone, cacheIRStubs);
   }
+  *objectFusesArg += objectFuses.sizeOfExcludingThis(mallocSizeOf);
   *uniqueIdMap += uniqueIds().shallowSizeOfExcludingThis(mallocSizeOf);
   shapeZone().addSizeOfExcludingThis(mallocSizeOf, initialPropMapTable,
                                      shapeTables);
