@@ -16,6 +16,8 @@
 
 namespace js {
 
+class NamedLambdaObject;
+
 namespace jit {
 
 class BaselineSnapshot;
@@ -87,6 +89,8 @@ class BaselineCodeGen {
   // stored in the script) as argument for a VM function.
   void loadScriptGCThing(ScriptGCThingType type, Register dest,
                          Register scratch);
+  void loadScriptGCThingInternal(ScriptGCThingType type, Register dest,
+                                 Register scratch);
   void pushScriptGCThingArg(ScriptGCThingType type, Register scratch1,
                             Register scratch2);
   void pushScriptNameArg(Register scratch1, Register scratch2);
@@ -100,6 +104,8 @@ class BaselineCodeGen {
 
   // Loads the current JSScript* in dest.
   void loadScript(Register dest);
+  // Loads the current JitScript* in dest
+  void loadJitScript(Register dest);
 
   void saveInterpreterPCReg();
   void restoreInterpreterPCReg();
@@ -107,6 +113,9 @@ class BaselineCodeGen {
   // Subtracts |script->nslots() * sizeof(Value)| from reg.
   void subtractScriptSlotsSize(Register reg, Register scratch);
 
+  // Loads the resume entries of the current BaselineScript* in dest
+  void loadBaselineScriptResumeEntries(Register dest, Register scratch);
+
   // Jump to the script's resume entry indicated by resumeIndex.
   void jumpToResumeEntry(Register resumeIndex, Register scratch1,
                          Register scratch2);
@@ -151,9 +160,8 @@ class BaselineCodeGen {
 
   bool emitSuspend(JSOp op);
 
-  template <typename F>
-  [[nodiscard]] bool emitAfterYieldDebugInstrumentation(const F& ifDebuggee,
-                                                        Register scratch);
+  [[nodiscard]] bool emitAfterYieldDebugInstrumentation(Register scratch);
+  [[nodiscard]] bool emitDebugAfterYield();
 
   // ifSet should be a function emitting code for when the script has |flag|
   // set. ifNotSet emits code for when the flag isn't set.
@@ -201,6 +209,8 @@ class BaselineCodeGen {
   // Handles JSOp::Lt, JSOp::Gt, and friends
   [[nodiscard]] bool emitCompare();
 
+  [[nodiscard]] bool emitConstantStrictEq(JSOp op);
+
   // Handles JSOp::NewObject and JSOp::NewInit.
   [[nodiscard]] bool emitNewObject();
 
@@ -295,9 +305,14 @@ class BaselineCompilerHandler {
   JSScript* script_;
   jsbytecode* pc_;
 
+  size_t nargs_;
+
   JSObject* globalLexicalEnvironment_;
   JSObject* globalThis_;
 
+  CallObject* callObjectTemplate_;
+  NamedLambdaObject* namedLambdaTemplate_;
+
   // Index of the current ICEntry in the script's JitScript.
   uint32_t icEntryIndex_;
 
@@ -308,6 +323,8 @@ class BaselineCompilerHandler {
 
   bool compilingOffThread_ = false;
 
+  bool needsEnvAllocSite_ = false;
+
  public:
   using FrameInfoT = CompilerFrameInfo;
 
@@ -336,8 +353,16 @@ class BaselineCompilerHandler {
   JSScript* script() const { return script_; }
   JSScript* maybeScript() const { return script_; }
 
-  JSFunction* function() const { return script_->function(); }
-  JSFunction* maybeFunction() const { return function(); }
+  size_t nargs() const {
+    MOZ_ASSERT(isFunction());
+    return nargs_;
+  }
+  CallObject* callObjectTemplate() const { return callObjectTemplate_; }
+  NamedLambdaObject* namedLambdaTemplate() const {
+    return namedLambdaTemplate_;
+  }
+
+  bool isFunction() const { return !!script_->function(); }
 
   ModuleObject* module() const { return script_->module(); }
 
@@ -368,7 +393,7 @@ class BaselineCompilerHandler {
 
   bool canHaveFixedSlots() const { return script()->nfixed() != 0; }
 
-  JSObject* globalLexicalEnvironment() const {
+  JSObject* maybeGlobalLexicalEnvironment() const {
     return globalLexicalEnvironment_;
   }
   JSObject* globalThis() const { return globalThis_; }
@@ -384,6 +409,16 @@ class BaselineCompilerHandler {
 
   bool compilingOffThread() const { return compilingOffThread_; }
   void setCompilingOffThread() { compilingOffThread_ = true; }
+
+  bool addEnvAllocSite() {
+    needsEnvAllocSite_ = true;
+    return true;
+  }
+
+  bool realmIndependentJitcode() const {
+    return JS::Prefs::experimental_self_hosted_cache() &&
+           script()->selfHosted();
+  }
 };
 
 using BaselineCompilerCodeGen = BaselineCodeGen<BaselineCompilerHandler>;
@@ -487,7 +522,6 @@ class BaselineInterpreterHandler {
   jsbytecode* maybePC() const { return nullptr; }
   bool isDefinitelyLastOp() const { return false; }
   JSScript* maybeScript() const { return nullptr; }
-  JSFunction* maybeFunction() const { return nullptr; }
 
   bool shouldEmitDebugEpilogueAtReturnOp() const {
     // The interpreter doesn't use the return address -> pc mapping and doesn't
@@ -512,6 +546,11 @@ class BaselineInterpreterHandler {
   bool mustIncludeSlotsInStackCheck() const { return true; }
 
   bool canHaveFixedSlots() const { return true; }
+  JSObject* maybeGlobalLexicalEnvironment() const { return nullptr; }
+
+  bool addEnvAllocSite() { return false; }  // Not supported.
+
+  bool realmIndependentJitcode() const { return true; }
 };
 
 using BaselineInterpreterCodeGen = BaselineCodeGen<BaselineInterpreterHandler>;
