@@ -86,7 +86,8 @@ class CodeGenerator final : public CodeGeneratorSpecific {
 
  public:
   CodeGenerator(MIRGenerator* gen, LIRGraph* graph,
-                MacroAssembler* masm = nullptr);
+                MacroAssembler* masm = nullptr,
+                const wasm::CodeMetadata* wasmCodeMeta = nullptr);
   ~CodeGenerator();
 
   [[nodiscard]] bool generate(const WarpSnapshot* snapshot);
@@ -159,7 +160,7 @@ class CodeGenerator final : public CodeGeneratorSpecific {
 
   void emitElementPostWriteBarrier(MInstruction* mir,
                                    const LiveRegisterSet& liveVolatileRegs,
-                                   Register obj, const LAllocation* index,
+                                   Register obj, Register index,
                                    Register scratch,
                                    const ConstantOrRegister& val,
                                    int32_t indexDiff = 0);
@@ -227,6 +228,18 @@ class CodeGenerator final : public CodeGeneratorSpecific {
 
   static RegisterOrInt32 ToRegisterOrInt32(const LAllocation* allocation);
 
+  using AddressOrBaseIndex = mozilla::Variant<Address, BaseIndex>;
+
+  static AddressOrBaseIndex ToAddressOrBaseIndex(Register elements,
+                                                 const LAllocation* index,
+                                                 Scalar::Type type);
+
+  using AddressOrBaseObjectElementIndex =
+      mozilla::Variant<Address, BaseObjectElementIndex>;
+
+  static AddressOrBaseObjectElementIndex ToAddressOrBaseObjectElementIndex(
+      Register elements, const LAllocation* index);
+
 #ifdef DEBUG
   void emitAssertArgumentsSliceBounds(const RegisterOrInt32& begin,
                                       const RegisterOrInt32& count,
@@ -293,14 +306,14 @@ class CodeGenerator final : public CodeGeneratorSpecific {
 
   IonScriptCounts* maybeCreateScriptCounts();
 
-  template <typename InstructionWithMaybeTrapSite, class AddressOrBaseIndex>
+  template <typename InstructionWithMaybeTrapSite, class AddressOrBaseIndexT>
   void emitWasmValueLoad(InstructionWithMaybeTrapSite* ins, MIRType type,
-                         MWideningOp wideningOp, AddressOrBaseIndex addr,
+                         MWideningOp wideningOp, AddressOrBaseIndexT addr,
                          AnyRegister dst);
-  template <typename InstructionWithMaybeTrapSite, class AddressOrBaseIndex>
+  template <typename InstructionWithMaybeTrapSite, class AddressOrBaseIndexT>
   void emitWasmValueStore(InstructionWithMaybeTrapSite* ins, MIRType type,
                           MNarrowingOp narrowingOp, AnyRegister src,
-                          AddressOrBaseIndex addr);
+                          AddressOrBaseIndexT addr);
 
   void testValueTruthyForType(JSValueType type, ScratchTagScope& tag,
                               const ValueOperand& value, Register tempToUnbox,
@@ -346,12 +359,9 @@ class CodeGenerator final : public CodeGeneratorSpecific {
                                    Label* ifDoesntEmulateUndefined,
                                    Register scratch, OutOfLineTestObject* ool);
 
-  void emitStoreElementTyped(const LAllocation* value, MIRType valueType,
-                             Register elements, const LAllocation* index);
-
   // Bailout if an element about to be written to is a hole.
-  void emitStoreHoleCheck(Register elements, const LAllocation* index,
-                          LSnapshot* snapshot);
+  void emitStoreHoleCheck(Address dest, LSnapshot* snapshot);
+  void emitStoreHoleCheck(BaseObjectElementIndex dest, LSnapshot* snapshot);
 
   void emitAssertRangeI(MIRType type, const Range* r, Register input);
   void emitAssertRangeD(const Range* r, FloatRegister input,
@@ -373,7 +383,21 @@ class CodeGenerator final : public CodeGeneratorSpecific {
     NurseryObjectLabel(CodeOffset offset, uint32_t nurseryIndex)
         : offset(offset), nurseryIndex(nurseryIndex) {}
   };
-  Vector<NurseryObjectLabel, 0, JitAllocPolicy> ionNurseryObjectLabels_;
+  Vector<NurseryObjectLabel, 0, JitAllocPolicy> nurseryObjectLabels_;
+
+  // Like NurseryObjectLabel but for Values. The Values in the IonScript will be
+  // stored in the constants-list so this also stores the constantPoolIndex for
+  // that list.
+  struct NurseryValueLabel {
+    CodeOffset offset;
+    uint32_t nurseryIndex;
+    uint32_t constantPoolIndex = UINT32_MAX;
+    NurseryValueLabel(CodeOffset offset, uint32_t nurseryIndex)
+        : offset(offset), nurseryIndex(nurseryIndex) {}
+  };
+  Vector<NurseryValueLabel, 0, JitAllocPolicy> nurseryValueLabels_;
+
+  Address getNurseryValueAddress(ValueOrNurseryValueIndex val, Register reg);
 
   void branchIfInvalidated(Register temp, Label* invalidated);
 
@@ -381,6 +405,7 @@ class CodeGenerator final : public CodeGeneratorSpecific {
   void emitDebugResultChecks(LInstruction* ins);
   void emitGCThingResultChecks(LInstruction* lir, MDefinition* mir);
   void emitValueResultChecks(LInstruction* lir, MDefinition* mir);
+  void emitWasmAnyrefResultChecks(LInstruction* lir, MDefinition* mir);
 #endif
 
   // Script counts created during code generation.
@@ -409,6 +434,9 @@ class CodeGenerator final : public CodeGeneratorSpecific {
   // Register a dependency on the HasSeenObjectEmulateUndefined fuse.
   bool addHasSeenObjectEmulateUndefinedFuseDependency();
 
+  // Register a dependency on the HasSeenArrayExceedsInt32Length fuse.
+  bool addHasSeenArrayExceedsInt32LengthFuseDependency();
+
   // Return true if the fuse is intact, and if the fuse is intact note the
   // dependency
   bool hasSeenObjectEmulateUndefinedFuseIntactAndDependencyNoted() {
@@ -420,6 +448,14 @@ class CodeGenerator final : public CodeGeneratorSpecific {
     }
     return false;
   }
+
+  bool hasSeenArrayExceedsInt32LengthFuseIntactAndDependencyNoted() {
+    bool intact = gen->outerInfo().hasSeenArrayExceedsInt32LengthFuseIntact();
+    if (intact) {
+      return addHasSeenArrayExceedsInt32LengthFuseDependency();
+    }
+    return false;
+  }
 };
 
 }  // namespace jit
