@@ -50,7 +50,9 @@
 #include "jit/RangeAnalysis.h"
 #include "jit/ScalarReplacement.h"
 #include "jit/ScriptFromCalleeToken.h"
+#include "jit/SimpleAllocator.h"
 #include "jit/Sink.h"
+#include "jit/UnrollLoops.h"
 #include "jit/ValueNumbering.h"
 #include "jit/WarpBuilder.h"
 #include "jit/WarpOracle.h"
@@ -416,8 +418,6 @@ uint8_t* jit::LazyLinkTopActivation(JSContext* cx,
 
 /* static */
 void JitRuntime::TraceAtomZoneRoots(JSTracer* trc) {
-  MOZ_ASSERT(!JS::RuntimeHeapIsMinorCollecting());
-
   // Shared stubs are allocated in the atoms zone, so do not iterate
   // them after the atoms heap after it has been "finished."
   if (trc->runtime()->atomsAreFinished()) {
@@ -450,27 +450,27 @@ void JitRuntime::TraceWeakJitcodeGlobalTable(JSRuntime* rt, JSTracer* trc) {
   }
 }
 
-bool JitZone::addInlinedCompilation(const RecompileInfo& info,
+bool JitZone::addInlinedCompilation(const IonScriptKey& ionScriptKey,
                                     JSScript* inlined) {
-  MOZ_ASSERT(inlined != info.script());
+  MOZ_ASSERT(inlined != ionScriptKey.script());
 
   auto p = inlinedCompilations_.lookupForAdd(inlined);
   if (p) {
     auto& compilations = p->value();
-    if (!compilations.empty() && compilations.back() == info) {
+    if (!compilations.empty() && compilations.back() == ionScriptKey) {
       return true;
     }
-    return compilations.append(info);
+    return compilations.append(ionScriptKey);
   }
 
-  RecompileInfoVector compilations;
-  if (!compilations.append(info)) {
+  IonScriptKeyVector compilations;
+  if (!compilations.append(ionScriptKey)) {
     return false;
   }
   return inlinedCompilations_.add(p, inlined, std::move(compilations));
 }
 
-void jit::AddPendingInvalidation(RecompileInfoVector& invalid,
+void jit::AddPendingInvalidation(IonScriptKeyVector& invalid,
                                  JSScript* script) {
   MOZ_ASSERT(script);
 
@@ -484,40 +484,46 @@ void jit::AddPendingInvalidation(RecompileInfoVector& invalid,
     return;
   }
 
-  auto addPendingInvalidation = [&invalid](const RecompileInfo& info) {
+  auto addPendingInvalidation = [&invalid](const IonScriptKey& ionScriptKey) {
     AutoEnterOOMUnsafeRegion oomUnsafe;
-    if (!invalid.append(info)) {
+    if (!invalid.append(ionScriptKey)) {
       // BUG 1536159: For diagnostics, compute the size of the failed
       // allocation. This presumes the vector growth strategy is to double. This
       // is only used for crash reporting so not a problem if we get it wrong.
-      size_t allocSize = 2 * sizeof(RecompileInfo) * invalid.capacity();
-      oomUnsafe.crash(allocSize, "Could not update RecompileInfoVector");
+      size_t allocSize = 2 * sizeof(IonScriptKey) * invalid.capacity();
+      oomUnsafe.crash(allocSize, "Could not update IonScriptKeyVector");
     }
   };
 
   // Trigger invalidation of the IonScript.
   if (jitScript->hasIonScript()) {
-    RecompileInfo info(script, jitScript->ionScript()->compilationId());
-    addPendingInvalidation(info);
+    IonScriptKey ionScriptKey(script, jitScript->ionScript()->compilationId());
+    addPendingInvalidation(ionScriptKey);
   }
 
   // Trigger invalidation of any callers inlining this script.
   auto* inlinedCompilations =
       script->zone()->jitZone()->maybeInlinedCompilations(script);
   if (inlinedCompilations) {
-    for (const RecompileInfo& info : *inlinedCompilations) {
-      addPendingInvalidation(info);
+    for (const auto& ionScriptKey : *inlinedCompilations) {
+      addPendingInvalidation(ionScriptKey);
     }
     script->zone()->jitZone()->removeInlinedCompilations(script);
   }
 }
 
-IonScript* RecompileInfo::maybeIonScriptToInvalidate() const {
+IonScript* IonScriptKey::maybeIonScriptToInvalidate() const {
+  // This must be called either on the main thread or when sweeping WeakCaches
+  // off-thread.
+  MOZ_ASSERT(CurrentThreadIsMainThread() || CurrentThreadIsGCSweeping());
+
+#ifdef DEBUG
   // Make sure this is not called under CodeGenerator::link (before the
-  // IonScript is created).
-  MOZ_ASSERT_IF(
-      script_->zone()->jitZone()->currentCompilationId(),
-      script_->zone()->jitZone()->currentCompilationId().ref() != id_);
+  // corresponding IonScript is created).
+  auto* jitZone = script_->zoneFromAnyThread()->jitZone();
+  MOZ_ASSERT_IF(jitZone->currentCompilationId(),
+                jitZone->currentCompilationId().ref() != id_);
+#endif
 
   if (!script_->hasIonScript() ||
       script_->ionScript()->compilationId() != id_) {
@@ -527,11 +533,11 @@ IonScript* RecompileInfo::maybeIonScriptToInvalidate() const {
   return script_->ionScript();
 }
 
-bool RecompileInfo::traceWeak(JSTracer* trc) {
-  // Sweep the RecompileInfo if either the script is dead or the IonScript has
+bool IonScriptKey::traceWeak(JSTracer* trc) {
+  // Sweep the IonScriptKey if either the script is dead or the IonScript has
   // been invalidated.
 
-  if (!TraceManuallyBarrieredWeakEdge(trc, &script_, "RecompileInfo::script")) {
+  if (!TraceManuallyBarrieredWeakEdge(trc, &script_, "IonScriptKey::script")) {
     return false;
   }
 
@@ -694,7 +700,7 @@ IonScript* IonScript::New(JSContext* cx, IonCompilationId compilationId,
                 "IonScript has wrong size for SafepointIndex");
 
   CheckedInt<Offset> allocSize = sizeof(IonScript);
-  allocSize += CheckedInt<Offset>(constants) * sizeof(Value);
+  allocSize += CheckedInt<Offset>(constants) * sizeof(HeapPtr<Value>);
   allocSize += CheckedInt<Offset>(runtimeSize);
   allocSize += CheckedInt<Offset>(nurseryObjects) * sizeof(HeapPtr<JSObject*>);
   allocSize += CheckedInt<Offset>(osiIndices) * sizeof(OsiIndex);
@@ -720,9 +726,10 @@ IonScript* IonScript::New(JSContext* cx, IonCompilationId compilationId,
 
   Offset offsetCursor = sizeof(IonScript);
 
-  MOZ_ASSERT(offsetCursor % alignof(Value) == 0);
+  MOZ_ASSERT(offsetCursor % alignof(HeapPtr<Value>) == 0);
+  script->initElements<HeapPtr<Value>>(offsetCursor, constants);
   script->constantTableOffset_ = offsetCursor;
-  offsetCursor += constants * sizeof(Value);
+  offsetCursor += constants * sizeof(HeapPtr<Value>);
 
   MOZ_ASSERT(offsetCursor % alignof(uint64_t) == 0);
   script->runtimeDataOffset_ = offsetCursor;
@@ -928,20 +935,24 @@ const OsiIndex* IonScript::getOsiIndex(uint8_t* retAddr) const {
 }
 
 void IonScript::Destroy(JS::GCContext* gcx, IonScript* script) {
-  // Make sure there are no pointers into the IonScript's nursery objects list
-  // in the store buffer. Because this can be called during sweeping when
-  // discarding JIT code, we have to lock the store buffer when we find an
-  // object that's (still) in the nursery.
+  // Destroy the HeapPtrs to ensure there are no pointers into the IonScript's
+  // nursery objects list or constants list in the store buffer. Because this
+  // can be called during sweeping when discarding JIT code, we have to lock the
+  // store buffer when we find a pointer that's (still) in the nursery.
   mozilla::Maybe<gc::AutoLockStoreBuffer> lock;
   for (size_t i = 0, len = script->numNurseryObjects(); i < len; i++) {
     JSObject* obj = script->nurseryObjects()[i];
-    if (!IsInsideNursery(obj)) {
-      continue;
-    }
-    if (lock.isNothing()) {
+    if (lock.isNothing() && IsInsideNursery(obj)) {
       lock.emplace(gcx->runtimeFromAnyThread());
     }
-    script->nurseryObjects()[i] = HeapPtr<JSObject*>();
+    script->nurseryObjects()[i].~HeapPtr<JSObject*>();
+  }
+  for (size_t i = 0, len = script->numConstants(); i < len; i++) {
+    Value v = script->getConstant(i);
+    if (lock.isNothing() && v.isGCThing() && IsInsideNursery(v.toGCThing())) {
+      lock.emplace(gcx->runtimeFromAnyThread());
+    }
+    script->getConstant(i).~HeapPtr<Value>();
   }
 
   // This allocation is tracked by JSScript::setIonScriptImpl.
@@ -993,7 +1004,8 @@ bool OptimizeMIR(MIRGenerator* mir) {
   }
 
   {
-    if (!FoldEmptyBlocks(graph)) {
+    bool dummy;
+    if (!FoldEmptyBlocks(graph, &dummy)) {
       return false;
     }
     gs.spewPass("Fold Empty Blocks");
@@ -1053,7 +1065,7 @@ bool OptimizeMIR(MIRGenerator* mir) {
   }
 
   {
-    if (!BuildDominatorTree(graph)) {
+    if (!BuildDominatorTree(mir, graph)) {
       return false;
     }
     // No spew: graph not changed.
@@ -1130,6 +1142,18 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
   }
 
+  if (mir->compilingWasm()) {
+    if (!TrackWasmRefTypes(graph)) {
+      return false;
+    }
+    gs.spewPass("Track Wasm ref types");
+    AssertExtendedGraphCoherency(graph);
+
+    if (mir->shouldCancel("Track Wasm ref types")) {
+      return false;
+    }
+  }
+
   if (mir->optimizationInfo().amaEnabled()) {
     AlignmentMaskAnalysis ama(graph);
     if (!ama.analyze()) {
@@ -1336,7 +1360,8 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
   }
 
-  if (mir->optimizationInfo().eaaEnabled()) {
+  // EAA, but only for wasm; it appears to be of minimal benefit for JS inputs.
+  if (mir->compilingWasm() && mir->optimizationInfo().eaaEnabled()) {
     EffectiveAddressAnalysis eaa(mir, graph);
     JitSpewCont(JitSpew_EAA, "\n");
     if (!eaa.analyze()) {
@@ -1388,7 +1413,7 @@ bool OptimizeMIR(MIRGenerator* mir) {
 
   if (mir->optimizationInfo().instructionReorderingEnabled() &&
       !mir->outerInfo().hadReorderingBailout()) {
-    if (!ReorderInstructions(graph)) {
+    if (!ReorderInstructions(mir, graph)) {
       return false;
     }
     gs.spewPass("Reordering");
@@ -1416,6 +1441,48 @@ bool OptimizeMIR(MIRGenerator* mir) {
   AssertExtendedGraphCoherency(graph, /* underValueNumberer = */ false,
                                /* force = */ true);
 
+  // Unroll and/or peel loops
+  if (mir->compilingWasm() && JS::Prefs::wasm_unroll_loops()) {
+    bool loopsChanged;
+    if (!UnrollLoops(mir, graph, &loopsChanged)) {
+      return false;
+    }
+
+    gs.spewPass("Unroll loops");
+
+    AssertExtendedGraphCoherency(graph);
+
+    if (mir->shouldCancel("Unroll loops")) {
+      return false;
+    }
+
+    if (loopsChanged) {
+      // Rerun GVN in the hope that unrolling exposed more optimization
+      // opportunities.
+      if (!gvn.run(ValueNumberer::DontUpdateAliasAnalysis)) {
+        return false;
+      }
+      // And tidy up any empty blocks.
+      bool blocksFolded;
+      if (!FoldEmptyBlocks(graph, &blocksFolded)) {
+        return false;
+      }
+      if (blocksFolded) {
+        // Redo the dominator tree.
+        ClearDominatorTree(graph);
+        if (!BuildDominatorTree(mir, graph)) {
+          return false;
+        }
+      }
+
+      AssertExtendedGraphCoherency(graph);
+
+      if (mir->shouldCancel("Rerun GVN after loop unrolling")) {
+        return false;
+      }
+    }
+  }
+
   // Remove unreachable blocks created by MBasicBlock::NewFakeLoopPredecessor
   // to ensure every loop header has two predecessors. (This only happens due
   // to OSR.)  After this point, it is no longer possible to build the
@@ -1456,6 +1523,10 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
     gs.spewPass("Bounds Check Elimination");
     AssertGraphCoherency(graph);
+
+    if (mir->shouldCancel("Bounds Check Elimination")) {
+      return false;
+    }
   }
 
   if (mir->optimizationInfo().eliminateRedundantShapeGuardsEnabled()) {
@@ -1464,6 +1535,10 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
     gs.spewPass("Shape Guard Elimination");
     AssertGraphCoherency(graph);
+
+    if (mir->shouldCancel("Shape Guard Elimination")) {
+      return false;
+    }
   }
 
   // Run the GC Barrier Elimination pass after instruction reordering, to
@@ -1475,6 +1550,10 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
     gs.spewPass("GC Barrier Elimination");
     AssertGraphCoherency(graph);
+
+    if (mir->shouldCancel("GC Barrier Elimination")) {
+      return false;
+    }
   }
 
   if (!mir->compilingWasm() && !mir->outerInfo().hadUnboxFoldingBailout()) {
@@ -1483,6 +1562,10 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
     gs.spewPass("FoldLoadsWithUnbox");
     AssertGraphCoherency(graph);
+
+    if (mir->shouldCancel("FoldLoadsWithUnbox")) {
+      return false;
+    }
   }
 
   if (!mir->compilingWasm()) {
@@ -1491,6 +1574,10 @@ bool OptimizeMIR(MIRGenerator* mir) {
     }
     gs.spewPass("Add KeepAlive Instructions");
     AssertGraphCoherency(graph);
+
+    if (mir->shouldCancel("Add KeepAlive Instructions")) {
+      return false;
+    }
   }
 
   AssertGraphCoherency(graph, /* force = */ true);
@@ -1527,49 +1614,46 @@ LIRGraph* GenerateLIR(MIRGenerator* mir) {
 
 #ifdef DEBUG
   AllocationIntegrityState integrity(*lir);
-#endif
-
-  {
-    IonRegisterAllocator allocator =
-        mir->optimizationInfo().registerAllocator();
-
-    switch (allocator) {
-      case RegisterAllocator_Backtracking:
-      case RegisterAllocator_Testbed: {
-#ifdef DEBUG
-        if (JitOptions.fullDebugChecks) {
-          if (!integrity.record()) {
-            return nullptr;
-          }
-        }
-#endif
-
-        BacktrackingAllocator regalloc(mir, &lirgen, *lir,
-                                       allocator == RegisterAllocator_Testbed);
-        if (!regalloc.go()) {
-          return nullptr;
-        }
-
-#ifdef DEBUG
-        if (JitOptions.fullDebugChecks) {
-          if (!integrity.check()) {
-            return nullptr;
-          }
-        }
-#endif
-
-        gs.spewPass("Allocate Registers [Backtracking]");
-        break;
-      }
-
-      default:
-        MOZ_CRASH("Bad regalloc");
-    }
-
-    if (mir->shouldCancel("Allocate Registers")) {
+  if (JitOptions.fullDebugChecks) {
+    if (!integrity.record()) {
       return nullptr;
     }
   }
+#endif
+
+  IonRegisterAllocator allocator = mir->optimizationInfo().registerAllocator();
+  switch (allocator) {
+    case RegisterAllocator_Backtracking: {
+      BacktrackingAllocator regalloc(mir, &lirgen, *lir);
+      if (!regalloc.go()) {
+        return nullptr;
+      }
+      gs.spewPass("Allocate Registers [Backtracking]");
+      break;
+    }
+    case RegisterAllocator_Simple: {
+      SimpleAllocator regalloc(mir, &lirgen, *lir);
+      if (!regalloc.go()) {
+        return nullptr;
+      }
+      gs.spewPass("Allocate Registers [Simple]");
+      break;
+    }
+    default:
+      MOZ_CRASH("Bad regalloc");
+  }
+
+#ifdef DEBUG
+  if (JitOptions.fullDebugChecks) {
+    if (!integrity.check()) {
+      return nullptr;
+    }
+  }
+#endif
+
+  if (mir->shouldCancel("Allocate Registers")) {
+    return nullptr;
+  }
 
   return lir;
 }
@@ -1706,9 +1790,9 @@ static AbortReason IonCompile(JSContext* cx, HandleScript script,
     return AbortReason::Alloc;
   }
 
-  CompileInfo* info = alloc->new_<CompileInfo>(
-      CompileRuntime::get(cx->runtime()), script, script->function(), osrPc,
-      script->needsArgsObj(), inlineScriptTree);
+  CompileInfo* info =
+      alloc->new_<CompileInfo>(CompileRuntime::get(cx->runtime()), script,
+                               osrPc, script->needsArgsObj(), inlineScriptTree);
   if (!info) {
     return AbortReason::Alloc;
   }
@@ -2298,7 +2382,7 @@ bool jit::IonCompileScriptForBaselineOSR(JSContext* cx, BaselineFrame* frame,
     return false;
   }
 
-  RootedScript script(cx, frame->script());
+  JSScript* script = frame->script();
   if (!script->hasIonScript() || script->ionScript()->osrPc() != pc ||
       frame->isDebuggee()) {
     return true;
@@ -2526,40 +2610,27 @@ static void ClearIonScriptAfterInvalidation(JSContext* cx, JSScript* script,
   }
 }
 
-// Remove this script from pending invalidation script caches.
-//
-// This is done to avoid an invalidation leaving behind dependencies which
-// may not actually be real on a recompilation, causing superflous invalidation
-// of recompiled code.
-//
-// Note: This must be kept in sync with the various WeakScriptCaches.
-static void ClearPendingInvalidationDependencies(JSScript* script) {
-  script->realm()->zone()->fuseDependencies.removeScript(script);
-  script->realm()->realmFuses.fuseDependencies.removeScript(script);
-}
-
-void jit::Invalidate(JSContext* cx, const RecompileInfoVector& invalid,
+void jit::Invalidate(JSContext* cx, const IonScriptKeyVector& invalid,
                      bool resetUses, bool cancelOffThread) {
   JitSpew(JitSpew_IonInvalidate, "Start invalidation.");
 
   // Add an invalidation reference to all invalidated IonScripts to indicate
   // to the traversal which frames have been invalidated.
   size_t numInvalidations = 0;
-  for (const RecompileInfo& info : invalid) {
+  for (const auto& ionScriptKey : invalid) {
+    JSScript* script = ionScriptKey.script();
     if (cancelOffThread) {
-      CancelOffThreadIonCompile(info.script());
+      CancelOffThreadIonCompile(script);
     }
 
-    IonScript* ionScript = info.maybeIonScriptToInvalidate();
+    IonScript* ionScript = ionScriptKey.maybeIonScriptToInvalidate();
     if (!ionScript) {
       continue;
     }
 
     JitSpew(JitSpew_IonInvalidate, " Invalidate %s:%u:%u, IonScript %p",
-            info.script()->filename(), info.script()->lineno(),
-            info.script()->column().oneOriginValue(), ionScript);
-
-    ClearPendingInvalidationDependencies(info.script());
+            script->filename(), script->lineno(),
+            script->column().oneOriginValue(), ionScript);
 
     // Keep the ion script alive during the invalidation and flag this
     // ionScript as being invalidated.  This increment is removed by the
@@ -2581,8 +2652,8 @@ void jit::Invalidate(JSContext* cx, const RecompileInfoVector& invalid,
   // Drop the references added above. If a script was never active, its
   // IonScript will be immediately destroyed. Otherwise, it will be held live
   // until its last invalidated frame is destroyed.
-  for (const RecompileInfo& info : invalid) {
-    IonScript* ionScript = info.maybeIonScriptToInvalidate();
+  for (const auto& ionScriptKey : invalid) {
+    IonScript* ionScript = ionScriptKey.maybeIonScriptToInvalidate();
     if (!ionScript) {
       continue;
     }
@@ -2592,7 +2663,8 @@ void jit::Invalidate(JSContext* cx, const RecompileInfoVector& invalid,
       // jitScript->ionScript_ now. We don't want to do this unconditionally
       // because maybeIonScriptToInvalidate depends on script->ionScript() (we
       // would leak the IonScript if |invalid| contains duplicates).
-      ClearIonScriptAfterInvalidation(cx, info.script(), ionScript, resetUses);
+      ClearIonScriptAfterInvalidation(cx, ionScriptKey.script(), ionScript,
+                                      resetUses);
     }
 
     ionScript->decrementInvalidationCount(gcx);
@@ -2605,9 +2677,10 @@ void jit::Invalidate(JSContext* cx, const RecompileInfoVector& invalid,
 
   // Finally, null out jitScript->ionScript_ for IonScripts that are still on
   // the stack.
-  for (const RecompileInfo& info : invalid) {
-    if (IonScript* ionScript = info.maybeIonScriptToInvalidate()) {
-      ClearIonScriptAfterInvalidation(cx, info.script(), ionScript, resetUses);
+  for (const auto& ionScriptKey : invalid) {
+    if (IonScript* ionScript = ionScriptKey.maybeIonScriptToInvalidate()) {
+      ClearIonScriptAfterInvalidation(cx, ionScriptKey.script(), ionScript,
+                                      resetUses);
     }
   }
 }
@@ -2621,8 +2694,8 @@ void jit::IonScript::invalidate(JSContext* cx, JSScript* script, bool resetUses,
 
   JitSpew(JitSpew_IonInvalidate, " Invalidate IonScript %p: %s", this, reason);
 
-  // RecompileInfoVector has inline space for at least one element.
-  RecompileInfoVector list;
+  // IonScriptKeyVector has inline space for at least one element.
+  IonScriptKeyVector list;
   MOZ_RELEASE_ASSERT(list.reserve(1));
   list.infallibleEmplaceBack(script, compilationId());
 
@@ -2654,8 +2727,8 @@ void jit::Invalidate(JSContext* cx, JSScript* script, bool resetUses,
     }
   }
 
-  // RecompileInfoVector has inline space for at least one element.
-  RecompileInfoVector scripts;
+  // IonScriptKeyVector has inline space for at least one element.
+  IonScriptKeyVector scripts;
   MOZ_ASSERT(script->hasIonScript());
   MOZ_RELEASE_ASSERT(scripts.reserve(1));
   scripts.infallibleEmplaceBack(script, script->ionScript()->compilationId());
