@@ -69,7 +69,7 @@ static_assert((CompilingScript & SpecialScriptBit) != 0);
 static BaselineScript* const BaselineDisabledScriptPtr =
     reinterpret_cast<BaselineScript*>(DisabledScript);
 static BaselineScript* const BaselineQueuedScriptPtr =
-    reinterpret_cast<BaselineScript*>(DisabledScript);
+    reinterpret_cast<BaselineScript*>(QueuedScript);
 static BaselineScript* const BaselineCompilingScriptPtr =
     reinterpret_cast<BaselineScript*>(CompilingScript);
 
@@ -180,6 +180,10 @@ class alignas(uintptr_t) ICScript final : public TrailingArray<ICScript> {
     return numElements<ICEntry>(icEntriesOffset(), fallbackStubsOffset());
   }
 
+  static constexpr Offset offsetOfEnvAllocSite() {
+    return offsetof(ICScript, envAllocSite_);
+  }
+
   ICEntry* interpreterICEntryFromPCOffset(uint32_t pcOffset);
 
   ICEntry& icEntryFromPCOffset(uint32_t pcOffset);
@@ -201,6 +205,9 @@ class alignas(uintptr_t) ICScript final : public TrailingArray<ICScript> {
 
   gc::AllocSite* getOrCreateAllocSite(JSScript* outerScript, uint32_t pcOffset);
 
+  void ensureEnvAllocSite(JSScript* outerScript);
+  gc::AllocSite* maybeEnvAllocSite() const { return envAllocSite_; }
+
   void prepareForDestruction(Zone* zone);
 
   void trace(JSTracer* trc);
@@ -232,6 +239,9 @@ class alignas(uintptr_t) ICScript final : public TrailingArray<ICScript> {
   LifoAlloc allocSitesSpace_{AllocSiteChunkSize, js::BackgroundMallocArena};
   Vector<gc::AllocSite*, 0, SystemAllocPolicy> allocSites_;
 
+  // Optional alloc site to use when allocating environment chain objects.
+  gc::AllocSite* envAllocSite_ = nullptr;
+
   // Number of times this copy of the script has been called or has had
   // backedges taken.  Reset if the script's JIT code is forcibly discarded.
   // See also the ScriptWarmUpData class.
@@ -470,6 +480,8 @@ class alignas(uintptr_t) JitScript final
   void updateLastICStubCounter() { warmUpCountAtLastICStub_ = warmUpCount(); }
   uint32_t warmUpCountAtLastICStub() const { return warmUpCountAtLastICStub_; }
 
+  bool hasEnvAllocSite() const { return icScript_.envAllocSite_; }
+
  private:
   // Methods to set baselineScript_ to a BaselineScript*, nullptr, or
   // BaselineDisabledScriptPtr.
