@@ -13,6 +13,7 @@
 #include "jit/MIR-wasm.h"
 #include "jit/MIR.h"
 #include "jit/MIRGenerator.h"
+#include "wasm/WasmMetadata.h"
 
 using namespace js;
 using namespace js::jit;
@@ -21,11 +22,13 @@ MIRGenerator::MIRGenerator(CompileRealm* realm,
                            const JitCompileOptions& options,
                            TempAllocator* alloc, MIRGraph* graph,
                            const CompileInfo* info,
-                           const OptimizationInfo* optimizationInfo)
+                           const OptimizationInfo* optimizationInfo,
+                           const wasm::CodeMetadata* wasmCodeMeta)
     : realm(realm),
       runtime(realm ? realm->runtime() : nullptr),
       outerInfo_(info),
       optimizationInfo_(optimizationInfo),
+      wasmCodeMeta_(wasmCodeMeta),
       alloc_(alloc),
       graph_(graph),
       offThreadStatus_(Ok()),
@@ -39,9 +42,8 @@ MIRGenerator::MIRGenerator(CompileRealm* realm,
                                    : false),
       bigIntsCanBeInNursery_(realm ? realm->zone()->canNurseryAllocateBigInts()
                                    : false),
-      minWasmMemory0Length_(0),
       options(options),
-      gs_(alloc) {}
+      gs_(alloc, wasmCodeMeta) {}
 
 bool MIRGenerator::licmEnabled() const {
   return optimizationInfo().licmEnabled() && !disableLICM_ &&
@@ -746,7 +748,7 @@ MConstant* MBasicBlock::optimizedOutConstant(TempAllocator& alloc) {
     return ins->toConstant();
   }
 
-  MConstant* constant = MConstant::New(alloc, MagicValue(JS_OPTIMIZED_OUT));
+  MConstant* constant = MConstant::NewMagic(alloc, JS_OPTIMIZED_OUT);
   insertBefore(ins, constant);
   return constant;
 }
