@@ -1239,6 +1239,60 @@ bool RCompare::recover(JSContext* cx, SnapshotIterator& iter) const {
   return true;
 }
 
+bool MStrictConstantCompareInt32::writeRecoverData(
+    CompactBufferWriter& writer) const {
+  MOZ_ASSERT(canRecoverOnBailout());
+  writer.writeUnsigned(
+      uint32_t(RInstruction::Recover_StrictConstantCompareInt32));
+
+  writer.writeByte(uint8_t(jsop_));
+  writer.writeSigned(constant_);
+  return true;
+}
+
+RStrictConstantCompareInt32::RStrictConstantCompareInt32(
+    CompactBufferReader& reader) {
+  jsop_ = JSOp(reader.readByte());
+  constant_ = reader.readSigned();
+  MOZ_ASSERT(IsStrictEqualityOp(jsop_));
+}
+
+bool RStrictConstantCompareInt32::recover(JSContext* cx,
+                                          SnapshotIterator& iter) const {
+  JS::Value lhs = iter.read();
+
+  iter.storeInstructionResult(
+      BooleanValue(lhs.isNumber() && lhs.toNumber() == constant_));
+  return true;
+}
+
+bool MStrictConstantCompareBoolean::writeRecoverData(
+    CompactBufferWriter& writer) const {
+  MOZ_ASSERT(canRecoverOnBailout());
+  writer.writeUnsigned(
+      uint32_t(RInstruction::Recover_StrictConstantCompareBoolean));
+
+  writer.writeByte(uint8_t(jsop_));
+  writer.writeUnsigned(constant_);
+  return true;
+}
+
+RStrictConstantCompareBoolean::RStrictConstantCompareBoolean(
+    CompactBufferReader& reader) {
+  jsop_ = JSOp(reader.readByte());
+  constant_ = reader.readUnsigned();
+  MOZ_ASSERT(IsStrictEqualityOp(jsop_));
+}
+
+bool RStrictConstantCompareBoolean::recover(JSContext* cx,
+                                            SnapshotIterator& iter) const {
+  JS::Value lhs = iter.read();
+
+  iter.storeInstructionResult(
+      BooleanValue(lhs.isBoolean() && lhs.toBoolean() == constant_));
+  return true;
+}
+
 bool MConcat::writeRecoverData(CompactBufferWriter& writer) const {
   MOZ_ASSERT(canRecoverOnBailout());
   writer.writeUnsigned(uint32_t(RInstruction::Recover_Concat));
@@ -2051,7 +2105,7 @@ RNewArray::RNewArray(CompactBufferReader& reader) {
 }
 
 bool RNewArray::recover(JSContext* cx, SnapshotIterator& iter) const {
-  RootedObject templateObject(cx, iter.readObject());
+  JSObject* templateObject = iter.readObject();
   Rooted<Shape*> shape(cx, templateObject->shape());
 
   ArrayObject* resultObject = NewArrayWithShape(cx, count_, shape);
@@ -2075,7 +2129,8 @@ RNewIterator::RNewIterator(CompactBufferReader& reader) {
 }
 
 bool RNewIterator::recover(JSContext* cx, SnapshotIterator& iter) const {
-  RootedObject templateObject(cx, iter.readObject());
+  // Template object is not used when recovering MNewIterator.
+  (void)iter.readObject();
 
   JSObject* resultObject = nullptr;
   switch (MNewIterator::Type(type_)) {
@@ -2151,7 +2206,7 @@ bool MNewCallObject::writeRecoverData(CompactBufferWriter& writer) const {
 RNewCallObject::RNewCallObject(CompactBufferReader& reader) {}
 
 bool RNewCallObject::recover(JSContext* cx, SnapshotIterator& iter) const {
-  Rooted<CallObject*> templateObj(cx, &iter.readObject()->as<CallObject>());
+  CallObject* templateObj = &iter.readObject()->as<CallObject>();
 
   Rooted<SharedShape*> shape(cx, templateObj->sharedShape());
 
@@ -2202,8 +2257,8 @@ RObjectState::RObjectState(CompactBufferReader& reader) {
 }
 
 bool RObjectState::recover(JSContext* cx, SnapshotIterator& iter) const {
-  RootedObject object(cx, iter.readObject());
-  Handle<NativeObject*> nativeObject = object.as<NativeObject>();
+  JSObject* object = iter.readObject();
+  NativeObject* nativeObject = &object->as<NativeObject>();
   MOZ_ASSERT(!Watchtower::watchesPropertyValueChange(nativeObject));
   MOZ_ASSERT(nativeObject->slotSpan() == numSlots());
 
@@ -2440,8 +2495,8 @@ RCreateArgumentsObject::RCreateArgumentsObject(CompactBufferReader& reader) {}
 bool RCreateArgumentsObject::recover(JSContext* cx,
                                      SnapshotIterator& iter) const {
   RootedObject callObject(cx, iter.readObject());
-  RootedObject result(
-      cx, ArgumentsObject::createForIon(cx, iter.frame(), callObject));
+  ArgumentsObject* result =
+      ArgumentsObject::createForIon(cx, iter.frame(), callObject);
   if (!result) {
     return false;
   }
@@ -2514,3 +2569,25 @@ bool RRest::recover(JSContext* cx, SnapshotIterator& iter) const {
   iter.storeInstructionResult(ObjectValue(*rest));
   return true;
 }
+
+bool MTypedArraySubarray::writeRecoverData(CompactBufferWriter& writer) const {
+  MOZ_ASSERT(canRecoverOnBailout());
+  writer.writeUnsigned(uint32_t(RInstruction::Recover_TypedArraySubarray));
+  return true;
+}
+
+RTypedArraySubarray::RTypedArraySubarray(CompactBufferReader& reader) {}
+
+bool RTypedArraySubarray::recover(JSContext* cx, SnapshotIterator& iter) const {
+  Rooted<TypedArrayObject*> obj(cx, &iter.readObject()->as<TypedArrayObject>());
+  intptr_t start = iter.readIntPtr();
+  intptr_t length = iter.readIntPtr();
+
+  auto* result = TypedArraySubarrayRecover(cx, obj, start, length);
+  if (!result) {
+    return false;
+  }
+
+  iter.storeInstructionResult(ObjectValue(*result));
+  return true;
+}
