@@ -45,6 +45,7 @@ class CodeGeneratorShared : public LElementVisitor {
  protected:
   MIRGenerator* gen;
   LIRGraph& graph;
+  const wasm::CodeMetadata* wasmCodeMeta_;
   LBlock* current;
   SnapshotWriter snapshots_;
   RecoverWriter recovers_;
@@ -142,8 +143,7 @@ class CodeGeneratorShared : public LElementVisitor {
   inline Address ToAddress(const LInt64Allocation& a) const;
 
   static inline Address ToAddress(Register elements, const LAllocation* index,
-                                  Scalar::Type type,
-                                  int32_t offsetAdjustment = 0);
+                                  Scalar::Type type);
 
   uint32_t frameSize() const { return frameDepth_; }
 
@@ -154,6 +154,7 @@ class CodeGeneratorShared : public LElementVisitor {
 
  public:
   MIRGenerator& mirGen() const { return *gen; }
+  const wasm::CodeMetadata* wasmCodeMeta() const { return wasmCodeMeta_; }
 
   // When appending to runtimeData_, the vector might realloc, leaving pointers
   // int the origianl vector stale and unusable. DataPtr acts like a pointer,
@@ -241,8 +242,8 @@ class CodeGeneratorShared : public LElementVisitor {
   void emitTruncateDouble(FloatRegister src, Register dest, MInstruction* mir);
   void emitTruncateFloat32(FloatRegister src, Register dest, MInstruction* mir);
 
-  void emitPreBarrier(Register elements, const LAllocation* index);
   void emitPreBarrier(Address address);
+  void emitPreBarrier(BaseObjectElementIndex address);
 
   // We don't emit code for trivial blocks, so if we want to branch to the
   // given block, and it's trivial, return the ultimate block we should
@@ -389,7 +390,8 @@ class CodeGeneratorShared : public LElementVisitor {
   void generateInvalidateEpilogue();
 
  public:
-  CodeGeneratorShared(MIRGenerator* gen, LIRGraph* graph, MacroAssembler* masm);
+  CodeGeneratorShared(MIRGenerator* gen, LIRGraph* graph, MacroAssembler* masm,
+                      const wasm::CodeMetadata* wasmCodeMeta);
 
  public:
   void visitOutOfLineTruncateSlow(OutOfLineTruncateSlow* ool);
