@@ -34,7 +34,7 @@
 #include "builtin/MapObject.h"
 #include "builtin/ShadowRealm.h"
 #include "builtin/Symbol.h"
-#ifdef JS_HAS_TEMPORAL_API
+#ifdef JS_HAS_INTL_API
 #  include "builtin/temporal/Duration.h"
 #  include "builtin/temporal/Instant.h"
 #  include "builtin/temporal/PlainDate.h"
@@ -203,18 +203,9 @@ bool GlobalObject::skipDeselectedConstructor(JSContext* cx, JSProtoKey key) {
     case JSProto_NumberFormat:
     case JSProto_PluralRules:
     case JSProto_RelativeTimeFormat:
-      return false;
-
     case JSProto_Segmenter:
-#  if defined(MOZ_ICU4X)
       return false;
-#  else
-      return true;
-#  endif
 
-#endif
-
-#ifdef JS_HAS_TEMPORAL_API
     case JSProto_Temporal:
     case JSProto_Duration:
     case JSProto_Instant:
@@ -597,6 +588,9 @@ GlobalObject* GlobalObject::createInternal(JSContext* cx,
       ObjectFlag::QualifiedVarObj,
       ObjectFlag::GenerationCountedGlobal,
   };
+  if (JS::Prefs::objectfuse_for_global()) {
+    objectFlags.setFlag(ObjectFlag::HasObjectFuse);
+  }
 
   JSObject* obj =
       NewTenuredObjectWithGivenProto(cx, clasp, nullptr, objectFlags);
@@ -868,8 +862,8 @@ GlobalObject::getOrCreateFinalizationRegistryData() {
 /* static */
 bool GlobalObject::createIntrinsicsHolder(JSContext* cx,
                                           Handle<GlobalObject*> global) {
-  Rooted<NativeObject*> intrinsicsHolder(
-      cx, NewPlainObjectWithProto(cx, nullptr, TenuredObject));
+  NativeObject* intrinsicsHolder =
+      NewPlainObjectWithProto(cx, nullptr, TenuredObject);
   if (!intrinsicsHolder) {
     return false;
   }
@@ -886,7 +880,7 @@ bool GlobalObject::getSelfHostedFunction(JSContext* cx,
                                          Handle<JSAtom*> name, unsigned nargs,
                                          MutableHandleValue funVal) {
   if (global->maybeGetIntrinsicValue(selfHostedName, funVal.address(), cx)) {
-    RootedFunction fun(cx, &funVal.toObject().as<JSFunction>());
+    JSFunction* fun = &funVal.toObject().as<JSFunction>();
     if (fun->fullExplicitName() == name) {
       return true;
     }
@@ -1063,9 +1057,14 @@ void GlobalObjectData::trace(JSTracer* trc, GlobalObject* global) {
 
   TraceNullableEdge(trc, &boundFunctionShapeWithDefaultProto,
                     "global-bound-function-shape");
+  TraceNullableEdge(trc, &regExpShapeWithDefaultProto, "global-regexp-shape");
 
   regExpRealm.trace(trc);
 
+#ifdef JS_HAS_INTL_API
+  globalIntlData.trace(trc);
+#endif
+
   TraceNullableEdge(trc, &mappedArgumentsTemplate, "mapped-arguments-template");
   TraceNullableEdge(trc, &unmappedArgumentsTemplate,
                     "unmapped-arguments-template");
