@@ -21,6 +21,9 @@
 #include "jstypes.h"
 #include "NamespaceImports.h"
 
+#ifdef JS_HAS_INTL_API
+#  include "builtin/intl/GlobalIntlData.h"
+#endif
 #include "gc/AllocKind.h"
 #include "js/CallArgs.h"
 #include "js/Class.h"
@@ -199,9 +202,17 @@ class GlobalObjectData {
   // Shape for BoundFunctionObject with %Function.prototype% as proto.
   GCPtr<SharedShape*> boundFunctionShapeWithDefaultProto;
 
+  // Shape for RegExpObject with %RegExp.prototype% as proto.
+  GCPtr<SharedShape*> regExpShapeWithDefaultProto;
+
   // Global state for regular expressions.
   RegExpRealm regExpRealm;
 
+#ifdef JS_HAS_INTL_API
+  // Cache Intl formatters.
+  intl::GlobalIntlData globalIntlData;
+#endif
+
   GCPtr<ArgumentsObject*> mappedArgumentsTemplate;
   GCPtr<ArgumentsObject*> unmappedArgumentsTemplate;
 
@@ -280,10 +291,11 @@ class GlobalObject : public NativeObject {
 
   void initBuiltinProto(ProtoKind kind, JSObject* proto) {
     MOZ_ASSERT(proto);
-    // Catch double-initialization; however if this is too much of a burden due
-    // to OOM handling it could be removed.
-    MOZ_ASSERT(!hasBuiltinProto(kind));
-    data().builtinProtos[kind].init(proto);
+    // Use set, as it's possible to construct oomTest test
+    // cases where the proto is already initialized.
+    //
+    // See Bugs 1969353 and 1928852.
+    data().builtinProtos[kind].set(proto);
   }
 
   void setBuiltinProto(ProtoKind kind, JSObject* proto) {
@@ -985,6 +997,10 @@ class GlobalObject : public NativeObject {
 
   RegExpRealm& regExpRealm() { return data().regExpRealm; }
 
+#ifdef JS_HAS_INTL_API
+  intl::GlobalIntlData& globalIntlData() { return data().globalIntlData; }
+#endif
+
   // Infallibly test whether the given value is the eval function for this
   // global.
   bool valueIsEval(const Value& val);
@@ -1112,6 +1128,13 @@ class GlobalObject : public NativeObject {
     data().boundFunctionShapeWithDefaultProto = shape;
   }
 
+  SharedShape* maybeRegExpShapeWithDefaultProto() const {
+    return data().regExpShapeWithDefaultProto;
+  }
+  void setRegExpShapeWithDefaultProto(SharedShape* shape) {
+    data().regExpShapeWithDefaultProto = shape;
+  }
+
   static PropertyIteratorObject* getOrCreateEmptyIterator(JSContext* cx);
 
   // Returns an object that represents the realm, used by embedder.
@@ -1184,22 +1207,22 @@ JSObject* GenericCreatePrototype(JSContext* cx, JSProtoKey key) {
   return GlobalObject::createBlankPrototype(cx, cx->global(), &T::protoClass_);
 }
 
-// Which object(s) should be marked as having a fuse property in
+// Which object(s) should be marked as having a RealmFuse property in
 // GenericFinishInit.
-enum class WhichHasFuseProperty {
+enum class WhichHasRealmFuseProperty {
   Proto,
   ProtoAndCtor,
 };
 
-template <WhichHasFuseProperty FuseProperty>
+template <WhichHasRealmFuseProperty FuseProperty>
 inline bool GenericFinishInit(JSContext* cx, HandleObject ctor,
                               HandleObject proto) {
-  if constexpr (FuseProperty == WhichHasFuseProperty::ProtoAndCtor) {
-    if (!JSObject::setHasFuseProperty(cx, ctor)) {
+  if constexpr (FuseProperty == WhichHasRealmFuseProperty::ProtoAndCtor) {
+    if (!JSObject::setHasRealmFuseProperty(cx, ctor)) {
       return false;
     }
   }
-  return JSObject::setHasFuseProperty(cx, proto);
+  return JSObject::setHasRealmFuseProperty(cx, proto);
 }
 
 inline JSProtoKey StandardProtoKeyOrNull(const JSObject* obj) {
