@@ -7,6 +7,7 @@
 #include "vm/Watchtower.h"
 
 #include "js/CallAndConstruct.h"
+#include "js/experimental/TypedData.h"
 #include "vm/Compartment.h"
 #include "vm/JSContext.h"
 #include "vm/JSObject.h"
@@ -80,7 +81,15 @@ static bool ReshapeForShadowedProp(JSContext* cx, Handle<NativeObject*> obj,
     if (!proto->is<NativeObject>()) {
       break;
     }
-    if (proto->as<NativeObject>().contains(cx, id)) {
+
+    Handle<NativeObject*> nproto = proto.as<NativeObject>();
+
+    if (mozilla::Maybe<PropertyInfo> propInfo = nproto->lookup(cx, id)) {
+      if (proto->hasObjectFuse()) {
+        if (auto* objFuse = cx->zone()->objectFuses.get(nproto)) {
+          objFuse->handleTeleportingShadowedProperty(cx, *propInfo);
+        }
+      }
       if (useDictionaryTeleporting) {
         JS_LOG(teleporting, Debug,
                "Shadowed Prop: Dictionary Reshape for Teleporting");
@@ -138,6 +147,25 @@ void MaybePopReturnFuses(JSContext* cx, Handle<NativeObject*> nobj) {
   }
 }
 
+static void MaybePopStringPrototypeSymbolsFuse(JSContext* cx, NativeObject* obj,
+                                               PropertyKey key) {
+  if (!key.isSymbol()) {
+    return;
+  }
+  GlobalObject* global = &obj->global();
+  if (obj != global->maybeGetPrototype(JSProto_String) &&
+      obj != global->maybeGetPrototype(JSProto_Object)) {
+    return;
+  }
+  if (key.toSymbol() == cx->wellKnownSymbols().match ||
+      key.toSymbol() == cx->wellKnownSymbols().replace ||
+      key.toSymbol() == cx->wellKnownSymbols().search ||
+      key.toSymbol() == cx->wellKnownSymbols().split) {
+    obj->realm()->realmFuses.optimizeStringPrototypeSymbolsFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
 // static
 bool Watchtower::watchPropertyAddSlow(JSContext* cx, Handle<NativeObject*> obj,
                                       HandleId id) {
@@ -154,6 +182,8 @@ bool Watchtower::watchPropertyAddSlow(JSContext* cx, Handle<NativeObject*> obj,
     if (id == NameToId(cx->names().return_)) {
       MaybePopReturnFuses(cx, obj);
     }
+
+    MaybePopStringPrototypeSymbolsFuse(cx, obj, id);
   }
 
   if (MOZ_UNLIKELY(obj->useWatchtowerTestingLog())) {
@@ -203,6 +233,12 @@ static bool ReshapeForProtoMutation(JSContext* cx, HandleObject obj) {
       cx->zone()->shapeZone().useDictionaryModeTeleportation();
 
   while (pobj && pobj->is<NativeObject>()) {
+    if (pobj->hasObjectFuse()) {
+      if (auto* objFuse =
+              cx->zone()->objectFuses.get(pobj.as<NativeObject>())) {
+        objFuse->handleTeleportingProtoMutation(cx);
+      }
+    }
     if (useDictionaryTeleporting) {
       MOZ_ASSERT(!pobj->hasInvalidatedTeleporting(),
                  "Once we start using invalidation shouldn't do any more "
@@ -227,6 +263,22 @@ static bool ReshapeForProtoMutation(JSContext* cx, HandleObject obj) {
   return true;
 }
 
+static constexpr bool IsTypedArrayProtoKey(JSProtoKey protoKey) {
+  switch (protoKey) {
+#define PROTO_KEY(_, T, N) \
+  case JSProto_##N##Array: \
+    return true;
+    JS_FOR_EACH_TYPED_ARRAY(PROTO_KEY)
+#undef PROTO_KEY
+    default:
+      return false;
+  }
+}
+
+static_assert(
+    !IsTypedArrayProtoKey(JSProto_TypedArray),
+    "IsTypedArrayProtoKey(JSProto_TypedArray) is expected to return false");
+
 static bool WatchProtoChangeImpl(JSContext* cx, HandleObject obj) {
   if (!obj->isUsedAsPrototype()) {
     return true;
@@ -247,6 +299,18 @@ static bool WatchProtoChangeImpl(JSContext* cx, HandleObject obj) {
       nobj->realm()->realmFuses.iteratorPrototypeHasObjectProto.popFuse(
           cx, nobj->realm()->realmFuses);
     }
+
+    if (nobj == nobj->global().maybeGetPrototype(JSProto_String)) {
+      nobj->realm()->realmFuses.optimizeStringPrototypeSymbolsFuse.popFuse(
+          cx, nobj->realm()->realmFuses);
+    }
+
+    auto protoKey = StandardProtoKeyOrNull(nobj);
+    if (IsTypedArrayProtoKey(protoKey) &&
+        nobj == nobj->global().maybeGetPrototype(protoKey)) {
+      nobj->realm()->realmFuses.optimizeTypedArraySpeciesFuse.popFuse(
+          cx, nobj->realm()->realmFuses);
+    }
   }
 
   return true;
@@ -406,7 +470,99 @@ static void MaybePopPromisePrototypeFuses(JSContext* cx, NativeObject* obj,
   }
 }
 
-static void MaybePopFuses(JSContext* cx, NativeObject* obj, jsid id) {
+static void MaybePopRegExpPrototypeFuses(JSContext* cx, NativeObject* obj,
+                                         jsid id) {
+  if (obj != obj->global().maybeGetPrototype(JSProto_RegExp)) {
+    return;
+  }
+  if (id.isAtom(cx->names().flags) || id.isAtom(cx->names().global) ||
+      id.isAtom(cx->names().hasIndices) || id.isAtom(cx->names().ignoreCase) ||
+      id.isAtom(cx->names().multiline) || id.isAtom(cx->names().sticky) ||
+      id.isAtom(cx->names().unicode) || id.isAtom(cx->names().unicodeSets) ||
+      id.isAtom(cx->names().dotAll) || id.isAtom(cx->names().exec) ||
+      id.isWellKnownSymbol(JS::SymbolCode::match) ||
+      id.isWellKnownSymbol(JS::SymbolCode::matchAll) ||
+      id.isWellKnownSymbol(JS::SymbolCode::replace) ||
+      id.isWellKnownSymbol(JS::SymbolCode::search) ||
+      id.isWellKnownSymbol(JS::SymbolCode::split)) {
+    obj->realm()->realmFuses.optimizeRegExpPrototypeFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopArrayBufferConstructorFuses(JSContext* cx,
+                                                NativeObject* obj, jsid id) {
+  if (obj != obj->global().maybeGetConstructor(JSProto_ArrayBuffer)) {
+    return;
+  }
+  if (id.isWellKnownSymbol(JS::SymbolCode::species)) {
+    obj->realm()->realmFuses.optimizeArrayBufferSpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopArrayBufferPrototypeFuses(JSContext* cx, NativeObject* obj,
+                                              jsid id) {
+  if (obj != obj->global().maybeGetPrototype(JSProto_ArrayBuffer)) {
+    return;
+  }
+  if (id.isAtom(cx->names().constructor)) {
+    obj->realm()->realmFuses.optimizeArrayBufferSpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopSharedArrayBufferConstructorFuses(JSContext* cx,
+                                                      NativeObject* obj,
+                                                      jsid id) {
+  if (obj != obj->global().maybeGetConstructor(JSProto_SharedArrayBuffer)) {
+    return;
+  }
+  if (id.isWellKnownSymbol(JS::SymbolCode::species)) {
+    obj->realm()->realmFuses.optimizeSharedArrayBufferSpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopSharedArrayBufferPrototypeFuses(JSContext* cx,
+                                                    NativeObject* obj,
+                                                    jsid id) {
+  if (obj != obj->global().maybeGetPrototype(JSProto_SharedArrayBuffer)) {
+    return;
+  }
+  if (id.isAtom(cx->names().constructor)) {
+    obj->realm()->realmFuses.optimizeSharedArrayBufferSpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopTypedArrayConstructorFuses(JSContext* cx, NativeObject* obj,
+                                               jsid id) {
+  if (obj != obj->global().maybeGetConstructor(JSProto_TypedArray)) {
+    return;
+  }
+  if (id.isWellKnownSymbol(JS::SymbolCode::species)) {
+    obj->realm()->realmFuses.optimizeTypedArraySpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopTypedArrayPrototypeFuses(JSContext* cx, NativeObject* obj,
+                                             jsid id) {
+  auto protoKey = StandardProtoKeyOrNull(obj);
+  if (protoKey != JSProto_TypedArray && !IsTypedArrayProtoKey(protoKey)) {
+    return;
+  }
+  if (obj != obj->global().maybeGetPrototype(protoKey)) {
+    return;
+  }
+  if (id.isAtom(cx->names().constructor)) {
+    obj->realm()->realmFuses.optimizeTypedArraySpeciesFuse.popFuse(
+        cx, obj->realm()->realmFuses);
+  }
+}
+
+static void MaybePopRealmFuses(JSContext* cx, NativeObject* obj, jsid id) {
   // Handle writes to Array constructor fuse properties.
   MaybePopArrayConstructorFuses(cx, obj, id);
 
@@ -439,12 +595,34 @@ static void MaybePopFuses(JSContext* cx, NativeObject* obj, jsid id) {
 
   // Handle writes to Promise.prototype fuse properties.
   MaybePopPromisePrototypeFuses(cx, obj, id);
+
+  // Handle writes to RegExp.prototype fuse properties.
+  MaybePopRegExpPrototypeFuses(cx, obj, id);
+
+  // Handle writes to ArrayBuffer constructor fuse properties.
+  MaybePopArrayBufferConstructorFuses(cx, obj, id);
+
+  // Handle writes to ArrayBuffer.prototype fuse properties.
+  MaybePopArrayBufferPrototypeFuses(cx, obj, id);
+
+  // Handle writes to SharedArrayBuffer constructor fuse properties.
+  MaybePopSharedArrayBufferConstructorFuses(cx, obj, id);
+
+  // Handle writes to SharedArrayBuffer.prototype fuse properties.
+  MaybePopSharedArrayBufferPrototypeFuses(cx, obj, id);
+
+  // Handle writes to %TypedArray% constructor fuse properties.
+  MaybePopTypedArrayConstructorFuses(cx, obj, id);
+
+  // Handle writes to %TypedArray%.prototype and concrete TypedArray.prototype
+  // fuse properties.
+  MaybePopTypedArrayPrototypeFuses(cx, obj, id);
 }
 
 // static
 bool Watchtower::watchPropertyRemoveSlow(JSContext* cx,
-                                         Handle<NativeObject*> obj,
-                                         HandleId id) {
+                                         Handle<NativeObject*> obj, HandleId id,
+                                         PropertyInfo propInfo) {
   MOZ_ASSERT(watchesPropertyRemove(obj));
 
   if (obj->isUsedAsPrototype() && !id.isInt()) {
@@ -455,8 +633,13 @@ bool Watchtower::watchPropertyRemoveSlow(JSContext* cx,
     obj->as<GlobalObject>().bumpGenerationCount();
   }
 
-  if (MOZ_UNLIKELY(obj->hasFuseProperty())) {
-    MaybePopFuses(cx, obj, id);
+  if (MOZ_UNLIKELY(obj->hasRealmFuseProperty())) {
+    MaybePopRealmFuses(cx, obj, id);
+  }
+  if (obj->hasObjectFuse()) {
+    if (auto* objFuse = cx->zone()->objectFuses.get(obj)) {
+      objFuse->handlePropertyRemove(cx, propInfo);
+    }
   }
 
   if (MOZ_UNLIKELY(obj->useWatchtowerTestingLog())) {
@@ -517,13 +700,23 @@ void Watchtower::watchPropertyValueChangeSlow(
   // accessor property or when redefining a data property as an accessor
   // property and vice versa.
 
+  // Handle object fuses before the check for no-op changes below. We don't
+  // attach SetProp stubs for constant properties, so if a constant property is
+  // overwritten with the same value, we want to mark it non-constant.
+  // See Watchtower::canOptimizeSetSlotSlow.
+  if (obj->hasObjectFuse()) {
+    if (auto* objFuse = cx->zone()->objectFuses.get(obj)) {
+      objFuse->handlePropertyValueChange(cx, propInfo);
+    }
+  }
+
   if (propInfo.hasSlot() && obj->getSlot(propInfo.slot()) == value) {
     // We're not actually changing the property's value.
     return;
   }
 
-  if (MOZ_UNLIKELY(obj->hasFuseProperty())) {
-    MaybePopFuses(cx, obj, id);
+  if (MOZ_UNLIKELY(obj->hasRealmFuseProperty())) {
+    MaybePopRealmFuses(cx, obj, id);
   }
 
   // If we cannot GC, we can't manipulate the log, but we need to be able to
@@ -553,6 +746,30 @@ template void Watchtower::watchPropertyValueChangeSlow<AllowGC::NoGC>(
     typename MaybeRooted<Value, AllowGC::NoGC>::HandleType value,
     PropertyInfo propInfo);
 
+// static
+SetSlotOptimizable Watchtower::canOptimizeSetSlotSlow(JSContext* cx,
+                                                      NativeObject* obj,
+                                                      PropertyInfo prop) {
+  MOZ_ASSERT(obj->hasObjectFuse());
+
+  ObjectFuse* objFuse = cx->zone()->objectFuses.getOrCreate(cx, obj);
+  if (!objFuse) {
+    cx->recoverFromOutOfMemory();
+    return SetSlotOptimizable::No;
+  }
+
+  if (objFuse->canOptimizeSetSlot(prop)) {
+    return SetSlotOptimizable::Yes;
+  }
+
+  // If a property is constant, there's no point in attaching a SetProp IC stub.
+  // The next time we set this property, we have to call into the VM to mark
+  // it NotConstant and potentially pop fuses. After that, we can attach a
+  // regular SetProp IC stub. If we never set this property again, there's no
+  // need to optimize this SetProp.
+  return SetSlotOptimizable::NotYet;
+}
+
 // static
 bool Watchtower::watchFreezeOrSealSlow(JSContext* cx, Handle<NativeObject*> obj,
                                        IntegrityLevel level) {
@@ -590,6 +807,17 @@ bool Watchtower::watchObjectSwapSlow(JSContext* cx, HandleObject a,
     return false;
   }
 
+  if (a->hasObjectFuse()) {
+    if (auto* objFuse = cx->zone()->objectFuses.get(a.as<NativeObject>())) {
+      objFuse->handleObjectSwap(cx);
+    }
+  }
+  if (b->hasObjectFuse()) {
+    if (auto* objFuse = cx->zone()->objectFuses.get(b.as<NativeObject>())) {
+      objFuse->handleObjectSwap(cx);
+    }
+  }
+
   // Note: we don't invoke the testing callback for swap because the objects may
   // not be safe to expose to JS at this point. See bug 1754699.
 
