@@ -100,22 +100,6 @@ bool FeatureOptions::init(JSContext* cx, HandleValue val) {
     return true;
   }
 
-  bool jsStringBuiltinsAvailable = false;
-#ifdef ENABLE_WASM_JS_STRING_BUILTINS
-  jsStringBuiltinsAvailable = JSStringBuiltinsAvailable(cx);
-#endif  // ENABLE_WASM_JS_STRING_BUILTINS
-  bool isPrivilegedContext = IsPrivilegedContext(cx);
-
-  if (!jsStringBuiltinsAvailable && !isPrivilegedContext) {
-    // Skip checking for a compile options object if we don't have a feature
-    // enabled yet that requires it. Once js-string-builtins is standardized
-    // and shipped we will always need to check for it.
-    MOZ_ASSERT(!this->disableOptimizingCompiler);
-    MOZ_ASSERT(!this->jsStringConstants);
-    MOZ_ASSERT(!this->jsStringBuiltins);
-    return true;
-  }
-
   if (!val.isObject()) {
     JS_ReportErrorNumberUTF8(cx, GetErrorMessage, nullptr,
                              JSMSG_WASM_BAD_COMPILE_OPTIONS);
@@ -123,7 +107,7 @@ bool FeatureOptions::init(JSContext* cx, HandleValue val) {
   }
   RootedObject obj(cx, &val.toObject());
 
-  if (isPrivilegedContext) {
+  if (IsPrivilegedContext(cx)) {
     RootedValue disableOptimizingCompiler(cx);
     if (!JS_GetProperty(cx, obj, "disableOptimizingCompiler",
                         &disableOptimizingCompiler)) {
@@ -131,89 +115,92 @@ bool FeatureOptions::init(JSContext* cx, HandleValue val) {
     }
 
     this->disableOptimizingCompiler = JS::ToBoolean(disableOptimizingCompiler);
+
+    RootedValue mozIntGemm(cx);
+    if (!JS_GetProperty(cx, obj, "mozIntGemm", &mozIntGemm)) {
+      return false;
+    }
+
+    this->mozIntGemm = JS::ToBoolean(mozIntGemm);
   } else {
     MOZ_ASSERT(!this->disableOptimizingCompiler);
   }
 
-#ifdef ENABLE_WASM_JS_STRING_BUILTINS
-  if (jsStringBuiltinsAvailable) {
-    // Check the 'importedStringConstants' option
-    RootedValue importedStringConstants(cx);
-    if (!JS_GetProperty(cx, obj, "importedStringConstants",
-                        &importedStringConstants)) {
+  // Check the 'importedStringConstants' option
+  RootedValue importedStringConstants(cx);
+  if (!JS_GetProperty(cx, obj, "importedStringConstants",
+                      &importedStringConstants)) {
+    return false;
+  }
+
+  if (importedStringConstants.isNullOrUndefined()) {
+    this->jsStringConstants = false;
+  } else {
+    this->jsStringConstants = true;
+
+    RootedString importedStringConstantsString(
+        cx, JS::ToString(cx, importedStringConstants));
+    if (!importedStringConstantsString) {
       return false;
     }
 
-    if (importedStringConstants.isNullOrUndefined()) {
-      this->jsStringConstants = false;
-    } else {
-      this->jsStringConstants = true;
-
-      RootedString importedStringConstantsString(
-          cx, JS::ToString(cx, importedStringConstants));
-      if (!importedStringConstantsString) {
-        return false;
-      }
-
-      UniqueChars jsStringConstantsNamespace =
-          StringToNewUTF8CharsZ(cx, *importedStringConstantsString);
-      if (!jsStringConstantsNamespace) {
-        return false;
-      }
-
-      this->jsStringConstantsNamespace =
-          js_new<ShareableChars>(std::move(jsStringConstantsNamespace));
-      if (!this->jsStringConstantsNamespace) {
-        return false;
-      }
-    }
-
-    // Get the `builtins` iterable
-    RootedValue builtins(cx);
-    if (!JS_GetProperty(cx, obj, "builtins", &builtins)) {
+    UniqueChars jsStringConstantsNamespace =
+        StringToNewUTF8CharsZ(cx, *importedStringConstantsString);
+    if (!jsStringConstantsNamespace) {
       return false;
     }
 
-    if (!builtins.isUndefined()) {
-      JS::ForOfIterator iterator(cx);
+    this->jsStringConstantsNamespace =
+        cx->new_<ShareableChars>(std::move(jsStringConstantsNamespace));
+    if (!this->jsStringConstantsNamespace) {
+      return false;
+    }
+  }
 
-      if (!iterator.init(builtins, JS::ForOfIterator::ThrowOnNonIterable)) {
+  // Get the `builtins` iterable
+  RootedValue builtins(cx);
+  if (!JS_GetProperty(cx, obj, "builtins", &builtins)) {
+    return false;
+  }
+
+  if (!builtins.isUndefined()) {
+    JS::ForOfIterator iterator(cx);
+
+    if (!iterator.init(builtins, JS::ForOfIterator::ThrowOnNonIterable)) {
+      return false;
+    }
+
+    RootedValue jsStringModule(cx, StringValue(cx->names().jsStringModule));
+    RootedValue nextBuiltin(cx);
+    while (true) {
+      bool done;
+      if (!iterator.next(&nextBuiltin, &done)) {
         return false;
       }
+      if (done) {
+        break;
+      }
 
-      RootedValue jsStringModule(cx, StringValue(cx->names().jsStringModule));
-      RootedValue nextBuiltin(cx);
-      while (true) {
-        bool done;
-        if (!iterator.next(&nextBuiltin, &done)) {
-          return false;
-        }
-        if (done) {
-          break;
-        }
+      bool jsStringBuiltins;
+      if (!JS::LooselyEqual(cx, nextBuiltin, jsStringModule,
+                            &jsStringBuiltins)) {
+        return false;
+      }
 
-        bool jsStringBuiltins;
-        if (!JS::LooselyEqual(cx, nextBuiltin, jsStringModule,
-                              &jsStringBuiltins)) {
-          return false;
-        }
+      // We ignore unknown builtins
+      if (!jsStringBuiltins) {
+        continue;
+      }
 
-        // We ignore unknown builtins
-        if (!jsStringBuiltins) {
-          continue;
-        }
-
-        // You cannot request the same builtin twice
-        if (this->jsStringBuiltins && jsStringBuiltins) {
-          JS_ReportErrorNumberUTF8(cx, GetErrorMessage, nullptr,
-                                   JSMSG_WASM_DUPLICATE_BUILTIN);
-          return false;
-        }
-        this->jsStringBuiltins = jsStringBuiltins;
+      // You cannot request the same builtin twice
+      if (this->jsStringBuiltins && jsStringBuiltins) {
+        JS_ReportErrorNumberUTF8(cx, GetErrorMessage, nullptr,
+                                 JSMSG_WASM_DUPLICATE_BUILTIN);
+        return false;
       }
+      this->jsStringBuiltins = jsStringBuiltins;
     }
   }
-#endif
 
   return true;
 }
@@ -231,15 +218,12 @@ FeatureArgs FeatureArgs::build(JSContext* cx, const FeatureOptions& options) {
 
   features.simd = jit::JitSupportsWasmSimd();
   features.isBuiltinModule = options.isBuiltinModule;
-  if (features.jsStringBuiltins) {
-    features.builtinModules.jsString = options.jsStringBuiltins;
-    features.builtinModules.jsStringConstants = options.jsStringConstants;
-    features.builtinModules.jsStringConstantsNamespace =
-        options.jsStringConstantsNamespace;
-  }
-  if (options.requireExnref) {
-    features.exnref = true;
-  }
+  features.builtinModules.jsString = options.jsStringBuiltins;
+  features.builtinModules.jsStringConstants = options.jsStringConstants;
+  features.builtinModules.jsStringConstantsNamespace =
+      options.jsStringConstantsNamespace;
+  features.builtinModules.intGemm =
+      MozIntGemmAvailable(cx) && options.mozIntGemm;
 
   return features;
 }
@@ -380,16 +364,16 @@ BytecodeSource::BytecodeSource(const uint8_t* begin, size_t length) {
   BytecodeRange codeRange;
   BytecodeRange tailRange;
   if (StartsCodeSection(begin, begin + length, &codeRange)) {
-    if (codeRange.end() <= length) {
+    if (codeRange.end <= length) {
       envRange = BytecodeRange(0, codeRange.start);
-      tailRange = BytecodeRange(codeRange.end(), length - codeRange.end());
+      tailRange = BytecodeRange(codeRange.end, length - codeRange.end);
     } else {
       MOZ_RELEASE_ASSERT(codeRange.start <= length);
       // If the specified code range is larger than the buffer, clamp it to the
       // the buffer size. This buffer will be rejected later.
       envRange = BytecodeRange(0, codeRange.start);
       codeRange = BytecodeRange(codeRange.start, length - codeRange.start);
-      MOZ_RELEASE_ASSERT(codeRange.end() == length);
+      MOZ_RELEASE_ASSERT(codeRange.end == length);
       tailRange = BytecodeRange(length, 0);
     }
   } else {
@@ -417,7 +401,7 @@ BytecodeBuffer::BytecodeBuffer(const ShareableBytes* env,
 bool BytecodeBuffer::fromSource(const BytecodeSource& bytecodeSource,
                                 BytecodeBuffer* bytecodeBuffer) {
   SharedBytes env;
-  if (bytecodeSource.envRange().size) {
+  if (!bytecodeSource.envRange().isEmpty()) {
     env = ShareableBytes::fromSpan(bytecodeSource.envSpan());
     if (!env) {
       return false;
@@ -425,7 +409,8 @@ bool BytecodeBuffer::fromSource(const BytecodeSource& bytecodeSource,
   }
 
   SharedBytes code;
-  if (bytecodeSource.hasCodeSection() && bytecodeSource.codeRange().size) {
+  if (bytecodeSource.hasCodeSection() &&
+      !bytecodeSource.codeRange().isEmpty()) {
     code = ShareableBytes::fromSpan(bytecodeSource.codeSpan());
     if (!code) {
       return false;
@@ -433,7 +418,8 @@ bool BytecodeBuffer::fromSource(const BytecodeSource& bytecodeSource,
   }
 
   SharedBytes tail;
-  if (bytecodeSource.hasCodeSection() && bytecodeSource.tailRange().size) {
+  if (bytecodeSource.hasCodeSection() &&
+      !bytecodeSource.tailRange().isEmpty()) {
     tail = ShareableBytes::fromSpan(bytecodeSource.tailSpan());
     if (!tail) {
       return false;
@@ -835,6 +821,7 @@ static bool TieringBeneficial(bool lazyTiering, uint32_t codeSize) {
 
 // Ensure that we have the non-compiler requirements to tier safely.
 static bool PlatformCanTier(bool lazyTiering) {
+  // Note: ensure this function stays in sync with `WasmLazyTieringEnabled()`.
   // Tiering needs background threads if we're using eager tiering or we're
   // using lazy tiering without the synchronous flag.
   bool synchronousTiering =
@@ -884,9 +871,12 @@ void CompilerEnvironment::computeParameters(const ModuleMetadata& moduleMeta) {
   uint32_t codeSectionSize = moduleMeta.codeMeta->codeSectionSize();
 
   // We use lazy tiering if the 'for-all' pref is enabled, or the 'gc-only'
-  // pref is enabled and we're compiling a GC module.
-  bool lazyTiering = JS::Prefs::wasm_lazy_tiering() ||
-                     (JS::Prefs::wasm_lazy_tiering_for_gc() && isGcModule);
+  // pref is enabled and we're compiling a GC module.  However, forcing
+  // serialization-testing disables lazy tiering.
+  bool testSerialization = args_->features.testSerialization;
+  bool lazyTiering = (JS::Prefs::wasm_lazy_tiering() ||
+                      (JS::Prefs::wasm_lazy_tiering_for_gc() && isGcModule)) &&
+                     !testSerialization;
 
   if (baselineEnabled && hasSecondTier &&
       (TieringBeneficial(lazyTiering, codeSectionSize) || forceTiering) &&
@@ -1037,7 +1027,7 @@ SharedModule wasm::CompileBuffer(const CompileArgs& args,
     MOZ_RELEASE_ASSERT(envDecoder.done());
   }
 
-  return mg.finishModule(bytecode, moduleMeta, listener);
+  return mg.finishModule(bytecode, *moduleMeta, listener);
 }
 
 bool wasm::CompileCompleteTier2(const ShareableBytes* codeSection,
@@ -1088,8 +1078,8 @@ bool wasm::CompilePartialTier2(const Code& code, uint32_t funcIndex,
     return false;
   }
 
-  const BytecodeRange& funcRange = codeMeta.funcDefRange(funcIndex);
-  BytecodeSpan funcBytecode = codeMeta.funcDefBody(funcIndex);
+  const BytecodeRange& funcRange = code.codeTailMeta().funcDefRange(funcIndex);
+  BytecodeSpan funcBytecode = code.codeTailMeta().funcDefBody(funcIndex);
 
   // The following sequence will compile/finish this function, on this thread.
   // `error` (as stashed in `mg`) may get set to, for example, "stack frame too
@@ -1171,7 +1161,7 @@ SharedModule wasm::CompileStreaming(
       return nullptr;
     }
 
-    MOZ_RELEASE_ASSERT(codeMeta.codeSectionRange->size == codeBytes.length());
+    MOZ_RELEASE_ASSERT(codeMeta.codeSectionRange->size() == codeBytes.length());
     MOZ_RELEASE_ASSERT(d.done());
   }
 
@@ -1210,7 +1200,7 @@ SharedModule wasm::CompileStreaming(
   const ShareableBytes& tailBytes = *streamEnd.tailBytes;
 
   {
-    Decoder d(tailBytes.vector, codeMeta.codeSectionRange->end(), error,
+    Decoder d(tailBytes.vector, codeMeta.codeSectionRange->end, error,
               warnings);
 
     if (!DecodeModuleTail(d, &codeMeta, moduleMeta)) {
@@ -1221,7 +1211,7 @@ SharedModule wasm::CompileStreaming(
   }
 
   BytecodeBuffer bytecodeBuffer(&envBytes, &codeBytes, &tailBytes);
-  return mg.finishModule(BytecodeBufferOrSource(bytecodeBuffer), moduleMeta,
+  return mg.finishModule(BytecodeBufferOrSource(bytecodeBuffer), *moduleMeta,
                          streamEnd.completeTier2Listener);
 }
 
