@@ -199,7 +199,13 @@ void CookieServiceChild::RemoveSingleCookie(const CookieStruct& aCookie,
                                             const OriginAttributes& aAttrs,
                                             const Maybe<nsID>& aOperationID) {
   nsCString baseDomain;
-  CookieCommons::GetBaseDomainFromHost(mTLDService, aCookie.host(), baseDomain);
+  if (NS_FAILED(CookieCommons::GetBaseDomainFromHost(
+          mTLDService, aCookie.host(), baseDomain))) {
+    MOZ_ASSERT(false,
+               "CookieServiceChild::RemoveSingleCookie - GetBaseDomainFromHost "
+               "shouldn't fail");
+    return;
+  }
   CookieKey key(baseDomain, aAttrs);
   CookiesList* cookiesList = nullptr;
   mCookiesMap.Get(key, &cookiesList);
@@ -306,8 +312,13 @@ CookieServiceChild::CookieNotificationAction
 CookieServiceChild::RecordDocumentCookie(Cookie* aCookie,
                                          const OriginAttributes& aAttrs) {
   nsAutoCString baseDomain;
-  CookieCommons::GetBaseDomainFromHost(mTLDService, aCookie->Host(),
-                                       baseDomain);
+  if (NS_FAILED(CookieCommons::GetBaseDomainFromHost(
+          mTLDService, aCookie->Host(), baseDomain))) {
+    MOZ_ASSERT(false,
+               "CookieServiceChild::RecordDocumentCookie - "
+               "GetBaseDomainFromHost shouldn't fail");
+    return CookieNotificationAction::NoActionNeeded;
+  }
 
   if (CookieCommons::IsFirstPartyPartitionedCookieWithoutCHIPS(
           aCookie, baseDomain, aAttrs)) {
@@ -339,7 +350,6 @@ CookieServiceChild::RecordDocumentCookie(Cookie* aCookie,
           cookie->Expiry() == aCookie->Expiry() &&
           cookie->IsSecure() == aCookie->IsSecure() &&
           cookie->SameSite() == aCookie->SameSite() &&
-          cookie->RawSameSite() == aCookie->RawSameSite() &&
           cookie->IsSession() == aCookie->IsSession() &&
           cookie->IsHttpOnly() == aCookie->IsHttpOnly()) {
         cookie->SetLastAccessed(aCookie->LastAccessed());
@@ -351,7 +361,7 @@ CookieServiceChild::RecordDocumentCookie(Cookie* aCookie,
     }
   }
 
-  int64_t currentTime = PR_Now() / PR_USEC_PER_SEC;
+  int64_t currentTime = PR_Now() / PR_USEC_PER_MSEC;
   if (aCookie->Expiry() <= currentTime) {
     return cookieFound ? CookieNotificationAction::CookieDeleted
                        : CookieNotificationAction::NoActionNeeded;
@@ -366,6 +376,7 @@ NS_IMETHODIMP
 CookieServiceChild::GetCookieStringFromHttp(nsIURI* /*aHostURI*/,
                                             nsIChannel* /*aChannel*/,
                                             nsACString& /*aCookieString*/) {
+  MOZ_CRASH("This method should not be called");
   return NS_ERROR_NOT_IMPLEMENTED;
 }
 
@@ -373,170 +384,8 @@ NS_IMETHODIMP
 CookieServiceChild::SetCookieStringFromHttp(nsIURI* aHostURI,
                                             const nsACString& aCookieString,
                                             nsIChannel* aChannel) {
-  NS_ENSURE_ARG(aHostURI);
-  NS_ENSURE_ARG(aChannel);
-
-  if (!CookieCommons::IsSchemeSupported(aHostURI)) {
-    return NS_OK;
-  }
-
-  // Fast past: don't bother sending IPC messages about nullprincipal'd
-  // documents.
-  nsAutoCString scheme;
-  aHostURI->GetScheme(scheme);
-  if (scheme.EqualsLiteral("moz-nullprincipal")) {
-    return NS_OK;
-  }
-
-  nsCOMPtr<nsILoadInfo> loadInfo = aChannel->LoadInfo();
-
-  uint32_t rejectedReason = 0;
-  ThirdPartyAnalysisResult result = mThirdPartyUtil->AnalyzeChannel(
-      aChannel, false, aHostURI, RequireThirdPartyCheck, &rejectedReason);
-
-  nsCString cookieString(aCookieString);
-
-  OriginAttributes storagePrincipalOriginAttributes =
-      loadInfo->GetOriginAttributes();
-  StoragePrincipalHelper::PrepareEffectiveStoragePrincipalOriginAttributes(
-      aChannel, storagePrincipalOriginAttributes);
-
-  bool requireHostMatch;
-  nsCString baseDomain;
-  CookieCommons::GetBaseDomain(mTLDService, aHostURI, baseDomain,
-                               requireHostMatch);
-
-  nsCOMPtr<nsICookieJarSettings> cookieJarSettings =
-      CookieCommons::GetCookieJarSettings(aChannel);
-
-  nsCOMPtr<nsIConsoleReportCollector> crc = do_QueryInterface(aChannel);
-
-  CookieStatus cookieStatus = CookieService::CheckPrefs(
-      crc, cookieJarSettings, aHostURI,
-      result.contains(ThirdPartyAnalysis::IsForeign),
-      result.contains(ThirdPartyAnalysis::IsThirdPartyTrackingResource),
-      result.contains(ThirdPartyAnalysis::IsThirdPartySocialTrackingResource),
-      result.contains(ThirdPartyAnalysis::IsStorageAccessPermissionGranted),
-      aCookieString,
-      HasExistingCookies(baseDomain, storagePrincipalOriginAttributes),
-      storagePrincipalOriginAttributes, &rejectedReason);
-
-  if (cookieStatus != STATUS_ACCEPTED &&
-      cookieStatus != STATUS_ACCEPT_SESSION) {
-    return NS_OK;
-  }
-
-  int64_t currentTimeInUsec = PR_Now();
-
-  bool addonAllowsLoad = false;
-  nsCOMPtr<nsIURI> finalChannelURI;
-  NS_GetFinalChannelURI(aChannel, getter_AddRefs(finalChannelURI));
-  addonAllowsLoad = BasePrincipal::Cast(loadInfo->TriggeringPrincipal())
-                        ->AddonAllowsLoad(finalChannelURI);
-
-  bool isForeignAndNotAddon = false;
-  if (!addonAllowsLoad) {
-    mThirdPartyUtil->IsThirdPartyChannel(aChannel, aHostURI,
-                                         &isForeignAndNotAddon);
-
-    // include sub-document navigations from cross-site to same-site
-    // wrt top-level in our check for thirdparty-ness
-    if (StaticPrefs::network_cookie_sameSite_crossSiteIframeSetCheck() &&
-        !isForeignAndNotAddon &&
-        loadInfo->GetExternalContentPolicyType() ==
-            ExtContentPolicy::TYPE_SUBDOCUMENT) {
-      bool triggeringPrincipalIsThirdParty = false;
-      BasePrincipal::Cast(loadInfo->TriggeringPrincipal())
-          ->IsThirdPartyURI(finalChannelURI, &triggeringPrincipalIsThirdParty);
-      isForeignAndNotAddon |= triggeringPrincipalIsThirdParty;
-    }
-  }
-
-  bool mustBePartitioned =
-      isForeignAndNotAddon &&
-      cookieJarSettings->GetCookieBehavior() ==
-          nsICookieService::BEHAVIOR_REJECT_TRACKER_AND_PARTITION_FOREIGN &&
-      !result.contains(ThirdPartyAnalysis::IsStorageAccessPermissionGranted);
-
-  // CHIPS - The partitioned cookie jar is always available and it is always
-  // possible to store cookies in it using the "Partitioned" attribute.
-  // Prepare the partitioned principals OAs to enable possible partitioned
-  // cookie storing from first-party or with StorageAccess.
-  // Similar behavior to CookieService::SetCookieStringFromHttp().
-  OriginAttributes partitionedPrincipalOriginAttributes;
-  bool isPartitionedPrincipal =
-      !storagePrincipalOriginAttributes.mPartitionKey.IsEmpty();
-  bool isCHIPS = StaticPrefs::network_cookie_CHIPS_enabled() &&
-                 !cookieJarSettings->GetBlockingAllContexts();
-  // Only need to get OAs if we don't already use the partitioned principal.
-  if (isCHIPS && !isPartitionedPrincipal) {
-    StoragePrincipalHelper::GetOriginAttributes(
-        aChannel, partitionedPrincipalOriginAttributes,
-        StoragePrincipalHelper::ePartitionedPrincipal);
-  }
-
-  nsAutoCString dateHeader;
-  CookieCommons::GetServerDateHeader(aChannel, dateHeader);
-
-  CookieParser parser(crc, aHostURI);
-  parser.Parse(baseDomain, requireHostMatch, cookieStatus, cookieString,
-               dateHeader, true, isForeignAndNotAddon, mustBePartitioned,
-               storagePrincipalOriginAttributes.IsPrivateBrowsing(),
-               loadInfo->GetIsOn3PCBExceptionList());
-  if (!parser.ContainsCookie()) {
-    return NS_OK;
-  }
-
-  // check permissions from site permission list.
-  if (!CookieCommons::CheckCookiePermission(aChannel, parser.CookieData())) {
-    COOKIE_LOGFAILURE(SET_COOKIE, aHostURI, aCookieString,
-                      "cookie rejected by permission manager");
-    parser.RejectCookie(CookieParser::RejectedByPermissionManager);
-    CookieCommons::NotifyRejected(
-        aHostURI, aChannel,
-        nsIWebProgressListener::STATE_COOKIES_BLOCKED_BY_PERMISSION,
-        OPERATION_WRITE);
-    return NS_OK;
-  }
-
-  // CHIPS - If the partitioned attribute is set, store cookie in partitioned
-  // cookie jar independent of context. If the cookies are stored in the
-  // partitioned cookie jar anyway no special treatment of CHIPS cookies
-  // necessary.
-  bool needPartitioned =
-      isCHIPS && parser.CookieData().isPartitioned() && !isPartitionedPrincipal;
-  OriginAttributes& cookieOriginAttributes =
-      needPartitioned ? partitionedPrincipalOriginAttributes
-                      : storagePrincipalOriginAttributes;
-  // Assert that partitionedPrincipalOriginAttributes are initialized if used.
-  MOZ_ASSERT_IF(needPartitioned,
-                !partitionedPrincipalOriginAttributes.mPartitionKey.IsEmpty());
-
-  RefPtr<Cookie> cookie =
-      Cookie::Create(parser.CookieData(), cookieOriginAttributes);
-  MOZ_ASSERT(cookie);
-
-  cookie->SetLastAccessed(currentTimeInUsec);
-  cookie->SetCreationTime(
-      Cookie::GenerateUniqueCreationTime(currentTimeInUsec));
-
-  CookieNotificationAction action =
-      RecordDocumentCookie(cookie, cookieOriginAttributes);
-  NotifyObservers(cookie, cookieOriginAttributes, action);
-
-  // Asynchronously call the parent.
-  if (CanSend()) {
-    nsTArray<CookieStruct> cookies;
-    cookies.AppendElement(parser.CookieData());
-
-    RefPtr<HttpChannelChild> httpChannelChild = do_QueryObject(aChannel);
-    MOZ_ASSERT(httpChannelChild);
-    httpChannelChild->SendSetCookies(baseDomain, cookieOriginAttributes,
-                                     aHostURI, true, isForeignAndNotAddon,
-                                     cookies);
-  }
-
-  return NS_OK;
+  MOZ_CRASH("This method should not be called");
+  return NS_ERROR_NOT_IMPLEMENTED;
 }
 
 NS_IMETHODIMP
@@ -672,8 +521,13 @@ void CookieServiceChild::NotifyObservers(Cookie* aCookie,
   }
 
   nsAutoCString baseDomain;
-  CookieCommons::GetBaseDomainFromHost(mTLDService, aCookie->Host(),
-                                       baseDomain);
+  if (NS_FAILED(CookieCommons::GetBaseDomainFromHost(
+          mTLDService, aCookie->Host(), baseDomain))) {
+    MOZ_ASSERT(false,
+               "CookieServiceChild::NotifyObservers - GetBaseDomainFromHost "
+               "shouldn't fail");
+    return;
+  }
 
   nsCOMPtr<nsICookieNotification> notification =
       new CookieNotification(notificationAction, aCookie, baseDomain, false,
