@@ -5,6 +5,41 @@
 #include "AudioEncoderBinding.h"
 #include "BlobBinding.h"
 #include "BroadcastChannelBinding.h"
+#include "CSSColorBinding.h"
+#include "CSSColorValueBinding.h"
+#include "CSSHSLBinding.h"
+#include "CSSHWBBinding.h"
+#include "CSSImageValueBinding.h"
+#include "CSSKeywordValueBinding.h"
+#include "CSSLCHBinding.h"
+#include "CSSLabBinding.h"
+#include "CSSMathClampBinding.h"
+#include "CSSMathInvertBinding.h"
+#include "CSSMathMaxBinding.h"
+#include "CSSMathMinBinding.h"
+#include "CSSMathNegateBinding.h"
+#include "CSSMathProductBinding.h"
+#include "CSSMathSumBinding.h"
+#include "CSSMathValueBinding.h"
+#include "CSSMatrixComponentBinding.h"
+#include "CSSNumericArrayBinding.h"
+#include "CSSNumericValueBinding.h"
+#include "CSSOKLCHBinding.h"
+#include "CSSOKLabBinding.h"
+#include "CSSPerspectiveBinding.h"
+#include "CSSRGBBinding.h"
+#include "CSSRotateBinding.h"
+#include "CSSScaleBinding.h"
+#include "CSSSkewBinding.h"
+#include "CSSSkewXBinding.h"
+#include "CSSSkewYBinding.h"
+#include "CSSStyleValueBinding.h"
+#include "CSSTransformComponentBinding.h"
+#include "CSSTransformValueBinding.h"
+#include "CSSTranslateBinding.h"
+#include "CSSUnitValueBinding.h"
+#include "CSSUnparsedValueBinding.h"
+#include "CSSVariableReferenceValueBinding.h"
 #include "CacheBinding.h"
 #include "CacheStorageBinding.h"
 #include "CanvasRenderingContext2DBinding.h"
@@ -84,6 +119,7 @@
 #include "NetworkInformationBinding.h"
 #include "NotificationBinding.h"
 #include "NotificationEventBinding.h"
+#include "ONNXBinding.h"
 #include "OffscreenCanvasBinding.h"
 #include "OffscreenCanvasRenderingContext2DBinding.h"
 #include "PathUtilsBinding.h"
@@ -127,6 +163,7 @@
 #include "StorageManagerBinding.h"
 #include "StructuredCloneHolderBinding.h"
 #include "StructuredCloneTesterBinding.h"
+#include "StylePropertyMapReadOnlyBinding.h"
 #include "SubtleCryptoBinding.h"
 #include "TaskPriorityChangeEventBinding.h"
 #include "TestUtilsBinding.h"
@@ -138,6 +175,7 @@
 #include "TransformStreamDefaultControllerBinding.h"
 #include "TrustedTypesBinding.h"
 #include "URLBinding.h"
+#include "URLPatternBinding.h"
 #include "URLSearchParamsBinding.h"
 #include "VideoColorSpaceBinding.h"
 #include "VideoDecoderBinding.h"
@@ -201,6 +239,146 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (CSSColor_Binding::ConstructorEnabled(aCx, aObj) && !CSSColor_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSColorValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSColorValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSHSL_Binding::ConstructorEnabled(aCx, aObj) && !CSSHSL_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSHWB_Binding::ConstructorEnabled(aCx, aObj) && !CSSHWB_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSImageValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSImageValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSKeywordValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSKeywordValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSLCH_Binding::ConstructorEnabled(aCx, aObj) && !CSSLCH_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSLab_Binding::ConstructorEnabled(aCx, aObj) && !CSSLab_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathClamp_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathClamp_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathInvert_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathInvert_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathMax_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathMax_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathMin_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathMin_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathNegate_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathNegate_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathProduct_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathProduct_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathSum_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathSum_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMathValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSMathValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSMatrixComponent_Binding::ConstructorEnabled(aCx, aObj) && !CSSMatrixComponent_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSNumericArray_Binding::ConstructorEnabled(aCx, aObj) && !CSSNumericArray_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSNumericValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSNumericValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSOKLCH_Binding::ConstructorEnabled(aCx, aObj) && !CSSOKLCH_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSOKLab_Binding::ConstructorEnabled(aCx, aObj) && !CSSOKLab_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSPerspective_Binding::ConstructorEnabled(aCx, aObj) && !CSSPerspective_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSRGB_Binding::ConstructorEnabled(aCx, aObj) && !CSSRGB_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSRotate_Binding::ConstructorEnabled(aCx, aObj) && !CSSRotate_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSScale_Binding::ConstructorEnabled(aCx, aObj) && !CSSScale_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSSkew_Binding::ConstructorEnabled(aCx, aObj) && !CSSSkew_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSSkewX_Binding::ConstructorEnabled(aCx, aObj) && !CSSSkewX_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSSkewY_Binding::ConstructorEnabled(aCx, aObj) && !CSSSkewY_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSStyleValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSStyleValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSTransformComponent_Binding::ConstructorEnabled(aCx, aObj) && !CSSTransformComponent_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSTransformValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSTransformValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSTranslate_Binding::ConstructorEnabled(aCx, aObj) && !CSSTranslate_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSUnitValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSUnitValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSUnparsedValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSUnparsedValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
+  if (CSSVariableReferenceValue_Binding::ConstructorEnabled(aCx, aObj) && !CSSVariableReferenceValue_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (Cache_Binding::ConstructorEnabled(aCx, aObj) && !Cache_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
@@ -513,6 +691,10 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (GPUPipelineError_Binding::ConstructorEnabled(aCx, aObj) && !GPUPipelineError_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (GPUPipelineLayout_Binding::ConstructorEnabled(aCx, aObj) && !GPUPipelineLayout_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
@@ -661,6 +843,10 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (InferenceSession_Binding::ConstructorEnabled(aCx, aObj) && !InferenceSession_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (Lock_Binding::ConstructorEnabled(aCx, aObj) && !Lock_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
@@ -893,6 +1079,10 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (StylePropertyMapReadOnly_Binding::ConstructorEnabled(aCx, aObj) && !StylePropertyMapReadOnly_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (SubtleCrypto_Binding::ConstructorEnabled(aCx, aObj) && !SubtleCrypto_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
@@ -913,6 +1103,10 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (Tensor_Binding::ConstructorEnabled(aCx, aObj) && !Tensor_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (TestUtils_Binding::ConstructorEnabled(aCx, aObj) && !TestUtils_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
@@ -973,6 +1167,10 @@ RegisterWorkerBindings(JSContext* aCx, JS::Handle<JSObject*> aObj)
     return false;
   }
 
+  if (URLPattern_Binding::ConstructorEnabled(aCx, aObj) && !URLPattern_Binding::CreateAndDefineOnGlobal(aCx)) {
+    return false;
+  }
+
   if (!URLSearchParams_Binding::CreateAndDefineOnGlobal(aCx)) {
     return false;
   }
