@@ -9,10 +9,10 @@ extern crate byteorder;
 extern crate pkcs11_bindings;
 #[macro_use]
 extern crate rsclientcerts;
-extern crate sha2;
 
 use pkcs11_bindings::*;
-use rsclientcerts::manager::{Manager, SlotType};
+use rsclientcerts::manager::Manager;
+use std::convert::TryInto;
 use std::sync::Mutex;
 
 mod backend;
@@ -52,7 +52,7 @@ macro_rules! manager_guard_to_manager {
 /// instantiating the `Manager`.
 extern "C" fn C_Initialize(_pInitArgs: CK_VOID_PTR) -> CK_RV {
     let mut manager_guard = try_to_get_manager_guard!();
-    let _unexpected_previous_manager = manager_guard.replace(Manager::new(Backend::new()));
+    let _unexpected_previous_manager = manager_guard.replace(Manager::new(vec![Backend::new()]));
     CKR_OK
 }
 
@@ -79,24 +79,19 @@ extern "C" fn C_GetInfo(pInfo: CK_INFO_PTR) -> CK_RV {
     if pInfo.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
-    let mut info = CK_INFO::default();
-    info.cryptokiVersion.major = 2;
-    info.cryptokiVersion.minor = 2;
-    info.manufacturerID = *MANUFACTURER_ID_BYTES;
-    info.libraryDescription = *LIBRARY_DESCRIPTION_BYTES;
+    let info = CK_INFO {
+        cryptokiVersion: CK_VERSION { major: 2, minor: 2 },
+        manufacturerID: *MANUFACTURER_ID_BYTES,
+        flags: 0,
+        libraryDescription: *LIBRARY_DESCRIPTION_BYTES,
+        libraryVersion: CK_VERSION::default(),
+    };
     unsafe {
         *pInfo = info;
     }
     CKR_OK
 }
 
-/// This module has two slots.
-const SLOT_COUNT: CK_ULONG = 2;
-/// The slot with ID 1 supports modern mechanisms like RSA-PSS.
-const SLOT_ID_MODERN: CK_SLOT_ID = 1;
-/// The slot with ID 2 only supports legacy mechanisms.
-const SLOT_ID_LEGACY: CK_SLOT_ID = 2;
-
 /// This gets called twice: once with a null `pSlotList` to get the number of slots (returned via
 /// `pulCount`) and a second time to get the ID for each slot.
 extern "C" fn C_GetSlotList(
@@ -107,43 +102,34 @@ extern "C" fn C_GetSlotList(
     if pulCount.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
+    let mut manager_guard = try_to_get_manager_guard!();
+    let manager = manager_guard_to_manager!(manager_guard);
+    let slot_ids = manager.get_slot_ids();
+    let slot_count: CK_ULONG = slot_ids.len().try_into().unwrap();
     if !pSlotList.is_null() {
-        if unsafe { *pulCount } < SLOT_COUNT {
+        if unsafe { *pulCount } < slot_count {
             return CKR_BUFFER_TOO_SMALL;
         }
         unsafe {
-            *pSlotList = SLOT_ID_MODERN;
-            *pSlotList.offset(1) = SLOT_ID_LEGACY;
+            std::ptr::copy_nonoverlapping(slot_ids.as_ptr(), pSlotList, slot_ids.len());
         }
     };
     unsafe {
-        *pulCount = SLOT_COUNT;
+        *pulCount = slot_count;
     }
     CKR_OK
 }
 
-const SLOT_DESCRIPTION_MODERN_BYTES: &[u8; 64] =
-    b"IPC Client Cert Slot (Modern)                                   ";
-const SLOT_DESCRIPTION_LEGACY_BYTES: &[u8; 64] =
-    b"IPC Client Cert Slot (Legacy)                                   ";
-
 /// This gets called to obtain information about slots. In this implementation, the tokens are
 /// always present in the slots.
 extern "C" fn C_GetSlotInfo(slotID: CK_SLOT_ID, pInfo: CK_SLOT_INFO_PTR) -> CK_RV {
-    if (slotID != SLOT_ID_MODERN && slotID != SLOT_ID_LEGACY) || pInfo.is_null() {
+    if pInfo.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
-    let description = if slotID == SLOT_ID_MODERN {
-        SLOT_DESCRIPTION_MODERN_BYTES
-    } else {
-        SLOT_DESCRIPTION_LEGACY_BYTES
-    };
-    let slot_info = CK_SLOT_INFO {
-        slotDescription: *description,
-        manufacturerID: *MANUFACTURER_ID_BYTES,
-        flags: CKF_TOKEN_PRESENT,
-        hardwareVersion: CK_VERSION::default(),
-        firmwareVersion: CK_VERSION::default(),
+    let mut manager_guard = try_to_get_manager_guard!();
+    let manager = manager_guard_to_manager!(manager_guard);
+    let Ok(slot_info) = manager.get_slot_info(slotID) else {
+        return CKR_ARGUMENTS_BAD;
     };
     unsafe {
         *pInfo = slot_info;
@@ -151,60 +137,47 @@ extern "C" fn C_GetSlotInfo(slotID: CK_SLOT_ID, pInfo: CK_SLOT_INFO_PTR) -> CK_R
     CKR_OK
 }
 
-const TOKEN_LABEL_MODERN_BYTES: &[u8; 32] = b"IPC Client Cert Token (Modern)  ";
-const TOKEN_LABEL_LEGACY_BYTES: &[u8; 32] = b"IPC Client Cert Token (Legacy)  ";
-const TOKEN_MODEL_BYTES: &[u8; 16] = b"ipcclientcerts  ";
-const TOKEN_SERIAL_NUMBER_BYTES: &[u8; 16] = b"0000000000000000";
-
 /// This gets called to obtain some information about tokens. This implementation has two slots,
 /// so it has two tokens. This information is primarily for display purposes.
 extern "C" fn C_GetTokenInfo(slotID: CK_SLOT_ID, pInfo: CK_TOKEN_INFO_PTR) -> CK_RV {
-    if (slotID != SLOT_ID_MODERN && slotID != SLOT_ID_LEGACY) || pInfo.is_null() {
+    if pInfo.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
-    let mut token_info = CK_TOKEN_INFO::default();
-    let label = if slotID == SLOT_ID_MODERN {
-        TOKEN_LABEL_MODERN_BYTES
-    } else {
-        TOKEN_LABEL_LEGACY_BYTES
+    let mut manager_guard = try_to_get_manager_guard!();
+    let manager = manager_guard_to_manager!(manager_guard);
+    let Ok(token_info) = manager.get_token_info(slotID) else {
+        return CKR_ARGUMENTS_BAD;
     };
-    token_info.label = *label;
-    token_info.manufacturerID = *MANUFACTURER_ID_BYTES;
-    token_info.model = *TOKEN_MODEL_BYTES;
-    token_info.serialNumber = *TOKEN_SERIAL_NUMBER_BYTES;
     unsafe {
         *pInfo = token_info;
     }
     CKR_OK
 }
 
-/// This gets called to determine what mechanisms a slot supports. The modern slot supports ECDSA,
-/// RSA PKCS, and RSA PSS. The legacy slot only supports RSA PKCS.
 extern "C" fn C_GetMechanismList(
     slotID: CK_SLOT_ID,
     pMechanismList: CK_MECHANISM_TYPE_PTR,
     pulCount: CK_ULONG_PTR,
 ) -> CK_RV {
-    if (slotID != SLOT_ID_MODERN && slotID != SLOT_ID_LEGACY) || pulCount.is_null() {
+    if pulCount.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
-    let mechanisms = if slotID == SLOT_ID_MODERN {
-        vec![CKM_ECDSA, CKM_RSA_PKCS, CKM_RSA_PKCS_PSS]
-    } else {
-        vec![CKM_RSA_PKCS]
+    let mut manager_guard = try_to_get_manager_guard!();
+    let manager = manager_guard_to_manager!(manager_guard);
+    let Ok(mechanisms) = manager.get_mechanism_list(slotID) else {
+        return CKR_ARGUMENTS_BAD;
     };
+    let mechanisms_len: CK_ULONG = mechanisms.len().try_into().unwrap();
     if !pMechanismList.is_null() {
-        if unsafe { *pulCount as usize } < mechanisms.len() {
+        if unsafe { *pulCount } < mechanisms_len {
             return CKR_ARGUMENTS_BAD;
         }
-        for i in 0..mechanisms.len() {
-            unsafe {
-                *pMechanismList.offset(i as isize) = mechanisms[i];
-            }
+        unsafe {
+            std::ptr::copy_nonoverlapping(mechanisms.as_ptr(), pMechanismList, mechanisms.len());
         }
     }
     unsafe {
-        *pulCount = mechanisms.len() as CK_ULONG;
+        *pulCount = mechanisms_len;
     }
     CKR_OK
 }
@@ -253,17 +226,12 @@ extern "C" fn C_OpenSession(
     _Notify: CK_NOTIFY,
     phSession: CK_SESSION_HANDLE_PTR,
 ) -> CK_RV {
-    if (slotID != SLOT_ID_MODERN && slotID != SLOT_ID_LEGACY) || phSession.is_null() {
+    if phSession.is_null() {
         return CKR_ARGUMENTS_BAD;
     }
     let mut manager_guard = try_to_get_manager_guard!();
     let manager = manager_guard_to_manager!(manager_guard);
-    let slot_type = if slotID == SLOT_ID_MODERN {
-        SlotType::Modern
-    } else {
-        SlotType::Legacy
-    };
-    let session_handle = match manager.open_session(slot_type) {
+    let session_handle = match manager.open_session(slotID) {
         Ok(session_handle) => session_handle,
         Err(_) => return CKR_DEVICE_ERROR,
     };
@@ -285,17 +253,9 @@ extern "C" fn C_CloseSession(hSession: CK_SESSION_HANDLE) -> CK_RV {
 
 /// This gets called to close all open sessions at once. This is handled by the `ManagerProxy`.
 extern "C" fn C_CloseAllSessions(slotID: CK_SLOT_ID) -> CK_RV {
-    if slotID != SLOT_ID_MODERN && slotID != SLOT_ID_LEGACY {
-        return CKR_ARGUMENTS_BAD;
-    }
     let mut manager_guard = try_to_get_manager_guard!();
     let manager = manager_guard_to_manager!(manager_guard);
-    let slot_type = if slotID == SLOT_ID_MODERN {
-        SlotType::Modern
-    } else {
-        SlotType::Legacy
-    };
-    match manager.close_all_sessions(slot_type) {
+    match manager.close_all_sessions(slotID) {
         Ok(()) => CKR_OK,
         Err(_) => CKR_DEVICE_ERROR,
     }
@@ -377,7 +337,7 @@ extern "C" fn C_GetObjectSize(
 /// This gets called twice: once to obtain the lengths of the attributes and again to get the
 /// values.
 extern "C" fn C_GetAttributeValue(
-    _hSession: CK_SESSION_HANDLE,
+    hSession: CK_SESSION_HANDLE,
     hObject: CK_OBJECT_HANDLE,
     pTemplate: CK_ATTRIBUTE_PTR,
     ulCount: CK_ULONG,
@@ -392,7 +352,7 @@ extern "C" fn C_GetAttributeValue(
     }
     let mut manager_guard = try_to_get_manager_guard!();
     let manager = manager_guard_to_manager!(manager_guard);
-    let values = match manager.get_attributes(hObject, attr_types) {
+    let values = match manager.get_attributes(hSession, hObject, attr_types) {
         Ok(values) => values,
         Err(_) => return CKR_ARGUMENTS_BAD,
     };
