@@ -23,6 +23,38 @@ pub enum InlineBaseDirection {
     RightToLeft,
 }
 
+/// The writing-mode property (different from the WritingMode enum).
+/// https://drafts.csswg.org/css-writing-modes/#block-flow
+/// Aliases come from https://drafts.csswg.org/css-writing-modes-4/#svg-writing-mode
+#[allow(missing_docs)]
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Eq,
+    FromPrimitive,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(u8)]
+pub enum WritingModeProperty {
+    #[parse(aliases = "lr,lr-tb,rl,rl-tb")]
+    HorizontalTb,
+    #[parse(aliases = "tb,tb-rl")]
+    VerticalRl,
+    VerticalLr,
+    #[cfg(feature = "gecko")]
+    SidewaysRl,
+    #[cfg(feature = "gecko")]
+    SidewaysLr,
+}
+
 // TODO: improve the readability of the WritingMode serialization, refer to the Debug:fmt()
 #[derive(Clone, Copy, Debug, Eq, MallocSizeOf, PartialEq, Serialize)]
 #[repr(C)]
@@ -75,6 +107,23 @@ bitflags!(
         ///
         /// Never set without VERTICAL.
         const UPRIGHT = 1 << 7;
+        /// Writing mode combinations that can be specified in CSS.
+        ///
+        /// * writing-mode: horizontal-tb;
+        const WRITING_MODE_HORIZONTAL_TB = 0;
+        /// * writing-mode: vertical_rl;
+        const WRITING_MODE_VERTICAL_RL = WritingMode::VERTICAL.bits();
+        /// * writing-mode: vertcail-lr;
+        const WRITING_MODE_VERTICAL_LR = WritingMode::VERTICAL.bits() |
+                                         WritingMode::VERTICAL_LR.bits() |
+                                         WritingMode::LINE_INVERTED.bits();
+        /// * writing-mode: sideways-rl;
+        const WRITING_MODE_SIDEWAYS_RL = WritingMode::VERTICAL.bits() |
+                                         WritingMode::VERTICAL_SIDEWAYS.bits();
+        /// * writing-mode: sideways-lr;
+        const WRITING_MODE_SIDEWAYS_LR = WritingMode::VERTICAL.bits() |
+                                         WritingMode::VERTICAL_LR.bits() |
+                                         WritingMode::VERTICAL_SIDEWAYS.bits();
     }
 );
 
@@ -82,7 +131,6 @@ impl WritingMode {
     /// Return a WritingMode bitflags from the relevant CSS properties.
     pub fn new(inheritedbox_style: &style_structs::InheritedBox) -> Self {
         use crate::properties::longhands::direction::computed_value::T as Direction;
-        use crate::properties::longhands::writing_mode::computed_value::T as SpecifiedWritingMode;
 
         let mut flags = WritingMode::empty();
 
@@ -97,38 +145,33 @@ impl WritingMode {
         }
 
         match writing_mode {
-            SpecifiedWritingMode::HorizontalTb => {
+            WritingModeProperty::HorizontalTb => {
                 if direction == Direction::Rtl {
                     flags.insert(WritingMode::INLINE_REVERSED);
                 }
             },
-            SpecifiedWritingMode::VerticalRl => {
-                flags.insert(WritingMode::VERTICAL);
+            WritingModeProperty::VerticalRl => {
+                flags.insert(WritingMode::WRITING_MODE_VERTICAL_RL);
                 if direction == Direction::Rtl {
                     flags.insert(WritingMode::INLINE_REVERSED);
                 }
             },
-            SpecifiedWritingMode::VerticalLr => {
-                flags.insert(WritingMode::VERTICAL);
-                flags.insert(WritingMode::VERTICAL_LR);
-                flags.insert(WritingMode::LINE_INVERTED);
+            WritingModeProperty::VerticalLr => {
+                flags.insert(WritingMode::WRITING_MODE_VERTICAL_LR);
                 if direction == Direction::Rtl {
                     flags.insert(WritingMode::INLINE_REVERSED);
                 }
             },
             #[cfg(feature = "gecko")]
-            SpecifiedWritingMode::SidewaysRl => {
-                flags.insert(WritingMode::VERTICAL);
-                flags.insert(WritingMode::VERTICAL_SIDEWAYS);
+            WritingModeProperty::SidewaysRl => {
+                flags.insert(WritingMode::WRITING_MODE_SIDEWAYS_RL);
                 if direction == Direction::Rtl {
                     flags.insert(WritingMode::INLINE_REVERSED);
                 }
             },
             #[cfg(feature = "gecko")]
-            SpecifiedWritingMode::SidewaysLr => {
-                flags.insert(WritingMode::VERTICAL);
-                flags.insert(WritingMode::VERTICAL_LR);
-                flags.insert(WritingMode::VERTICAL_SIDEWAYS);
+            WritingModeProperty::SidewaysLr => {
+                flags.insert(WritingMode::WRITING_MODE_SIDEWAYS_LR);
                 if direction == Direction::Ltr {
                     flags.insert(WritingMode::INLINE_REVERSED);
                 }
@@ -142,7 +185,7 @@ impl WritingMode {
             // text-orientation only has an effect for vertical-rl and
             // vertical-lr values of writing-mode.
             match writing_mode {
-                SpecifiedWritingMode::VerticalRl | SpecifiedWritingMode::VerticalLr => {
+                WritingModeProperty::VerticalRl | WritingModeProperty::VerticalLr => {
                     match inheritedbox_style.clone_text_orientation() {
                         TextOrientation::Mixed => {},
                         TextOrientation::Upright => {
