@@ -10,21 +10,24 @@
 #include <wayland-egl.h>
 #include "nsGtkUtils.h"
 #include "mozilla/StaticPrefs_widget.h"
+#include "mozilla/ToString.h"
 #include <dlfcn.h>
 #include <fcntl.h>
 #include "ScreenHelperGTK.h"
 #include "DMABufFormats.h"
+#include "mozilla/gfx/gfxVars.h"
+#ifdef MOZ_LOGGING
+#  include "EncoderConfig.h"
+#endif
 
 #undef LOG
 #ifdef MOZ_LOGGING
 #  include "mozilla/Logging.h"
 #  include "nsTArray.h"
 #  include "Units.h"
-#  include "nsWindow.h"
 #  undef LOGWAYLAND
 #  undef LOGVERBOSE
 extern mozilla::LazyLogModule gWidgetWaylandLog;
-extern mozilla::LazyLogModule gWidgetLog;
 #  define LOGWAYLAND(str, ...)                           \
     MOZ_LOG(gWidgetWaylandLog, mozilla::LogLevel::Debug, \
             ("%s: " str, GetDebugTag().get(), ##__VA_ARGS__))
@@ -90,7 +93,7 @@ WaylandSurface::~WaylandSurface() {
   LOGWAYLAND("WaylandSurface::~WaylandSurface()");
   MOZ_RELEASE_ASSERT(!mIsMapped, "We can't release mapped WaylandSurface!");
   MOZ_RELEASE_ASSERT(!mSurfaceLock, "We can't release locked WaylandSurface!");
-  MOZ_RELEASE_ASSERT(mAttachedBuffers.Length() == 0,
+  MOZ_RELEASE_ASSERT(mBufferTransactions.Length() == 0,
                      "We can't release surface with buffers tracked!");
   MOZ_RELEASE_ASSERT(!mEmulatedFrameCallbackTimerID,
                      "We can't release WaylandSurface with active timer");
@@ -105,19 +108,20 @@ WaylandSurface::~WaylandSurface() {
                      "We can't release WaylandSurface with numap callback!");
 }
 
-void WaylandSurface::InitialFrameCallbackHandler(struct wl_callback* callback) {
+void WaylandSurface::ReadyToDrawFrameCallbackHandler(
+    struct wl_callback* callback) {
   LOGWAYLAND(
-      "WaylandSurface::InitialFrameCallbackHandler() "
+      "WaylandSurface::ReadyToDrawFrameCallbackHandler() "
       "mReadyToDrawFrameCallback %p mIsReadyToDraw %d initial_draw callback "
       "%zd\n",
       (void*)mReadyToDrawFrameCallback, (bool)mIsReadyToDraw,
-      mReadToDrawCallbacks.size());
+      mReadyToDrawCallbacks.size());
 
   // We're supposed to run on main thread only.
   AssertIsOnMainThread();
 
   // mReadyToDrawFrameCallback/callback can be nullptr when redering directly
-  // to GtkWidget and InitialFrameCallbackHandler is called by us from main
+  // to GtkWidget and ReadyToDrawFrameCallbackHandler is called by us from main
   // thread by WaylandSurface::Map().
   MOZ_RELEASE_ASSERT(mReadyToDrawFrameCallback == callback);
 
@@ -128,9 +132,9 @@ void WaylandSurface::InitialFrameCallbackHandler(struct wl_callback* callback) {
     // It's possible that we're already unmapped so quit in such case.
     if (!mSurface) {
       LOGWAYLAND("  WaylandSurface is unmapped, quit.");
-      if (!mReadToDrawCallbacks.empty()) {
+      if (!mReadyToDrawCallbacks.empty()) {
         NS_WARNING("Unmapping WaylandSurface with active draw callback!");
-        mReadToDrawCallbacks.clear();
+        mReadyToDrawCallbacks.clear();
       }
       return;
     }
@@ -138,7 +142,7 @@ void WaylandSurface::InitialFrameCallbackHandler(struct wl_callback* callback) {
       return;
     }
     mIsReadyToDraw = true;
-    cbs = std::move(mReadToDrawCallbacks);
+    cbs = std::move(mReadyToDrawCallbacks);
 
     RequestFrameCallbackLocked(lock);
   }
@@ -153,15 +157,16 @@ void WaylandSurface::InitialFrameCallbackHandler(struct wl_callback* callback) {
   }
 }
 
-static void InitialFrameCallbackHandler(void* aWaylandSurface,
-                                        struct wl_callback* callback,
-                                        uint32_t time) {
+static void ReadyToDrawFrameCallbackHandler(void* aWaylandSurface,
+                                            struct wl_callback* callback,
+                                            uint32_t time) {
   auto* waylandSurface = static_cast<WaylandSurface*>(aWaylandSurface);
-  waylandSurface->InitialFrameCallbackHandler(callback);
+  waylandSurface->ReadyToDrawFrameCallbackHandler(callback);
 }
 
-static const struct wl_callback_listener sWaylandSurfaceInitialFrameListener = {
-    ::InitialFrameCallbackHandler};
+static const struct wl_callback_listener
+    sWaylandSurfaceReadyToDrawFrameListener = {
+        ::ReadyToDrawFrameCallbackHandler};
 
 void WaylandSurface::AddReadyToDrawCallbackLocked(
     const WaylandSurfaceLock& aProofOfLock,
@@ -175,7 +180,7 @@ void WaylandSurface::AddReadyToDrawCallbackLocked(
         " ready to draw without wayland surface!");
   }
   MOZ_DIAGNOSTIC_ASSERT(!mIsReadyToDraw || !mSurface);
-  mReadToDrawCallbacks.push_back(aDrawCB);
+  mReadyToDrawCallbacks.push_back(aDrawCB);
 }
 
 void WaylandSurface::AddOrFireReadyToDrawCallback(
@@ -190,7 +195,7 @@ void WaylandSurface::AddOrFireReadyToDrawCallback(
     if (!mIsReadyToDraw || !mSurface) {
       LOGVERBOSE(
           "WaylandSurface::AddOrFireReadyToDrawCallback() callback stored");
-      mReadToDrawCallbacks.push_back(aDrawCB);
+      mReadyToDrawCallbacks.push_back(aDrawCB);
       return;
     }
   }
@@ -205,7 +210,7 @@ void WaylandSurface::ClearReadyToDrawCallbacksLocked(
     const WaylandSurfaceLock& aProofOfLock) {
   MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
   MozClearPointer(mReadyToDrawFrameCallback, wl_callback_destroy);
-  mReadToDrawCallbacks.clear();
+  mReadyToDrawCallbacks.clear();
 }
 
 void WaylandSurface::ClearReadyToDrawCallbacks() {
@@ -215,12 +220,7 @@ void WaylandSurface::ClearReadyToDrawCallbacks() {
 
 bool WaylandSurface::HasEmulatedFrameCallbackLocked(
     const WaylandSurfaceLock& aProofOfLock) const {
-  for (auto const& cb : mPersistentFrameCallbackHandlers) {
-    if (cb.mEmulated) {
-      return true;
-    }
-  }
-  return false;
+  return mFrameCallbackHandler.IsSet() && mFrameCallbackHandler.mEmulated;
 }
 
 void WaylandSurface::FrameCallbackHandler(struct wl_callback* aCallback,
@@ -231,7 +231,7 @@ void WaylandSurface::FrameCallbackHandler(struct wl_callback* aCallback,
 
   bool emulatedCallback = !aCallback && !aTime;
 
-  std::vector<FrameCallback> cbs;
+  FrameCallback cb;
   {
     WaylandSurfaceLock lock(this);
 
@@ -241,10 +241,9 @@ void WaylandSurface::FrameCallbackHandler(struct wl_callback* aCallback,
     }
 
     LOGVERBOSE(
-        "WaylandSurface::FrameCallbackHandler() one-time %zd "
-        "persistent %zd emulated %d routed %d",
-        mOneTimeFrameCallbackHandlers.size(),
-        mPersistentFrameCallbackHandlers.size(), emulatedCallback,
+        "WaylandSurface::FrameCallbackHandler() "
+        "set %d emulated %d routed %d",
+        mFrameCallbackHandler.IsSet(), emulatedCallback,
         aRoutedFromChildSurface);
 
     // It's possible to get regular frame callback right after unmap
@@ -266,28 +265,19 @@ void WaylandSurface::FrameCallbackHandler(struct wl_callback* aCallback,
       mBufferAttached = true;
     }
 
-    // We don't support emulated one-time callbacks
-    if (!emulatedCallback) {
-      cbs = std::move(mOneTimeFrameCallbackHandlers);
-    }
-    std::copy(mPersistentFrameCallbackHandlers.begin(),
-              mPersistentFrameCallbackHandlers.end(), back_inserter(cbs));
+    cb = mFrameCallbackHandler;
 
     // Fire frame callback again if there's any pending frame callback
     RequestFrameCallbackLocked(lock);
   }
 
   // We can't run the callbacks under WaylandSurfaceLock
-#ifdef MOZ_LOGGING
-  int callbackNum = 0;
-#endif
-  for (auto const& cb : cbs) {
-    LOGVERBOSE("  frame callback fire [%d]", callbackNum++);
-    const auto& [callback, runEmulated] = cb;
-    if (emulatedCallback && !runEmulated) {
-      continue;
-    }
-    callback(aCallback, aTime);
+  LOGVERBOSE("  frame callback fire");
+  if (emulatedCallback && !cb.mEmulated) {
+    return;
+  }
+  if (cb.IsSet()) {
+    cb.mCb(aCallback, aTime);
   }
 }
 
@@ -312,16 +302,7 @@ void WaylandSurface::RequestFrameCallbackLocked(
   MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
 
   // Frame callback will be added by Map.
-  if (!mIsMapped) {
-    return;
-  }
-
-  if (!mFrameCallbackEnabled) {
-    return;
-  }
-
-  if (mPersistentFrameCallbackHandlers.empty() &&
-      mOneTimeFrameCallbackHandlers.empty()) {
+  if (!mIsMapped || !mFrameCallbackEnabled || !mFrameCallbackHandler.IsSet()) {
     return;
   }
 
@@ -381,33 +362,28 @@ void WaylandSurface::ClearFrameCallbackLocked(
   MozClearPointer(mFrameCallback, wl_callback_destroy);
 }
 
-void WaylandSurface::AddOneTimeFrameCallbackLocked(
-    const WaylandSurfaceLock& aProofOfLock,
-    const std::function<void(wl_callback*, uint32_t)>& aFrameCallbackHandler) {
+void WaylandSurface::ClearFrameCallbackHandlerLocked(
+    const WaylandSurfaceLock& aProofOfLock) {
   MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
-  LOGWAYLAND("WaylandSurface::AddOneTimeFrameCallbackLocked()");
-
-  mOneTimeFrameCallbackHandlers.push_back(
-      FrameCallback{aFrameCallbackHandler, false});
-  RequestFrameCallbackLocked(aProofOfLock);
+  mFrameCallbackHandler = FrameCallback{};
 }
 
-void WaylandSurface::AddPersistentFrameCallbackLocked(
+void WaylandSurface::SetFrameCallbackLocked(
     const WaylandSurfaceLock& aProofOfLock,
     const std::function<void(wl_callback*, uint32_t)>& aFrameCallbackHandler,
     bool aEmulateFrameCallback) {
   MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
-  LOGWAYLAND("WaylandSurface::AddPersistentFrameCallbackLocked()");
 
-  mPersistentFrameCallbackHandlers.push_back(
-      FrameCallback{aFrameCallbackHandler, aEmulateFrameCallback});
+  LOGWAYLAND("WaylandSurface::SetFrameCallbackLocked()");
+
+  mFrameCallbackHandler =
+      FrameCallback{aFrameCallbackHandler, aEmulateFrameCallback};
   RequestFrameCallbackLocked(aProofOfLock);
 }
 
-void WaylandSurface::SetFrameCallbackState(bool aEnabled) {
+void WaylandSurface::SetFrameCallbackStateLocked(
+    const WaylandSurfaceLock& aProofOfLock, bool aEnabled) {
   LOGWAYLAND("WaylandSurface::SetFrameCallbackState() state %d", aEnabled);
-
-  WaylandSurfaceLock lock(this);
   if (mFrameCallbackEnabled == aEnabled) {
     return;
   }
@@ -415,10 +391,20 @@ void WaylandSurface::SetFrameCallbackState(bool aEnabled) {
 
   // If there's any frame callback waiting, register the handler.
   if (mFrameCallbackEnabled) {
-    RequestFrameCallbackLocked(lock);
+    RequestFrameCallbackLocked(aProofOfLock);
   } else {
-    ClearFrameCallbackLocked(lock);
+    ClearFrameCallbackLocked(aProofOfLock);
   }
+  if (mFrameCallbackStateHandler) {
+    mFrameCallbackStateHandler(aEnabled);
+  }
+}
+
+void WaylandSurface::SetFrameCallbackStateHandlerLocked(
+    const WaylandSurfaceLock& aProofOfLock,
+    const std::function<void(bool)>& aFrameCallbackStateHandler) {
+  MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
+  mFrameCallbackStateHandler = aFrameCallbackStateHandler;
 }
 
 bool WaylandSurface::CreateViewportLocked(
@@ -447,6 +433,12 @@ bool WaylandSurface::CreateViewportLocked(
 void WaylandSurface::EnableDMABufFormatsLocked(
     const WaylandSurfaceLock& aProofOfLock,
     const std::function<void(DMABufFormats*)>& aFormatRefreshCB) {
+  // Ignore DMABuf feedback requests if we export dmabuf surfaces
+  // directly from EGLImage.
+  if (gfx::gfxVars::UseDMABufSurfaceExport()) {
+    return;
+  }
+
   mUseDMABufFormats = true;
   mDMABufFormatRefreshCallback = aFormatRefreshCB;
 
@@ -496,6 +488,8 @@ bool WaylandSurface::MapLocked(const WaylandSurfaceLock& aProofOfLock,
 
   // Created wl_surface is without buffer attached
   mBufferAttached = false;
+  mLatestAttachedBuffer = 0;
+
   struct wl_compositor* compositor = WaylandDisplayGet()->GetCompositor();
   mSurface = wl_compositor_create_surface(compositor);
   if (!mSurface) {
@@ -510,6 +504,7 @@ bool WaylandSurface::MapLocked(const WaylandSurfaceLock& aProofOfLock,
     LOGWAYLAND("    Failed - can't create sub-surface!");
     return false;
   }
+  mSubsurfaceDesync = aSubsurfaceDesync;
   if (aSubsurfaceDesync) {
     wl_subsurface_set_desync(mSubsurface);
   }
@@ -519,7 +514,7 @@ bool WaylandSurface::MapLocked(const WaylandSurfaceLock& aProofOfLock,
   if (aUseReadyToDrawCallback) {
     mReadyToDrawFrameCallback = wl_surface_frame(mParentSurface);
     wl_callback_add_listener(mReadyToDrawFrameCallback,
-                             &sWaylandSurfaceInitialFrameListener, this);
+                             &sWaylandSurfaceReadyToDrawFrameListener, this);
     LOGWAYLAND("    created ready to draw frame callback ID %d\n",
                wl_proxy_get_id((struct wl_proxy*)mReadyToDrawFrameCallback));
   }
@@ -593,6 +588,18 @@ void WaylandSurface::GdkCleanUpLocked(const WaylandSurfaceLock& aProofOfLock) {
   }
 }
 
+void WaylandSurface::ReleaseAllWaylandTransactionsLocked(
+    WaylandSurfaceLock& aSurfaceLock) {
+  LOGWAYLAND("WaylandSurface::ReleaseAllWaylandTransactionsLocked(), num %d",
+             (int)mBufferTransactions.Length());
+  MOZ_DIAGNOSTIC_ASSERT(!mIsMapped);
+  auto transactions = std::move(mBufferTransactions);
+  MOZ_ASSERT(mBufferTransactions.IsEmpty());
+  for (auto& transaction : transactions) {
+    transaction->DeleteTransactionLocked(aSurfaceLock);
+  }
+}
+
 void WaylandSurface::UnmapLocked(WaylandSurfaceLock& aSurfaceLock) {
   if (!mIsMapped) {
     return;
@@ -615,11 +622,14 @@ void WaylandSurface::UnmapLocked(WaylandSurfaceLock& aSurfaceLock) {
   MozClearPointer(mFractionalScaleListener, wp_fractional_scale_v1_destroy);
   MozClearPointer(mSubsurface, wl_subsurface_destroy);
   MozClearPointer(mColorSurface, wp_color_management_surface_v1_destroy);
+  MozClearPointer(mColorRepresentationSurface,
+                  wp_color_representation_surface_v1_destroy);
   MozClearPointer(mImageDescription, wp_image_description_v1_destroy);
   mParentSurface = nullptr;
   mFormats = nullptr;
 
-  // We can't release mSurface if it's used by Gdk for frame callback routing.
+  // We can't release mSurface if it's used by Gdk for frame callback routing,
+  // it will be released at GdkCleanUpLocked().
   if (!mIsPendingGdkCleanup) {
     MozClearPointer(mSurface, wl_surface_destroy);
   }
@@ -629,7 +639,7 @@ void WaylandSurface::UnmapLocked(WaylandSurfaceLock& aSurfaceLock) {
 
   // Remove references to WaylandBuffers attached to mSurface,
   // we don't want to get any buffer release callback when we're unmapped.
-  ReleaseAllWaylandBuffersLocked(aSurfaceLock);
+  ReleaseAllWaylandTransactionsLocked(aSurfaceLock);
 }
 
 void WaylandSurface::Commit(WaylandSurfaceLock* aProofOfLock, bool aForceCommit,
@@ -639,8 +649,17 @@ void WaylandSurface::Commit(WaylandSurfaceLock* aProofOfLock, bool aForceCommit,
   // mSurface may be already deleted, see WaylandSurface::Unmap();
   if (mSurface && (aForceCommit || mSurfaceNeedsCommit)) {
     LOGVERBOSE(
-        "WaylandSurface::Commit() needs commit %d, force commit %d flush %d",
-        mSurfaceNeedsCommit, aForceCommit, aForceDisplayFlush);
+        "WaylandSurface::Commit() allowed [%d] needs commit %d, force commit "
+        "%d flush %d",
+        mCommitAllowed, !!mSurfaceNeedsCommit, aForceCommit,
+        aForceDisplayFlush);
+    if (!mCommitAllowed) {
+      return;
+    }
+    if (!mSubsurfaceDesync && mParent) {
+      LOGVERBOSE("  request force commit to parent [%p]", mParent.get());
+      mParent->ForceCommit();
+    }
     mSurfaceNeedsCommit = false;
     wl_surface_commit(mSurface);
     if (aForceDisplayFlush) {
@@ -846,21 +865,26 @@ void WaylandSurface::SetSizeLocked(const WaylandSurfaceLock& aProofOfLock,
 void WaylandSurface::SetViewPortDestLocked(
     const WaylandSurfaceLock& aProofOfLock, gfx::IntSize aDestSize) {
   MOZ_DIAGNOSTIC_ASSERT(&aProofOfLock == mSurfaceLock);
-  if (mViewport) {
-    if (mViewportDestinationSize == aDestSize) {
-      return;
-    }
-    LOGWAYLAND("WaylandSurface::SetViewPortDestLocked(): Size [%d x %d]",
-               aDestSize.width, aDestSize.height);
-    if (aDestSize.width < 1 || aDestSize.height < 1) {
-      NS_WARNING("WaylandSurface::SetViewPortDestLocked(): Wrong coordinates!");
-      aDestSize.width = aDestSize.height = -1;
-    }
-    mViewportDestinationSize = aDestSize;
-    wp_viewport_set_destination(mViewport, mViewportDestinationSize.width,
-                                mViewportDestinationSize.height);
-    mSurfaceNeedsCommit = true;
+  if (!mViewport) {
+    return;
   }
+  if (mViewportDestinationSize == aDestSize) {
+    return;
+  }
+  LOGWAYLAND("WaylandSurface::SetViewPortDestLocked(): Size [%d x %d]",
+             aDestSize.width, aDestSize.height);
+  if (aDestSize.width < 1 || aDestSize.height < 1) {
+    NS_WARNING(
+        nsPrintfCString(
+            "WaylandSurface::SetViewPortDestLocked(%s): Wrong coordinates!",
+            ToString(aDestSize).c_str())
+            .get());
+    aDestSize.width = aDestSize.height = -1;
+  }
+  mViewportDestinationSize = aDestSize;
+  wp_viewport_set_destination(mViewport, mViewportDestinationSize.width,
+                              mViewportDestinationSize.height);
+  mSurfaceNeedsCommit = true;
 }
 
 void WaylandSurface::SetViewPortSourceRectLocked(
@@ -877,8 +901,10 @@ void WaylandSurface::SetViewPortSourceRectLocked(
 
   // Don't throw protocol error with bad coords
   if (aRect.x < 0 || aRect.y < 0 || aRect.width < 1 || aRect.height < 1) {
-    NS_WARNING(
-        "WaylandSurface::SetViewPortSourceRectLocked(): Wrong coordinates!");
+    NS_WARNING(nsPrintfCString("WaylandSurface::SetViewPortSourceRectLocked(%s)"
+                               ": Wrong coordinates!",
+                               ToString(aRect).c_str())
+                   .get());
     aRect.x = aRect.y = aRect.width = aRect.height = -1;
   }
 
@@ -895,6 +921,7 @@ wl_surface* WaylandSurface::Lock(WaylandSurfaceLock* aWaylandSurfaceLock)
     // which we want.
     MOZ_NO_THREAD_SAFETY_ANALYSIS {
   mMutex.Lock();
+  MOZ_DIAGNOSTIC_ASSERT(!mSurfaceLock);
   mSurfaceLock = aWaylandSurfaceLock;
   return mIsReadyToDraw ? mSurface : nullptr;
 }
@@ -1069,113 +1096,100 @@ void WaylandSurface::InvalidateLocked(const WaylandSurfaceLock& aProofOfLock) {
   mSurfaceNeedsCommit = true;
 }
 
-void WaylandSurface::ReleaseAllWaylandBuffersLocked(
-    WaylandSurfaceLock& aSurfaceLock) {
-  LOGWAYLAND("WaylandSurface::ReleaseAllWaylandBuffersLocked(), buffers num %d",
-             (int)mAttachedBuffers.Length());
-  MOZ_DIAGNOSTIC_ASSERT(!mIsMapped);
-  for (auto& buffer : mAttachedBuffers) {
-    buffer->ReturnBufferAttached(aSurfaceLock);
+void WaylandSurface::RemoveTransactionLocked(
+    const WaylandSurfaceLock& aProofOfLock,
+    RefPtr<BufferTransaction> aTransaction) {
+  // If we're called from
+  // ReleaseAllWaylandTransactionsLocked -> DeleteTransactionLocked
+  // then mBufferTransactions is empty.
+  if (mBufferTransactions.IsEmpty()) {
+    return;
   }
+  LOGWAYLAND("WaylandSurface::RemoveTransactionLocked() [%p]",
+             (void*)aTransaction);
+  MOZ_DIAGNOSTIC_ASSERT(aTransaction->IsDeleted());
+  [[maybe_unused]] bool removed =
+      mBufferTransactions.RemoveElement(aTransaction);
+  MOZ_DIAGNOSTIC_ASSERT(!mBufferTransactions.Contains(aTransaction));
 }
 
-// BufferFreeCallbackHandler is called when WaylandBuffer is detached by
-// compositor or we delete it explicitly. The two events can happen in any
-// order.
-void WaylandSurface::BufferFreeCallbackHandler(uintptr_t aWlBufferID,
-                                               bool aWlBufferDelete) {
-  LOGWAYLAND("WaylandSurface::BufferFreeCallbackHandler() wl_buffer [%" PRIxPTR
-             "] buffer %s",
-             aWlBufferID, aWlBufferDelete ? "delete" : "detach");
-  WaylandSurfaceLock lock(this);
+BufferTransaction* WaylandSurface::GetNextTransactionLocked(
+    const WaylandSurfaceLock& aProofOfLock, WaylandBuffer* aBuffer) {
+  auto* nextTransaction = aBuffer->GetTransaction();
+  if (!nextTransaction) {
+    return nullptr;
+  }
 
-  // BufferFreeCallbackHandler() should be caled by Wayland compostor
-  // on main thread only.
-  AssertIsOnMainThread();
-
-  for (size_t i = 0; i < mAttachedBuffers.Length(); i++) {
-    if (mAttachedBuffers[i]->Matches(aWlBufferID)) {
-      mAttachedBuffers[i]->ReturnBufferDetached(lock);
-      mAttachedBuffers.RemoveElementAt(i);
-      return;
+  auto transactions = std::move(mBufferTransactions);
+  bool addedNext = false;
+  // DeleteTransactionLocked() may delete BufferTransaction so
+  // iterate with ref taken.
+  for (auto t : transactions) {
+    if (t == nextTransaction) {
+      mBufferTransactions.AppendElement(nextTransaction);
+      addedNext = true;
+      continue;
+    }
+    MOZ_DIAGNOSTIC_ASSERT(!t->IsDeleted());
+    // Remove detached transactions from unused buffers.
+    if (t->IsDetached() && !t->MatchesBuffer(mLatestAttachedBuffer)) {
+      t->DeleteTransactionLocked(aProofOfLock);
+    } else {
+      mBufferTransactions.AppendElement(t);
     }
   }
-
-  // It's possible that buffer was already freed by previous detach call
-  // and we're on synced delete now. In such case just quit.
-  // Reversed order (delete, detach) is not possible - we can't get detach
-  // for deleted buffers.
-  MOZ_DIAGNOSTIC_ASSERT(
-      aWlBufferDelete,
-      "Wayland compositor detach call after wl_buffer delete?");
+  if (!addedNext) {
+    mBufferTransactions.AppendElement(nextTransaction);
+  }
+  return nextTransaction;
 }
 
-static void BufferDetachedCallbackHandler(void* aData, wl_buffer* aBuffer) {
-  LOGS("BufferDetachedCallbackHandler() [%p] received wl_buffer [%p]", aData,
-       aBuffer);
-  RefPtr surface = static_cast<WaylandSurface*>(aData);
-  surface->BufferFreeCallbackHandler(reinterpret_cast<uintptr_t>(aBuffer),
-                                     /* aWlBufferDelete */ false);
+bool WaylandSurface::IsBufferAttached(WaylandBuffer* aBuffer) {
+  return mLatestAttachedBuffer == reinterpret_cast<uintptr_t>(aBuffer) &&
+         mBufferAttached;
 }
 
-static const struct wl_buffer_listener sBufferDetachListener = {
-    BufferDetachedCallbackHandler};
-
-bool WaylandSurface::AttachLocked(WaylandSurfaceLock& aSurfaceLock,
-                                  RefPtr<WaylandBuffer> aWaylandBuffer) {
+bool WaylandSurface::AttachLocked(const WaylandSurfaceLock& aSurfaceLock,
+                                  RefPtr<WaylandBuffer> aBuffer) {
   MOZ_DIAGNOSTIC_ASSERT(&aSurfaceLock == mSurfaceLock);
   MOZ_DIAGNOSTIC_ASSERT(mSurface);
 
   auto scale = GetScaleSafe();
-  LayoutDeviceIntSize bufferSize = aWaylandBuffer->GetSize();
+  LayoutDeviceIntSize bufferSize = aBuffer->GetSize();
   // TODO: rounding?
   SetSizeLocked(aSurfaceLock, gfx::IntSize(bufferSize.width, bufferSize.height),
                 gfx::IntSize((int)round(bufferSize.width / scale),
                              (int)round(bufferSize.height / scale)));
 
   LOGWAYLAND(
-      "WaylandSurface::AttachLocked() WaylandBuffer [%p] size [%d x %d] "
-      "fractional scale %f",
-      aWaylandBuffer.get(), bufferSize.width, bufferSize.height, scale);
+      "WaylandSurface::AttachLocked() transactions [%d] WaylandBuffer [%p] "
+      "attached [%d] size [%d x %d] fractional scale %f",
+      (int)mBufferTransactions.Length(), aBuffer.get(), aBuffer->IsAttached(),
+      bufferSize.width, bufferSize.height, scale);
 
-  wl_buffer* buffer = aWaylandBuffer->BorrowBuffer(aSurfaceLock);
-  if (!buffer) {
-    LOGWAYLAND("WaylandSurface::AttachLocked() failed, BorrowBuffer() failed");
+  auto* transaction = GetNextTransactionLocked(aSurfaceLock, aBuffer);
+  if (!transaction) {
     return false;
   }
 
-  // We don't take reference to this. Some compositors doesn't send
-  // buffer release callback and we may leak WaylandSurface then.
-  // Rather we destroy wl_buffer at end which makes sure no release callback
-  // comes after WaylandSurface release.
-  if (wl_buffer_add_listener(buffer, &sBufferDetachListener, this) < 0) {
-    LOGWAYLAND(
-        "WaylandSurface::AttachLocked() failed to attach buffer listener");
-    aWaylandBuffer->ReturnBufferDetached(aSurfaceLock);
-    return false;
-  }
-
-  if (!mAttachedBuffers.Contains(aWaylandBuffer)) {
-    mAttachedBuffers.AppendElement(aWaylandBuffer);
-  }
-
-  wl_surface_attach(mSurface, buffer, 0, 0);
-  aWaylandBuffer->SetAttachedLocked(aSurfaceLock);
-  mBufferAttached = true;
+  wl_surface_attach(mSurface, transaction->BufferBorrowLocked(aSurfaceLock), 0,
+                    0);
+  mLatestAttachedBuffer = reinterpret_cast<uintptr_t>(aBuffer.get());
   mSurfaceNeedsCommit = true;
+  mBufferAttached = true;
   return true;
 }
 
 void WaylandSurface::RemoveAttachedBufferLocked(
-    WaylandSurfaceLock& aSurfaceLock) {
+    const WaylandSurfaceLock& aSurfaceLock) {
   MOZ_DIAGNOSTIC_ASSERT(&aSurfaceLock == mSurfaceLock);
   MOZ_DIAGNOSTIC_ASSERT(mSurface);
 
   LOGWAYLAND("WaylandSurface::RemoveAttachedBufferLocked()");
 
-  // Set our size according to attached buffer
   SetSizeLocked(aSurfaceLock, gfx::IntSize(0, 0), gfx::IntSize(0, 0));
   wl_surface_attach(mSurface, nullptr, 0, 0);
+  mLatestAttachedBuffer = 0;
   mSurfaceNeedsCommit = true;
   mBufferAttached = false;
 }
@@ -1274,8 +1288,7 @@ void WaylandSurface::SetParentLocked(const WaylandSurfaceLock& aProofOfLock,
 void WaylandSurface::ImageDescriptionFailed(
     void* aData, struct wp_image_description_v1* aImageDescription,
     uint32_t aCause, const char* aMsg) {
-  RefPtr waylandSurface =
-      already_AddRefed(reinterpret_cast<WaylandSurface*>(aData));
+  RefPtr waylandSurface = dont_AddRef(static_cast<WaylandSurface*>(aData));
   WaylandSurfaceLock lock(waylandSurface);
   waylandSurface->mHDRSet = false;
   LOGS("[%p] WaylandSurface::ImageDescriptionFailed()",
@@ -1331,4 +1344,52 @@ bool WaylandSurface::EnableColorManagementLocked(
   return true;
 }
 
+static int YUVColorSpaceToWLColorCoeficients(
+    mozilla::gfx::YUVColorSpace aColorSpace) {
+  switch (aColorSpace) {
+    case gfx::YUVColorSpace::BT601:
+      return WP_COLOR_REPRESENTATION_SURFACE_V1_COEFFICIENTS_BT601;
+    case gfx::YUVColorSpace::BT709:
+      return WP_COLOR_REPRESENTATION_SURFACE_V1_COEFFICIENTS_BT709;
+    case gfx::YUVColorSpace::BT2020:
+      return WP_COLOR_REPRESENTATION_SURFACE_V1_COEFFICIENTS_BT2020;
+    default:
+      MOZ_DIAGNOSTIC_CRASH("Unsupported YUV color space!");
+      return 0;
+  }
+}
+
+void WaylandSurface::SetColorRepresentationLocked(
+    const WaylandSurfaceLock& aProofOfLock,
+    mozilla::gfx::YUVColorSpace aColorSpace, bool aFullRange) {
+  auto* colorRepresentation =
+      WaylandDisplayGet()->GetColorRepresentationManager();
+  if (!colorRepresentation) {
+    return;
+  }
+
+  auto coefficients = YUVColorSpaceToWLColorCoeficients(aColorSpace);
+  if (!coefficients) {
+    return;
+  }
+  auto range = WaylandDisplayGet()->GetColorRange(coefficients, aFullRange);
+  if (!range) {
+    return;
+  }
+
+  LOGWAYLAND(
+      "WaylandSurface::SetColorRepresentationLocked() colorspace %s full range "
+      "%d",
+      YUVColorSpaceToString(aColorSpace), aFullRange);
+  MOZ_DIAGNOSTIC_ASSERT(!mColorRepresentationSurface);
+  mColorRepresentationSurface = wp_color_representation_manager_v1_get_surface(
+      colorRepresentation, mSurface);
+  wp_color_representation_surface_v1_set_coefficients_and_range(
+      mColorRepresentationSurface, coefficients, range);
+}
+
+void WaylandSurface::AssertCurrentThreadOwnsMutex() {
+  mMutex.AssertCurrentThreadOwns();
+}
+
 }  // namespace mozilla::widget
