@@ -31,7 +31,7 @@
 #include "nsColor.h"
 #include "nsDebug.h"
 #include "nsID.h"
-#include "nsIObserver.h"
+#include "nsIDOMWindowUtils.h"
 #include "nsISupports.h"
 #include "nsITheme.h"
 #include "nsITimer.h"
@@ -82,6 +82,7 @@ namespace layers {
 class AsyncDragMetrics;
 class Compositor;
 class CompositorBridgeChild;
+struct CompositorScrollUpdate;
 struct FrameMetrics;
 class LayerManager;
 class WebRenderBridgeChild;
@@ -139,9 +140,6 @@ typedef void* nsNativeWidget;
 #define NS_RAW_NATIVE_IME_CONTEXT 14
 #define NS_NATIVE_WINDOW_WEBRTC_DEVICE_ID 15
 #ifdef XP_WIN
-#  define NS_NATIVE_TSF_THREAD_MGR 100
-#  define NS_NATIVE_TSF_CATEGORY_MGR 101
-#  define NS_NATIVE_TSF_DISPLAY_ATTR_MGR 102
 #  define NS_NATIVE_ICOREWINDOW 103  // winrt specific
 #endif
 #if defined(MOZ_WIDGET_GTK)
@@ -277,48 +275,48 @@ struct SizeConstraints {
   mozilla::DesktopToLayoutDeviceScale mScale;
 };
 
-struct AutoObserverNotifier {
-  AutoObserverNotifier(nsIObserver* aObserver, const char* aTopic)
-      : mObserver(aObserver), mTopic(aTopic) {}
+class MOZ_RAII AutoSynthesizedEventCallbackNotifier final {
+ public:
+  explicit AutoSynthesizedEventCallbackNotifier(
+      nsISynthesizedEventCallback* aCallback)
+      : mCallback(aCallback) {}
 
-  void SkipNotification() { mObserver = nullptr; }
+  void SkipNotification() { mCallback = nullptr; }
 
-  uint64_t SaveObserver() {
-    if (!mObserver) {
-      return 0;
+  Maybe<uint64_t> SaveCallback() {
+    if (!mCallback) {
+      return Nothing();
     }
-    uint64_t observerId = ++sObserverId;
-    sSavedObservers.InsertOrUpdate(observerId, mObserver);
+    uint64_t callbackId = ++sCallbackId;
+    sSavedCallbacks.InsertOrUpdate(callbackId, mCallback);
     SkipNotification();
-    return observerId;
+    return Some(callbackId);
   }
 
-  ~AutoObserverNotifier() {
-    if (mObserver) {
-      mObserver->Observe(nullptr, mTopic, nullptr);
+  ~AutoSynthesizedEventCallbackNotifier() {
+    if (mCallback) {
+      mCallback->OnCompleteDispatch();
     }
   }
 
-  static void NotifySavedObserver(const uint64_t& aObserverId,
-                                  const char* aTopic) {
-    nsCOMPtr<nsIObserver> observer = sSavedObservers.Get(aObserverId);
-    if (!observer) {
-      MOZ_ASSERT(aObserverId == 0,
-                 "We should always find a saved observer for nonzero IDs");
+  static void NotifySavedCallback(const uint64_t& aCallbackId) {
+    MOZ_ASSERT(aCallbackId > 0, "Callback ID must be non-zero");
+
+    auto entry = sSavedCallbacks.Extract(aCallbackId);
+    if (!entry) {
+      MOZ_ASSERT_UNREACHABLE("We should always find a saved callback");
       return;
     }
 
-    sSavedObservers.Remove(aObserverId);
-    observer->Observe(nullptr, aTopic, nullptr);
+    entry.value()->OnCompleteDispatch();
   }
 
  private:
-  nsCOMPtr<nsIObserver> mObserver;
-  const char* mTopic;
+  nsCOMPtr<nsISynthesizedEventCallback> mCallback;
 
- private:
-  static uint64_t sObserverId;
-  static nsTHashMap<uint64_t, nsCOMPtr<nsIObserver>> sSavedObservers;
+  static uint64_t sCallbackId;
+  static nsTHashMap<uint64_t, nsCOMPtr<nsISynthesizedEventCallback>>
+      sSavedCallbacks;
 };
 
 }  // namespace mozilla::widget
@@ -393,7 +391,7 @@ class nsIWidget : public nsISupports {
         : mType(aType), mRect(aRect) {}
   };
 
-  NS_DECLARE_STATIC_IID_ACCESSOR(NS_IWIDGET_IID)
+  NS_INLINE_DECL_STATIC_IID(NS_IWIDGET_IID)
 
   nsIWidget() = default;
 
@@ -566,6 +564,8 @@ class nsIWidget : public nsISupports {
   virtual mozilla::DesktopToLayoutDeviceScale
   GetDesktopToDeviceScaleByScreen() = 0;
 
+  virtual void DynamicToolbarOffsetChanged(mozilla::ScreenIntCoord aOffset) = 0;
+
   /**
    * Return the default scale factor for the window. This is the
    * default number of device pixels per CSS pixel to use. This should
@@ -1449,7 +1449,7 @@ class nsIWidget : public nsISupports {
    * @param aUnmodifiedCharacters characters that the OS would decide
    * to generate from the event if modifier keys (other than shift)
    * were assumed inactive. Needed on Mac, ignored on Windows.
-   * @param aObserver the observer that will get notified once the events
+   * @param aCallback the callback that will get notified once the events
    * have been dispatched.
    * @return NS_ERROR_NOT_AVAILABLE to indicate that the keyboard
    * layout is not supported and the event was not fired
@@ -1457,7 +1457,8 @@ class nsIWidget : public nsISupports {
   virtual nsresult SynthesizeNativeKeyEvent(
       int32_t aNativeKeyboardLayout, int32_t aNativeKeyCode,
       uint32_t aModifierFlags, const nsAString& aCharacters,
-      const nsAString& aUnmodifiedCharacters, nsIObserver* aObserver) = 0;
+      const nsAString& aUnmodifiedCharacters,
+      nsISynthesizedEventCallback* aCallback) = 0;
 
   /**
    * Utility method intended for testing. Dispatches native mouse events
@@ -1473,7 +1474,7 @@ class nsIWidget : public nsISupports {
    * @param aModifierFlags Some values of nsIWidget::Modifiers.
    *                       FYI: On Windows, Android and Headless widget on all
    *                       platroms, this hasn't been handled yet.
-   * @param aObserver the observer that will get notified once the events
+   * @param aCallback the callback that will get notified once the events
    * have been dispatched.
    */
   enum class NativeMouseMessage : uint32_t {
@@ -1486,17 +1487,17 @@ class nsIWidget : public nsISupports {
   virtual nsresult SynthesizeNativeMouseEvent(
       LayoutDeviceIntPoint aPoint, NativeMouseMessage aNativeMessage,
       mozilla::MouseButton aButton, nsIWidget::Modifiers aModifierFlags,
-      nsIObserver* aObserver) = 0;
+      nsISynthesizedEventCallback* aCallback) = 0;
 
   /**
    * A shortcut to SynthesizeNativeMouseEvent, abstracting away the native
    * message. aPoint is location in device pixels to which the mouse pointer
    * moves to.
-   * @param aObserver the observer that will get notified once the events
+   * @param aCallback the callback that will get notified once the events
    * have been dispatched.
    */
-  virtual nsresult SynthesizeNativeMouseMove(LayoutDeviceIntPoint aPoint,
-                                             nsIObserver* aObserver) = 0;
+  virtual nsresult SynthesizeNativeMouseMove(
+      LayoutDeviceIntPoint aPoint, nsISynthesizedEventCallback* aCallback) = 0;
 
   /**
    * Utility method intended for testing. Dispatching native mouse scroll
@@ -1518,13 +1519,13 @@ class nsIWidget : public nsISupports {
    * @param aModifierFlags    Must be values of Modifiers, or zero.
    * @param aAdditionalFlags  See nsIDOMWidnowUtils' consts and their
    *                          document.
-   * @param aObserver         The observer that will get notified once the
+   * @param aCallback         The callback that will get notified once the
    *                          events have been dispatched.
    */
   virtual nsresult SynthesizeNativeMouseScrollEvent(
       LayoutDeviceIntPoint aPoint, uint32_t aNativeMessage, double aDeltaX,
       double aDeltaY, double aDeltaZ, uint32_t aModifierFlags,
-      uint32_t aAdditionalFlags, nsIObserver* aObserver) = 0;
+      uint32_t aAdditionalFlags, nsISynthesizedEventCallback* aCallback) = 0;
 
   /*
    * TouchPointerState states for SynthesizeNativeTouchPoint. Match
@@ -1567,15 +1568,13 @@ class nsIWidget : public nsISupports {
    * @param aPressure 0.0 -> 1.0 float val indicating pressure
    * @param aOrientation 0 -> 359 degree value indicating the
    * orientation of the pointer. Use 90 for normal taps.
-   * @param aObserver The observer that will get notified once the events
+   * @param aCallback The callback that will get notified once the events
    * have been dispatched.
    */
-  virtual nsresult SynthesizeNativeTouchPoint(uint32_t aPointerId,
-                                              TouchPointerState aPointerState,
-                                              LayoutDeviceIntPoint aPoint,
-                                              double aPointerPressure,
-                                              uint32_t aPointerOrientation,
-                                              nsIObserver* aObserver) = 0;
+  virtual nsresult SynthesizeNativeTouchPoint(
+      uint32_t aPointerId, TouchPointerState aPointerState,
+      LayoutDeviceIntPoint aPoint, double aPointerPressure,
+      uint32_t aPointerOrientation, nsISynthesizedEventCallback* aCallback) = 0;
   /*
    * See nsIDOMWindowUtils.sendNativeTouchpadPinch().
    */
@@ -1588,29 +1587,18 @@ class nsIWidget : public nsISupports {
    * aLongTap is true, simulates a native long tap with a duration equal to
    * ui.click_hold_context_menus.delay. This pref is compatible with the
    * apzc long tap duration. Defaults to 1.5 seconds.
-   * @param aObserver The observer that will get notified once the events
+   * @param aCallback The callback that will get notified once the events
    * have been dispatched.
    */
-  virtual nsresult SynthesizeNativeTouchTap(LayoutDeviceIntPoint aPoint,
-                                            bool aLongTap,
-                                            nsIObserver* aObserver);
+  virtual nsresult SynthesizeNativeTouchTap(
+      LayoutDeviceIntPoint aPoint, bool aLongTap,
+      nsISynthesizedEventCallback* aCallback);
 
-  virtual nsresult SynthesizeNativePenInput(uint32_t aPointerId,
-                                            TouchPointerState aPointerState,
-                                            LayoutDeviceIntPoint aPoint,
-                                            double aPressure,
-                                            uint32_t aRotation, int32_t aTiltX,
-                                            int32_t aTiltY, int32_t aButton,
-                                            nsIObserver* aObserver) = 0;
-
-  /*
-   * Cancels all active simulated touch input points and pending long taps.
-   * Native widgets should track existing points such that they can clear the
-   * digitizer state when this call is made.
-   * @param aObserver The observer that will get notified once the touch
-   * sequence has been cleared.
-   */
-  virtual nsresult ClearNativeTouchSequence(nsIObserver* aObserver);
+  virtual nsresult SynthesizeNativePenInput(
+      uint32_t aPointerId, TouchPointerState aPointerState,
+      LayoutDeviceIntPoint aPoint, double aPressure, uint32_t aRotation,
+      int32_t aTiltX, int32_t aTiltY, int32_t aButton,
+      nsISynthesizedEventCallback* aCallback) = 0;
 
   /*
    * Send a native event as if the user double tapped the touchpad with two
@@ -1622,11 +1610,10 @@ class nsIWidget : public nsISupports {
   /*
    * See nsIDOMWindowUtils.sendNativeTouchpadPan().
    */
-  virtual nsresult SynthesizeNativeTouchpadPan(TouchpadGesturePhase aEventPhase,
-                                               LayoutDeviceIntPoint aPoint,
-                                               double aDeltaX, double aDeltaY,
-                                               int32_t aModifierFlags,
-                                               nsIObserver* aObserver) = 0;
+  virtual nsresult SynthesizeNativeTouchpadPan(
+      TouchpadGesturePhase aEventPhase, LayoutDeviceIntPoint aPoint,
+      double aDeltaX, double aDeltaY, int32_t aModifierFlags,
+      nsISynthesizedEventCallback* aCallback) = 0;
 
   virtual void StartAsyncScrollbarDrag(
       const AsyncDragMetrics& aDragMetrics) = 0;
@@ -1693,17 +1680,18 @@ class nsIWidget : public nsISupports {
   class LongTapInfo {
    public:
     LongTapInfo(int32_t aPointerId, LayoutDeviceIntPoint& aPoint,
-                mozilla::TimeDuration aDuration, nsIObserver* aObserver)
+                mozilla::TimeDuration aDuration,
+                nsISynthesizedEventCallback* aCallback)
         : mPointerId(aPointerId),
           mPosition(aPoint),
           mDuration(aDuration),
-          mObserver(aObserver),
+          mCallback(aCallback),
           mStamp(mozilla::TimeStamp::Now()) {}
 
     int32_t mPointerId;
     LayoutDeviceIntPoint mPosition;
     mozilla::TimeDuration mDuration;
-    nsCOMPtr<nsIObserver> mObserver;
+    nsCOMPtr<nsISynthesizedEventCallback> mCallback;
     mozilla::TimeStamp mStamp;
   };
 
@@ -1944,11 +1932,6 @@ class nsIWidget : public nsISupports {
    */
   virtual void ClearCachedWebrenderResources() {}
 
-  /**
-   * Clear WebRender animation resources
-   */
-  virtual void ClearWebrenderAnimationResources() {}
-
   /**
    * Request fast snapshot at RenderCompositor of WebRender.
    * Since readback of Windows DirectComposition is very slow.
@@ -2012,6 +1995,13 @@ class nsIWidget : public nsISupports {
     MOZ_ASSERT(false, "This function should only execute in Windows");
   }
 
+  /**
+   * NotifyCompositorScrollUpdate notify widget about an update to the
+   * composited scroll offset and zoom
+   */
+  virtual void NotifyCompositorScrollUpdate(
+      const mozilla::layers::CompositorScrollUpdate& aUpdate) = 0;
+
 #if defined(MOZ_WIDGET_ANDROID)
   /**
    * RecvToolbarAnimatorMessageFromCompositor receive message from compositor
@@ -2021,16 +2011,6 @@ class nsIWidget : public nsISupports {
    */
   virtual void RecvToolbarAnimatorMessageFromCompositor(int32_t aMessage) = 0;
 
-  /**
-   * UpdateRootFrameMetrics steady state frame metrics send from compositor
-   * thread
-   *
-   * @param aScrollOffset  page scroll offset value in screen pixels.
-   * @param aZoom          current page zoom.
-   */
-  virtual void UpdateRootFrameMetrics(const ScreenPoint& aScrollOffset,
-                                      const CSSToScreenScale& aZoom) = 0;
-
   /**
    * RecvScreenPixels Buffer containing the pixel from the frame buffer. Used
    * for android robocop tests.
@@ -2099,10 +2079,8 @@ class nsIWidget : public nsISupports {
   nsIWidget* MOZ_NON_OWNING_REF mParent = nullptr;
   // When Destroy() is called, the sub class should set this true.
   bool mOnDestroyCalled = false;
-  WindowType mWindowType = WindowType::Child;
+  WindowType mWindowType = WindowType::TopLevel;
   WidgetType mWidgetType = WidgetType::Native;
 };
 
-NS_DEFINE_STATIC_IID_ACCESSOR(nsIWidget, NS_IWIDGET_IID)
-
 #endif  // nsIWidget_h__
